/*
 *
 *                 #####    #####   ######  ######  ###   ###
 *               ##   ##  ##   ##  ##      ##      ## ### ##
 *              ##   ##  ##   ##  ####    ####    ##  #  ##
 *             ##   ##  ##   ##  ##      ##      ##     ##
 *            ##   ##  ##   ##  ##      ##      ##     ##
 *            #####    #####   ##      ######  ##     ##
 *
 *
 *             OOFEM : Object Oriented Finite Element Code
 *
 *               Copyright (C) 1993 - 2013   Borek Patzak
 *
 *
 *
 *       Czech Technical University, Faculty of Civil Engineering,
 *   Department of Structural Mechanics, 166 29 Prague, Czech Republic
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "idm1sla.h"
#include "isolinearelasticmaterial.h"
#include "gausspoint.h"
#include "floatmatrix.h"
#include "floatarray.h"
#include "datastream.h"
#include "contextioerr.h"
#include "mathfem.h"
#include "strainvector.h"
#include "stressvector.h"
#include "classfactory.h"
#include "dynamicinputrecord.h"

namespace oofem {
REGISTER_Material(IsotropicDamageMaterial1Sla);


IsotropicDamageMaterial1Sla :: IsotropicDamageMaterial1Sla(int n, Domain *d) : IsotropicDamageMaterial1(n, d),
                               SlaMaterialExtensionInterface()
    //
    // constructor
    //
{

}

IRResultType
IsotropicDamageMaterial1Sla :: initializeFrom(InputRecord *ir)
{
    IsotropicDamageMaterial1 :: initializeFrom(ir);

    SlaMaterialExtensionInterface :: initializeFrom(ir);

    this->mapper.initializeFrom(ir);

    return IRRT_OK;
}


double
IsotropicDamageMaterial1Sla::computeLambda(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep, double minLambda)
{
    if( this->softType == ST_Disable_Damage ) {
        return INFINITY;
    }

    // lambda is between 0 and DBL_MAX, but because we load with full unscaled bc => max value of lambda used = 1
    double minL;
    if( minLambda == INFINITY ) {
        minL = 1.01;
    } else {
        minL = minLambda;
    }

    this->initSLA(gp,totalStrain, tStep);

    // finding lambda by division of interval
    IsotropicDamageMaterial1StatusSla *status = static_cast< IsotropicDamageMaterial1StatusSla * >( this->giveStatus(gp) );

    int nIter;
    double kappa, dLambda = minL;
    double targetKappa = status->giveTempKappa();
    FloatArray stress;
    // this->initSLA(gp,totalStrain, tStep);
    FloatArray scaledStrain(status->giveTempStrainVector());
    FloatArray dStrain(totalStrain);
    dStrain.subtract(status->giveTempStrainVector());

    if( dStrain.computeNorm() < EPS_SLA ) {
        return INFINITY;
    }

    nIter = 0;
    while( nIter < MAX_ITER_SLA ) {
        nIter++;

        scaledStrain = status->giveTempStrainVector();
        scaledStrain.add(minL,dStrain);

        computeEquivalentStrain(kappa,scaledStrain,gp,tStep);

        if( (minL >= minLambda || minL > 1.0) && kappa < targetKappa ) {
            minL = INFINITY;
            break;
        } else if( fabs(kappa-targetKappa) < EPS_SLA ){
            break;
        } else if( kappa > targetKappa ) {
            dLambda /= 2.;
            minL -= dLambda;
        } else {
            dLambda /= 2.;
            minL += dLambda;
        }
    }

    if( nIter == MAX_ITER_SLA ) {
        OOFEM_ERROR( "idm1sla :: computeLambda :: maximum number of iterations exceeded")
    }

    return minL;
}


void
IsotropicDamageMaterial1Sla::initSLA(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep)
{
    IsotropicDamageMaterial1StatusSla *status = static_cast< IsotropicDamageMaterial1StatusSla * >( this->giveStatus(gp) );

    if( status->giveStrengthAt(1) >= 0. ) {
        return;
    }

    const double e0 = this->give(e0_ID, gp);  // e0 is the strain at the peak stress
    const double E = this->giveLinearElasticMaterial()->give('E', gp);

    // setup of element size if needed
    if ( isCrackBandApproachUsed() ) {
        initDamaged(INFINITY,totalStrain,gp);
    }

    switch ( damageLaw ) {
    case 0:     // exponential softening - default
        status->letStrengthAtBe((1.0+giveStrengthRange())*E*e0,1); // simplified should be changed to make it correct
        status->setTempKappa((1.0+giveStrengthRange())*e0);
        break;
    case 1:     // linear softening law
        status->letStrengthAtBe((1.0+giveStrengthRange())*E*e0,1);
        status->setTempKappa((1.0+giveStrengthRange())*e0);
        break;
    case 2:     // bilinear softening
        status->letStrengthAtBe((1.0+giveStrengthRange())*E*e0,1);
        status->setTempKappa((1.0+giveStrengthRange())*e0);
        break;
    case 3:     // reserved for Hordijk's law
        status->letStrengthAtBe((1.0+giveStrengthRange())*E*e0,1);
        status->setTempKappa((1.0+giveStrengthRange())*e0);
        break;
    case 4:     // Mazars damage law
        status->letStrengthAtBe((1.0+giveStrengthRange())*E*e0,1);
        status->setTempKappa((1.0+giveStrengthRange())*e0);
        break;
    case 5:     // Smooth exponential stress-strain law
        OOFEM_ERROR("Softening type is not implemented for Sequential linear analysis");
        break;
    case 6:     // Disabled damage
        status->letStrengthAtBe(INFINITY,1);
        status->setTempKappa(INFINITY);
        status->setTempDamage(0.0);
        break;
    case 7:     // Extended smooth damage law
        status->letStrengthAtBe((1.0+giveStrengthRange())*E*e0,1);
        status->setTempKappa((1.0+giveStrengthRange())*e0);
        break;
    default:
        OOFEM_ERROR("Softening type number %d is unknown", damageLaw);
    }

    status->updateYourself(tStep);
}


void
IsotropicDamageMaterial1Sla :: giveRealStressVector(FloatArray &answer, GaussPoint *gp,
                                                    const FloatArray &totalStrain,
                                                    TimeStep *tStep)
//
// returns real stress vector in 3d stress space of receiver according to
// previous level of stress and current
// strain increment, the only way, how to correctly update gp records
//
{
    IsotropicDamageMaterial1StatusSla *status = static_cast< IsotropicDamageMaterial1StatusSla * >( this->giveStatus(gp) );
    LinearElasticMaterial *lmat = this->giveLinearElasticMaterial();
    FloatArray reducedTotalStrainVector, reducedTotalStrainVectorOld;
    FloatMatrix de;
    double omega = 0.0;

    // subtract stress independent part
    // note: eigenStrains (temperature) is not contained in mechanical strain stored in gp
    // therefore it is necessary to subtract always the total eigen strain value
    this->giveStressDependentPartOfStrainVector(reducedTotalStrainVector, gp, totalStrain, tStep, VM_Total);
    this->giveStressDependentPartOfStrainVector(reducedTotalStrainVectorOld, gp, status->giveTempStrainVector(), tStep->givePreviousStep(), VM_Total);


    omega = status->giveTempDamage();

    lmat->giveStiffnessMatrix(de, SecantStiffness, gp, tStep);
    // damage deactivation in compression for 1D model
    if ( ( reducedTotalStrainVector.giveSize() > 1 ) || ( reducedTotalStrainVector.at(1) > 0. ) ) {
        de.times(1.0 - omega);
    }

    FloatArray dStrain = reducedTotalStrainVector;
    dStrain.subtract(reducedTotalStrainVectorOld);
    answer.beProductOf(de, dStrain);
    answer.add(status->giveTempStressVector());
}


void
IsotropicDamageMaterial1Sla :: updateHistoryParameters(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep, bool activeDamage)
{

    const double E = this->giveLinearElasticMaterial()->give('E', gp);
    IsotropicDamageMaterial1StatusSla *status = static_cast< IsotropicDamageMaterial1StatusSla * >( this->giveStatus(gp) );
    double newStrength;
    int rSize = totalStrain.giveSize();
    FloatArray stress(rSize), dStress(rSize), strain(rSize);
    FloatMatrix de,ce;

    this->giveRealStressVector(stress,gp,totalStrain,tStep);

    status->letTempStrainVectorBe(totalStrain);
    status->letTempStressVectorBe(stress);

    initSLA(gp,totalStrain,tStep);

    if( !activeDamage ) {
        status->updateYourself(tStep);
        return;
    }

    // setup of element size if needed
    if ( isCrackBandApproachUsed() ) {
        this->initDamaged(INFINITY,totalStrain,gp);
        if ( this->gf != 0. ) { //cohesive crack model
            if ( softType == ST_Exponential_Cohesive_Crack ) { // exponential softening
                wf = this->gf / E / e0; // wf is the crack opening
            } else if ( softType == ST_Linear_Cohesive_Crack || softType == ST_BiLinear_Cohesive_Crack ) { // (bi)linear softening law
                wf = 2. * gf / E / e0; // wf is the crack opening
            } else {
                OOFEM_ERROR("Gf unsupported for softening type softType = %d", softType);
            }
        } else if ( softType == ST_BiLinear_Cohesive_Crack ) {
            wf = this->wk/(e0*E-this->sk) * (e0*E);
        }


        const double Le = status->giveLe();
        ef = wf / Le;    //ef is the fracturing strain
        if ( ef < e0 ) { //check that no snapback occurs
            double minGf = 0.;
            OOFEM_WARNING("ef %e < e0 %e, this leads to material snapback in element %d, characteristic length %f", ef, e0, gp->giveElement()->giveNumber(), Le);
            if ( gf != 0. ) { //cohesive crack
                if ( softType == ST_Exponential_Cohesive_Crack ) { //exponential softening
                    minGf = E * e0 * e0 * Le;
                } else if ( softType == ST_Linear_Cohesive_Crack || softType == ST_BiLinear_Cohesive_Crack ) { //(bi)linear softening law
                    minGf = E * e0 * e0 * Le / 2.;
                } else {
                    OOFEM_WARNING("Gf unsupported for softening type softType = %d", softType);
                }

                if ( checkSnapBack ) {
                    OOFEM_ERROR("Material number %d, decrease e0, or increase Gf from %f to Gf=%f", this->giveNumber(), gf, minGf);
                }
            }

            if ( checkSnapBack ) { //given fracturing strain
                OOFEM_ERROR("Material number %d, increase ef %f to minimum e0 %f", this->giveNumber(), ef, e0);
            }
        }
    }

    switch ( damageLaw ) {
        case 0:     // exponential softening - default
            OOFEM_ERROR("Softening type is not implemented for Sequential linear analysis");
            break;
        case 1:     // linear softening law
        {
            // upper limit = (1.0+giveStrengthRange())*ft + (e-e0)*Et
            // lower limit = (1.0-giveStrengthRange())*ft + (e-e0)*Et

            double kappa = status->giveTempKappa();
            double Ed = (1-status->giveTempDamage()) * E;
            const double ft = e0 * E;
            const double Et = -ft / (ef-e0);

            // sigma at given strain on the lower limit
            double sig = (1.0-giveStrengthRange())*ft + (kappa-e0)*Et;
            if( sig < 0 ) sig = 0.0;

            // new damage
            double omega = 1.0-sig/(kappa*E);

            if( omega >= 1.0 ) {
                omega = 1.0;
                kappa = INFINITY;       // to ensure that this will not be damaged anymore
                newStrength = INFINITY; // to ensure that this will not be damaged anymore
            } else {
                // kappa for new max strength
                Ed = (1.-omega) * E;
                kappa = ((1.0+giveStrengthRange())*ft - Et*e0) / (Ed-Et);
                newStrength = (1.0+giveStrengthRange())*ft + (kappa-e0)*Et;
            }

            status->setTempDamage(omega);
            status->setTempKappa(kappa);

            break;
        }
        case 2:     // bilinear softening
            OOFEM_ERROR("Softening type is not implemented for Sequential linear analysis");
            break;
        case 3:     // reserved for Hordijk's law
            OOFEM_ERROR("Softening type is not implemented for Sequential linear analysis");
            break;
        case 4:     // Mazars damage law
            OOFEM_ERROR("Softening type is not implemented for Sequential linear analysis");
            break;
        case 5:     // Smooth exponential stress-strain law
            OOFEM_ERROR("Softening type is not implemented for Sequential linear analysis");
            break;
        case 6:     // Disabled damage
            OOFEM_ERROR("Damage law doesn't allow any decrease of stiffnes, this should not happen");
            break;
        case 7:     // Extended smooth damage law
            OOFEM_ERROR("Softening type is not implemented for Sequential linear analysis");
            break;
        default:
            OOFEM_ERROR("Softening type number %d is unknown", damageLaw);
    }


    this->giveStiffnessMatrix(de,ElasticStiffness,gp,tStep);
    ce.beInverseOf(de);
    ce.times(1.0/(1.0-status->giveDamage()));
    de.times(1.0-status->giveTempDamage());

    strain.beProductOf(ce,stress);
    dStress.beProductOf(de,strain);
    dStress.subtract(stress);

    status->updateYourself(tStep);
    status->letStrengthAtBe(newStrength,1);
    status->letDStressVectorBe(dStress);

}


// status
MaterialStatus *
IsotropicDamageMaterial1Sla :: CreateStatus(GaussPoint *gp) const
{
    IsotropicDamageMaterial1StatusSla *answer = new IsotropicDamageMaterial1StatusSla(1, IsotropicDamageMaterial1Sla :: domain, gp);
    return answer;
}

MaterialStatus *
IsotropicDamageMaterial1Sla :: giveStatus(GaussPoint *gp) const
{
    MaterialStatus *status = static_cast< MaterialStatus * >( gp->giveMaterialStatus() );
    if ( status == NULL ) {
        // create a new one
        status = this->CreateStatus(gp);

        if ( status != NULL ) {
            gp->setMaterialStatus( status, this->giveNumber() );
            this->_generateStatusVariables(gp);
        }
    }

    return status;
}

void
IsotropicDamageMaterial1StatusSla :: printOutputAt(FILE *file, TimeStep *tStep)
{
    StructuralMaterialStatus :: printOutputAt(file, tStep);
    fprintf(file, "status { ");
    if ( this->damage > 0.0 ) {
        fprintf( file, "damage %f crackVector %f %f %f strength %f", this->damage, this->crackVector.at(1), this->crackVector.at(2), this->crackVector.at(3), this->giveStrengthAt(1) );
    }

    fprintf(file, "}\n");
}

}     // end namespace oofem
