/*
 *
 *                 #####    #####   ######  ######  ###   ###
 *               ##   ##  ##   ##  ##      ##      ## ### ##
 *              ##   ##  ##   ##  ####    ####    ##  #  ##
 *             ##   ##  ##   ##  ##      ##      ##     ##
 *            ##   ##  ##   ##  ##      ##      ##     ##
 *            #####    #####   ##      ######  ##     ##
 *
 *
 *             OOFEM : Object Oriented Finite Element Code
 *
 *               Copyright (C) 1993 - 2013   Borek Patzak
 *
 *
 *
 *       Czech Technical University, Faculty of Civil Engineering,
 *   Department of Structural Mechanics, 166 29 Prague, Czech Republic
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "intmatisodamagetable.h"
#include "gausspoint.h"
#include "floatmatrix.h"
#include "floatarray.h"
#include "mathfem.h"
#include "datastream.h"
#include "contextioerr.h"
#include "classfactory.h"
#include "dynamicinputrecord.h"

#include <algorithm>

namespace oofem {
REGISTER_Material(IntMatIsoDamageTable);

IntMatIsoDamageTable :: IntMatIsoDamageTable(int n, Domain *d) : IntMatIsoDamage(n, d){}



IntMatIsoDamageTable :: ~IntMatIsoDamageTable(){}




IRResultType
IntMatIsoDamageTable :: initializeFrom(InputRecord *ir)
{
    IRResultType result;                // Required by IR_GIVE_FIELD macro
    std :: ifstream is;
    int nbrOfLinesToRead;

    IR_GIVE_FIELD(ir, kn, _IFT_IntMatIsoDamageTable_kn);
    IR_GIVE_FIELD(ir, ks, _IFT_IntMatIsoDamageTable_ks);
    IR_GIVE_FIELD(ir, ft, _IFT_IntMatIsoDamageTable_ft);

    this->e0 = ft / kn;

    // Set value of weight factor for the influence of shear component of displacement jump on equivalent strain
    IR_GIVE_OPTIONAL_FIELD(ir, beta, _IFT_IntMatIsoDamageTable_beta);

    //Set value of weight factor for the influence of shear component of displacement jump on equivalent strain
    IR_GIVE_OPTIONAL_FIELD(ir, absNormal, _IFT_IntMatIsoDamage_absNormal);

    //Set value of weight factor for the influence of normal component of displacement jump on equivalent strain
    IR_GIVE_OPTIONAL_FIELD(ir, betaNormal, _IFT_IntMatIsoDamage_betaNormal);
    if( betaNormal != 1 ) {
      OOFEM_ERROR( "Must be checked if the softening is correctly defined with 'e0'" );
    }

    // Set type of the input data
    int dum = 0;
    IR_GIVE_OPTIONAL_FIELD(ir, dum, _IFT_IntMatIsoDamageTable_tableType);
    switch( dum ) {
    case TT_jump_normalTraction:
      tabType = TT_jump_normalTraction;
      break;
    default:
      tabType = TT_jump_damage;
    }

    //Set limit on the maximum isotropic damage parameter if needed
    IR_GIVE_OPTIONAL_FIELD(ir, maxOmega, _IFT_IntMatIsoDamageTable_maxOmega);
    maxOmega = min(maxOmega, 0.999999);
    maxOmega = max(maxOmega, 0.0);

    // Parse the table file
    IR_GIVE_FIELD(ir, tablename, _IFT_IntMatIsoDamageTable_tablename);

    is.open(tablename.c_str(), std :: ifstream :: in);
    if ( !is.is_open() ) {
        OOFEM_ERROR(" Can't open table file %s.", tablename.c_str());
    }

    // Read first line
    if ( is >> nbrOfLinesToRead ) {
        printf("NumberofLinestoRead: %d\n", nbrOfLinesToRead);
    } else {
        OOFEM_ERROR("IntMatIsoDamageTable :: initializeFrom: Error reading table file, first line should be "
                    "an integer stating how many strain damage pairs that exist in the file.");
    }

    // Insert a (0,0) pair
    switch( tabType ) {
    case TT_jump_damage:
      tableSecondColumn.resize(nbrOfLinesToRead + 1);
      tableFirstColumn.resize(nbrOfLinesToRead + 1);

      tableFirstColumn(0) = tableSecondColumn(0) = 0.0;

      for ( int i = 0; i < nbrOfLinesToRead; i++ ) {
          if ( !( is >> tableFirstColumn(i + 1) >> tableSecondColumn(i + 1) ) ) {
              OOFEM_ERROR("Error reading table file at line %d, expected a "
                           "strain damage pair.", i + 2);
          }

          if ( ( tableSecondColumn(i + 1) < tableSecondColumn(i) ) || ( tableFirstColumn(i + 1) < tableFirstColumn(i) ) ) {
              OOFEM_ERROR("Error reading table file at line %d, strain "
                           "and damage must be given in an increasing order and be positive.", i + 2);
          }
      }

      // We add e0 to the tableJumps since these should be given as the increase in
      // strain relative to e0.
      tableFirstColumn.add(e0);

      break;
    case TT_jump_normalTraction:
      tableSecondColumn.resize(nbrOfLinesToRead + 2);
      tableFirstColumn.resize(nbrOfLinesToRead + 2);

      tableFirstColumn(0) = 0.0;
      tableSecondColumn(0) = ft;

      for ( int i = 0; i < nbrOfLinesToRead; i++ ) {
          if ( !( is >> tableFirstColumn(i + 1) >> tableSecondColumn(i + 1) ) ) {
              OOFEM_ERROR("Error reading table file at line %d, expected a "
                           "jump normalTration pair.", i + 2);
          }

          if ( ( tableFirstColumn(i + 1) < tableFirstColumn(i) ) ) {
              OOFEM_ERROR("Error reading table file at line %d,"
                           "jump must be given in an increasing order and be positive.", i + 2);
          }

          if ( (tableSecondColumn(i + 1)-tableSecondColumn(i))/(tableFirstColumn(i + 1)-tableFirstColumn(i)) >
               tableSecondColumn(i)/(tableFirstColumn(i)+e0)) {
              OOFEM_ERROR("Error reading table file at line %d,"
                           "wrong slope of stiffness => decreasing damage.", i + 2);
          }

          if ( ( tableSecondColumn(i + 1) < 0.0 ) ) {
              OOFEM_ERROR("Error reading table file at line %d,"
                           "normalTraction should have only positive values.", i + 2);
          }
      }

      // We add e0 to the tableJumps since these should be given as the increase in
      // strain relative to e0.
      tableFirstColumn.add(e0);

      // Kappa = first column If out of bounds stress is kept constant => evolution of damage
      tableFirstColumn(nbrOfLinesToRead + 1) = INFINITY;
      tableSecondColumn(nbrOfLinesToRead + 1) = tableSecondColumn(nbrOfLinesToRead);

      break;
    default:
      OOFEM_ERROR( "Unknown tableType" );
    }

    return StructuralInterfaceMaterial :: initializeFrom(ir);
}


void
IntMatIsoDamageTable :: giveInputRecord(DynamicInputRecord &input)
{
    StructuralInterfaceMaterial :: giveInputRecord(input);

    input.setField(this->kn, _IFT_IntMatIsoDamageTable_kn);
    input.setField(this->ks, _IFT_IntMatIsoDamageTable_ks);
    input.setField(this->ft, _IFT_IntMatIsoDamageTable_ft);
    input.setField(this->tablename, _IFT_IntMatIsoDamageTable_tablename);
    input.setField(this->maxOmega, _IFT_IntMatIsoDamageTable_maxOmega);
    input.setField(this->beta, _IFT_IntMatIsoDamageTable_beta);
    input.setField(this->tabType, _IFT_IntMatIsoDamageTable_tableType);
}



void
IntMatIsoDamageTable :: computeDamageParam(double &omega, double kappa)
{
    double val;
    if ( kappa > this->e0 ) {
        // Linear interpolation between table values.

        if ( kappa >= tableFirstColumn.at( tableFirstColumn.giveSize() ) && tabType == TT_jump_damage ) {
            // If out of bounds damage is set to the last given damage value in the table
            omega = tableSecondColumn.at( tableSecondColumn.giveSize() );
        } else {
            // std::lower_bound uses binary search to find index with value bounding kappa from above
            int index = (int)(std :: lower_bound(tableFirstColumn.givePointer(), tableFirstColumn.givePointer() + tableFirstColumn.giveSize(), kappa) - tableFirstColumn.givePointer());

#if 0
            printf("e0 %lf\n", e0);
            printf( "sizeof %d\n", sizeof( tableJumps.givePointer() ) );
            printf("pointer: %d\n", index);
            printf("value: %lf\n", * index);
            printf( "Index found: %d\n", index - tableJumps.givePointer() );
            printf( "First index: %d\n", tableJumps.givePointer() );
            printf( "Last index: %d\n", tableJumps.givePointer() + tableJumps.giveSize() );
#endif

            // Pointer arithmetic to find the values used in interpolation
            double x0 = tableFirstColumn(index - 1);
            double x1 = tableFirstColumn(index );
            double y0 = tableSecondColumn(index - 1);
            double y1 = tableSecondColumn(index );

            // Interpolation formula
            val = y0 + ( y1 - y0 ) * ( kappa - x0 ) / ( x1 - x0 );

            switch( tabType ) {
            case TT_jump_damage:
              omega = val;
              break;
            case TT_jump_normalTraction:
              omega = 1.0 - val / (kappa*kn);
              break;
            default:
              OOFEM_ERROR( "Unknown tableType" );
            }
        }
    } else {
        omega = 0.0;
    }
}




} // end namespace oofem
