/*
 *
 *                 #####    #####   ######  ######  ###   ###
 *               ##   ##  ##   ##  ##      ##      ## ### ##
 *              ##   ##  ##   ##  ####    ####    ##  #  ##
 *             ##   ##  ##   ##  ##      ##      ##     ##
 *            ##   ##  ##   ##  ##      ##      ##     ##
 *            #####    #####   ##      ######  ##     ##
 *
 *
 *             OOFEM : Object Oriented Finite Element Code
 *
 *               Copyright (C) 1993 - 2013   Borek Patzak
 *
 *
 *
 *       Czech Technical University, Faculty of Civil Engineering,
 *   Department of Structural Mechanics, 166 29 Prague, Czech Republic
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "intmatisodamagetablesla.h"
#include "gausspoint.h"
#include "floatmatrix.h"
#include "floatarray.h"
#include "mathfem.h"
#include "datastream.h"
#include "contextioerr.h"
#include "classfactory.h"
#include "dynamicinputrecord.h"

#include <algorithm>

namespace oofem {
REGISTER_Material(IntMatIsoDamageTableSla);

IntMatIsoDamageTableSla :: IntMatIsoDamageTableSla(int n, Domain *d) : IntMatIsoDamageTable(n, d), SlaMaterialExtensionInterface()
{}

IRResultType
IntMatIsoDamageTableSla :: initializeFrom(InputRecord *ir)
{
  IntMatIsoDamageTable :: initializeFrom(ir);

  SlaMaterialExtensionInterface :: initializeFrom(ir);

  return IRRT_OK;
}


void
IntMatIsoDamageTableSla :: giveInputRecord(DynamicInputRecord &input)
{
  StructuralInterfaceMaterial :: giveInputRecord(input);

  IntMatIsoDamageTable :: giveInputRecord(input);

  SlaMaterialExtensionInterface :: giveInputRecord(input);
}


void
IntMatIsoDamageTableSla :: giveEngTraction_3d(FloatArray &answer, GaussPoint *gp,
                                              const FloatArray &jump, TimeStep *tStep)
//
// returns 3d traction vector
{
  FloatMatrix D;
  this->give3dStiffnessMatrix_Eng(D, SecantStiffness, gp, tStep);

  answer.beProductOf(D, jump);
}

void
IntMatIsoDamageTableSla :: give2dStiffnessMatrix_Eng(FloatMatrix &answer, MatResponseMode rMode,
                                                     GaussPoint *gp, TimeStep *tStep)
{
  double om;
  IntMatIsoDamageStatus *status = static_cast< IntMatIsoDamageStatus * >( this->giveStatus(gp) );


  if ( ( rMode == ElasticStiffness ) || ( rMode == SecantStiffness ) || ( rMode == TangentStiffness ) ) {
    // assemble eleastic stiffness
    answer.resize(2, 2);
    answer.zero();
    answer.at(1, 1) = ks;
    answer.at(2, 2) = kn;

    if ( rMode == ElasticStiffness ) {
      return;
    }

    if ( rMode == SecantStiffness ) {
      // Secant stiffness
      om = status->giveTempDamage();
      om = min(om, maxOmega);
      if ( om > 0 ) {
        answer.times(1.0 - om);
      }

      return;
    } else {
      // Tangent Stiffness
      OOFEM_ERROR( "Tangential matrix should not be used during SL analysis" );
    }
  }  else {
    OOFEM_ERROR("Unknown MatResponseMode");
  }
}


void
IntMatIsoDamageTableSla :: give3dStiffnessMatrix_Eng(FloatMatrix &answer, MatResponseMode rMode,
                                                     GaussPoint *gp, TimeStep *tStep)
{
  double om;
  IntMatIsoDamageStatus *status = static_cast< IntMatIsoDamageStatus * >( this->giveStatus(gp) );


  if ( ( rMode == ElasticStiffness ) || ( rMode == SecantStiffness ) || ( rMode == TangentStiffness ) ) {
    // assemble eleastic stiffness
    answer.resize(3, 3);
    answer.zero();
    answer.at(1, 1) = ks;
    answer.at(2, 2) = ks;
    answer.at(3, 3) = kn;

    if ( rMode == ElasticStiffness ) {
      return;
    }

    if ( rMode == SecantStiffness ) {
      // Secant stiffness
      om = status->giveTempDamage();
      om = min(om, maxOmega);
      if ( om > 0 ) {
        answer.times(1.0 - om);
      }

      return;
    } else {
      OOFEM_ERROR( "Tangential matrix should not be used during SL analysis" );
    }
  }  else {
    OOFEM_ERROR("Unknown MatResponseMode");
  }
}


double
IntMatIsoDamageTableSla :: computeLambda(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep, double minLambda)
{

  // totalStrain = jump (reduced form)
  // calculation done in 3D notation

  IntMatIsoDamageTableStatusSla *status = static_cast< IntMatIsoDamageTableStatusSla * >( this->giveStatus(gp) );

  FloatArray jump3D(3), jump3D_old(3), dJump3D(3);

  jump3D = giveJump3D(totalStrain);
  jump3D_old = status->giveTempJump();
  dJump3D.beDifferenceOf(jump3D, jump3D_old);

  if( (beta == 0.0 && dJump3D.at(3) < EPS_SLA) ||
      dJump3D.computeNorm() < EPS_SLA ) {
    return INFINITY;
  }

  this->initSLA(gp,totalStrain, tStep);

  double minL, targetKappa, kappa, dLambda;
  int nIter;
  if( minLambda == INFINITY ) {
    minL = 1.01;
  } else {
    minL = minLambda;
  }


  targetKappa = status->giveTempKappa();
  dLambda = minL;
  nIter = 0;
  while( nIter < MAX_ITER_SLA ) {
    nIter++;

    jump3D = jump3D_old;
    jump3D.add(minL,dJump3D);

    computeEquivalentJump(kappa, jump3D);
    if( (minL >= minLambda || minL > 1.0) && kappa < targetKappa ) {
      minL = INFINITY;
      break;
    } else if( fabs(kappa-targetKappa) < EPS_SLA ){
      break;
    } else if( kappa > targetKappa ) {
      dLambda /= 2.;
      minL -= dLambda;
    } else {
      dLambda /= 2.;
      minL += dLambda;
    }
  }

  if( nIter == MAX_ITER_SLA ) {
    OOFEM_ERROR( "IntMatIsoDamageTableSla :: computeLambda :: maximum number of iterations exceeded")
  }

  return minL;
}


void
IntMatIsoDamageTableSla :: initSLA(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep)
{
  IntMatIsoDamageTableStatusSla *status = static_cast< IntMatIsoDamageTableStatusSla * >( this->giveStatus(gp) );

  if( status->giveStrengthAt(1) >= 0. ) {
    return;
  }

  switch( tabType ) {
  case TT_jump_normalTraction: {
    double kt, kappa = e0;

    // std::upper_bound uses binary search to find index with value bounding kappa from above
    int index = (int)(std :: upper_bound(tableFirstColumn.givePointer(), tableFirstColumn.givePointer() + tableFirstColumn.giveSize(), kappa) - tableFirstColumn.givePointer());

    // Pointer arithmetic to find the values used in interpolation
    double x0 = tableFirstColumn(index - 1);
    double x1 = tableFirstColumn(index);
    double y0 = tableSecondColumn(index - 1);
    double y1 = tableSecondColumn(index);

    kt = ( y1 - y0 ) / ( x1 - x0 );
    kappa = ((y0+this->giveStrengthRange()*ft) - x0 * kt)/ (kn-kt);

    status->letStrengthAtBe(kappa*kn, 1);
    status->setTempKappa(kappa);

    break;
  }
  default:
    OOFEM_ERROR( "Unknown tableType or not implemented" );
  }

  status->updateYourself(tStep);
}


void
IntMatIsoDamageTableSla :: updateHistoryParameters(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep, bool activeDamage)
{

  IntMatIsoDamageTableStatusSla *status = static_cast< IntMatIsoDamageTableStatusSla * >( this->giveStatus(gp) );
  FloatArray jump3D(3), traction3D(3);

  jump3D = giveJump3D(totalStrain);
  giveEngTraction_3d(traction3D,gp,jump3D,tStep);

  // update gp
  status->letTempJumpBe(jump3D);
  status->letTempTractionBe(traction3D);

  this->initSLA(gp,totalStrain, tStep);

  if( !activeDamage ) {
    status->updateYourself(tStep);
    return;
  }

  // update damage
  updateDamageParam(status);
  giveEngTraction_3d(traction3D,gp,jump3D,tStep);

  // traction increment
  traction3D.subtract(status->giveTempTraction());
  status->letDStressVectorBe(traction3D);

  status->updateYourself(tStep);
}


FloatArray
IntMatIsoDamageTableSla :: giveJump3D(const FloatArray &jump)
{
  FloatArray jump3D(3);

  switch( jump.giveSize() ) {
  case 1:
    jump3D = { 0.0, 0.0, jump.at(3) };
    break;
  case 2:
    jump3D = { jump.at(1), 0.0, jump.at(2) };
    break;
  case 3:
    jump3D = jump;
    break;
  default:
    OOFEM_ERROR( "Unsupported dimension" );
  }

  return jump3D;
}


FloatArray
IntMatIsoDamageTableSla :: giveJumpReducedForm(const FloatArray &jump3D, const int nDim)
{
  FloatArray jump(nDim);

  switch( nDim ) {
  case 1:
    jump = { jump3D.at(3) };
    break;
  case 2:
    jump = { jump3D.at(1), jump3D.at(3) };
    break;
  case 3:
    jump = jump3D;
    break;
  default:
    OOFEM_ERROR( "Unsupported dimension" );
  }

  return jump;
}


void
IntMatIsoDamageTableSla :: updateDamageParam(IntMatIsoDamageTableStatusSla *status)
{
  double val, omega, kappa;
  kappa = status->giveTempKappa();

  // Linear interpolation between table values.

  // If out of bounds damage is set to the last given damage value in the table
  if ( kappa >= tableFirstColumn.at( tableFirstColumn.giveSize() ) ) {
    omega = tableSecondColumn.at( tableSecondColumn.giveSize() );
  } else {
    // std::lower_bound uses binary search to find index with value bounding kappa from above
    int index = (int)(std :: lower_bound(tableFirstColumn.givePointer(), tableFirstColumn.givePointer() + tableFirstColumn.giveSize(), kappa) - tableFirstColumn.givePointer());

    // Pointer arithmetic to find the values used in interpolation
    double x0 = tableFirstColumn(index - 1);
    double x1 = tableFirstColumn(index);
    double y0 = tableSecondColumn(index - 1);
    double y1 = tableSecondColumn(index);

    // Interpolation formula
    double kt;
    kt = ( y1 - y0 ) / ( x1 - x0 );
    val = (y0-this->giveStrengthRange()*ft) +  ( kappa - x0 ) * kt;

    switch( tabType ) {
    case TT_jump_normalTraction:
      omega = 1.0 - val / (kappa*kn);

      // kappa for new upper bound
      kappa = ((y0+this->giveStrengthRange()*ft) - kt * x0) / (kn*(1.-omega) - kt);

      while( index+1 < tableSecondColumn.giveSize() && kappa > tableFirstColumn(index) ) {
        index++;
        x0 = tableFirstColumn(index - 1);
        x1 = tableFirstColumn(index);
        y0 = tableSecondColumn(index - 1);
        y1 = tableSecondColumn(index);

        kt = ( y1 - y0 ) / ( x1 - x0 );
        kappa = ((y0+this->giveStrengthRange()*ft) - kt * x0) / (kn*(1.-omega) - kt);
      }

      break;
    default:
      OOFEM_ERROR( "Unknown tableType or not implemented" );
    }

    status->setTempKappa(kappa);
    status->setTempDamage(omega);
    status->letStrengthAtBe(kappa*(1.-omega)*kn, 1);

  }
}


/*****************************************************************/
void
IntMatIsoDamageTableStatusSla :: printOutputAt(FILE *file, TimeStep *tStep)
{
    StructuralInterfaceMaterialStatus :: printOutputAt(file, tStep);
    fprintf(file, "status { ");
    fprintf(file, "ts %f, ts %f, tn %f ", this->traction.at(1), this->traction.at(2), this->traction.at(3));

    if ( this->damage > 0.0 ) {
        fprintf(file, ", kappa %f, damage %f ", this->kappa, this->damage);
    }

    fprintf(file, "}\n");
}

} // end namespace oofem
