/*
 *
 *                 #####    #####   ######  ######  ###   ###
 *               ##   ##  ##   ##  ##      ##      ## ### ##
 *              ##   ##  ##   ##  ####    ####    ##  #  ##
 *             ##   ##  ##   ##  ##      ##      ##     ##
 *            ##   ##  ##   ##  ##      ##      ##     ##
 *            #####    #####   ##      ######  ##     ##
 *
 *
 *             OOFEM : Object Oriented Finite Element Code
 *
 *               Copyright (C) 1993 - 2013   Borek Patzak
 *
 *
 *
 *       Czech Technical University, Faculty of Civil Engineering,
 *   Department of Structural Mechanics, 166 29 Prague, Czech Republic
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sequentiallylinearstatic.h"
#include "timestep.h"
#include "dof.h"
#include "sparsemtrx.h"
#include "dictionary.h"
#include "verbose.h"
#include "classfactory.h"
#include "datastream.h"
#include "contextioerr.h"
#include "dofmanager.h"
#include "activebc.h"
#include "structuralelement.h"
#include "structuralinterfaceelement.h"
#include "structuralms.h"
#include "slamaterialextensioninterface.h"
#include "gausspoint.h"

namespace oofem {
REGISTER_EngngModel(SequentiallyLinearStatic);

SequentiallyLinearStatic :: SequentiallyLinearStatic(int i, EngngModel *_master) : IncrementalLinearStatic(i, _master)
{
    printIterations  = false;
    outputStreamIterations = NULL;
    fileNameIterations = "SLAiterations.out";
}

SequentiallyLinearStatic :: ~SequentiallyLinearStatic()
{
    if ( outputStreamIterations ) {
        fclose(outputStreamIterations);
    }
}


IRResultType
SequentiallyLinearStatic :: initializeFrom(InputRecord *ir)
{
    IRResultType result;

    IncrementalLinearStatic :: initializeFrom(ir);

    int dum = 0;
    IR_GIVE_OPTIONAL_FIELD(ir, dum, _IFT_SequentiallyLinearStatic_PrintIterations);
    printIterations = dum != 0;
    
    if( printIterations )
    {
        std::string dumStr;
        IR_GIVE_OPTIONAL_FIELD(ir, dumStr, _IFT_SequentiallyLinearStatic_FileName);
        if( fileNameIterations.compare(dumStr) != 0 ) {
            fileNameIterations = dumStr;
        }

        if ( ( outputStreamIterations = fopen(fileNameIterations.c_str(), "w") ) == NULL ) {
                OOFEM_ERROR("Can't open output file for iterations %s", fileNameIterations.c_str());
        }
    }
    return IRRT_OK;
}


void SequentiallyLinearStatic :: solveYourselfAt(TimeStep *tStep)
{
    // Creates system of governing eq's and solves them at given time step
    // Each step is devided into subSteps coresponding to incremental single element failure

    Domain *d = this->giveDomain(1);
    int nnodes = d->giveNumberOfDofManagers();
    int nelem = d->giveNumberOfElements();
    int i;
    double tu = 0.0; // total time during the step <0;1>

    FloatArray strain;
    FloatMatrix rotationMatG2L;
    IntegrationRule *iRule;
    GaussPoint *gp;
    DofManager *dofman;
    SlaMaterialExtensionInterface *slaMat;
    NM_Status s;
    bool increaseLoad;
    double dumLam, minLam = INFINITY;
    int dumEl=0,dumGp=0, avalanche;

    // set solver
    this->giveNumericalMethod( this->giveCurrentMetaStep() );


    // Initiates the total displacement to zero.
    if ( tStep->isTheFirstStep() ) {
        for ( i = 1; i <= nnodes; i++ ) {
            dofman = d->giveDofManager(i);
            for ( Dof *dof: *dofman ) {
                dof->updateUnknownsDictionary(tStep->givePreviousStep(), VM_Total, 0.);
                dof->updateUnknownsDictionary(tStep, VM_Total, 0.);
            }
        }

        int nbc = d->giveNumberOfBoundaryConditions();
        for ( int ibc = 1; ibc <= nbc; ++ibc ) {
            GeneralBoundaryCondition *bc = d->giveBc(ibc);
            ActiveBoundaryCondition *abc;

            if ( ( abc = dynamic_cast< ActiveBoundaryCondition * >(bc) ) ) {
                int ndman = abc->giveNumberOfInternalDofManagers();
                for ( int i = 1; i <= ndman; i++ ) {
                    dofman = abc->giveInternalDofManager(i);
                    for ( Dof *dof: *dofman ) {
                        dof->updateUnknownsDictionary(tStep->givePreviousStep(), VM_Total, 0.);
                        dof->updateUnknownsDictionary(tStep, VM_Total, 0.);
                    }
                }
            }
        }
    }


    int neq = this->giveNumberOfDomainEquations( 1, EModelDefaultEquationNumbering() );

    #ifdef VERBOSE
        OOFEM_LOG_RELEVANT("Solving [step number %8d, time %15e, equations %d]\n", tStep->giveNumber(), tStep->giveTargetTime(), neq);
    #endif



    increaseLoad = true;
    incrementOfDisplacementVector.resize(neq);
    internalLoadVector.resize(neq);
    loadVector.resize(neq);

    if ( stiffnessMatrix ) {
        delete stiffnessMatrix;
    }
    stiffnessMatrix = classFactory.createSparseMtrx(sparseMtrxType);
    if ( stiffnessMatrix == NULL ) {
        OOFEM_ERROR("sparse matrix creation failed");
    }
    stiffnessMatrix->buildInternalStructure( this, 1, EModelDefaultEquationNumbering() );

    #ifdef VERBOSE
        OOFEM_LOG_INFO("Assembling general stiffness matrix\n");
    #endif
    // assembling stiffness matrix, afterwards it will be only updated
    stiffnessMatrix->zero();
    this->assemble( stiffnessMatrix, tStep, SecantStiffnessMatrix, EModelDefaultEquationNumbering(), d );

    // solve
    #ifdef VERBOSE
        OOFEM_LOG_INFO("Calculating ...\n");
    #endif
    while ( tu < 1.0 )
    {
        // Apply dirichlet b.c's on total values
        if( increaseLoad )
        {
            avalanche = 0;
            tStep->incrementSubStepNumber();

            for ( i = 1; i <= nnodes; i++ ) {
                dofman = d->giveDofManager(i);
                for ( Dof *dof: *dofman ) {
                    double tot = dof->giveUnknown( VM_Total, tStep->givePreviousStep() );
                    if ( dof->hasBc(tStep) ) {
                        tot += dof->giveBcValue(VM_Incremental, tStep);
                    }

                    dof->updateUnknownsDictionary(tStep, VM_Total, tot);
                }
            }

            loadVector.zero();
            this->assembleVector( loadVector, tStep, ExternalForcesVector,
                                 VM_Total, EModelDefaultEquationNumbering(), d );

        } else {
            avalanche++;
        }

        if ( neq == 0 ) { // Allows for fully prescribed/empty problems.
            OOFEM_ERROR("Must be checked what is needed to be done!!!");
            return;
        }

        incrementOfDisplacementVector.zero();

        internalLoadVector.zero();
        this->assembleVector( internalLoadVector, tStep, InternalForcesVector,
                             VM_Total, EModelDefaultEquationNumbering(), d );

        loadVector.subtract(internalLoadVector);

        #ifdef __PARALLEL_MODE
            this->updateSharedDofManagers(loadVector, EModelDefaultEquationNumbering(), ReactionExchangeTag);
        #endif

        // calculate increments of displacements
        s = nMethod->solve(stiffnessMatrix, & loadVector, & incrementOfDisplacementVector);
        if ( !( s & NM_Success ) ) {
            OOFEM_ERROR("No success in solving system.");
        }

        // update increments in nodes - needed for calculation of stresses in IP
        for ( i = 1; i <= nnodes; i++ ) {
            updateDofUnknownsDictionary(d->giveDofManager(i), tStep);
        }

        // calculate lambda
        minLam = INFINITY;
        dumEl = dumGp = 0;
        for ( i = 1; i <= nelem; i++ ) {
            Element *elem = d->giveElement(i);
            slaMat = dynamic_cast<SlaMaterialExtensionInterface*>(elem->giveMaterial());
            if ( !slaMat ) {
                OOFEM_ERROR("SLA :: Material doesn't implement the required SlaMaterialExtension interface!");
            }

            for ( int j = 0; j < elem->giveNumberOfIntegrationRules(); j++ ) {
                iRule = elem->giveIntegrationRule(j);
                for ( int k = 0; k < iRule->giveNumberOfIntegrationPoints(); k++ ) {
                    gp = iRule->getIntegrationPoint(k);
                    giveDeformationLoadForElement(strain, rotationMatG2L, elem, gp, tStep);

                    if( rotationMatG2L.giveNumberOfRows() ) {
                      strain.rotatedWith(rotationMatG2L, 'n'); // transform to local coord system
                    }

                    dumLam = slaMat->computeLambda(gp, strain, tStep, minLam);

                    if( dumLam < minLam ) {
                        minLam = dumLam;
                        dumEl = i;
                        dumGp = k;

                        if( minLam == 0.0 ) { // other elements don't need to be checked
                          break;
                        }
                    }
                }
            }
        }


        // reassembling of loadVector
        loadVector.add(internalLoadVector);

        // increment of total time in the step
        if( increaseLoad ) {
            loadVector.times(minLam);

            if( (tu + minLam) > 1.0 ) { // no element damaged and full load applied
                minLam = 1.0 - tu;
                tu = 1.0;
                dumEl = 0;
                increaseLoad = true;
            } else {
                tu += minLam;
                increaseLoad = false;
            }

            #ifdef VERBOSE
                OOFEM_LOG_INFO("  Increment of load no. %d: sum lambda = %.3f\n", tStep->giveSubStepNumber(), tu);
            #endif


            // scaling of prescribed displacements (Dirichlet BC)
            for ( i = 1; i <= nnodes; i++ ) {
                dofman = d->giveDofManager(i);
                for ( Dof *dof: *dofman ) {
                    double tot = dof->giveUnknown( VM_Total, tStep );
                    if ( dof->hasBc(tStep) ) {
                        tot += (minLam-1.0)*dof->giveBcValue(VM_Incremental, tStep);
                    }

                    dof->updateUnknownsDictionary(tStep->givePreviousStep(), VM_Total, tot);
                    dof->updateUnknownsDictionary(tStep, VM_Total, tot);
                }
            }
        } else if( minLam > 1.0 ) {
            minLam = 1.0;
            dumEl = 0;
            increaseLoad = true;
        }


        // scaling of previous increment by lambda => subtracting previously added value and addition of the scaled values
        incrementOfDisplacementVector.times(minLam-1.0);
        for ( i = 1; i <= nnodes; i++ ) {
            updateDofUnknownsDictionary(d->giveDofManager(i), tStep);
        }

        if( this->solverType == ST_IML ) {
            // subtract the stiffness of damaged element (only part of the stiff. matrix is updated)
            this->updateStiffnessMatrix(stiffnessMatrix, tStep, SecantStiffnessMatrix,
                                        EModelDefaultEquationNumbering(), d, dumEl, -1.0 );
        }

        for ( i = 1; i <= nelem; i++ ) {
            // update element history variables
            Element *elem = d->giveElement(i);
            slaMat = dynamic_cast<SlaMaterialExtensionInterface*>(elem->giveMaterial());

            for ( int j = 0; j < elem->giveNumberOfIntegrationRules(); j++ ) {
                iRule = elem->giveIntegrationRule(j);
                for ( int k = 0; k < iRule->giveNumberOfIntegrationPoints(); k++ ) {
                    gp = iRule->getIntegrationPoint(k);

                    giveDeformationLoadForElement(strain, rotationMatG2L, elem, gp, tStep);
                    if( rotationMatG2L.giveNumberOfRows() ) {
                      strain.rotatedWith(rotationMatG2L, 'n'); // transform to local coord system
                    }

                    slaMat->updateHistoryParameters(gp, strain, tStep, ((i==dumEl) && (k==dumGp)) );
                }
            }
        }


        if( this->solverType == ST_IML ) {
            // add the updated stiffness of damaged element
            this->updateStiffnessMatrix(stiffnessMatrix, tStep, SecantStiffnessMatrix,
                                        EModelDefaultEquationNumbering(), d, dumEl, 1.0 );
        } else {
            // direct solver changes stiffMatrix during the solution
            if ( stiffnessMatrix ) {
                delete stiffnessMatrix;
            }
            stiffnessMatrix = classFactory.createSparseMtrx(sparseMtrxType);
            if ( stiffnessMatrix == NULL ) {
                OOFEM_ERROR("sparse matrix creation failed");
            }
            stiffnessMatrix->buildInternalStructure( this, 1, EModelDefaultEquationNumbering() );
            stiffnessMatrix->zero();
            this->assemble( stiffnessMatrix, tStep, SecantStiffnessMatrix, EModelDefaultEquationNumbering(), d );
            stiffnessMatrix->writeToFile("stiff.out");
        }

        if( increaseLoad ) {
            #ifdef VERBOSE
                OOFEM_LOG_INFO("    Number of steps in avalanche: %d\n", avalanche);
            #endif
        }

        if( printIterations ) {
            printIteration(tStep,avalanche);
        }

        // to ensure proper evaluation of stresses in damaged element, the stress must be decreased (damaged stress)
        if( dumEl != 0 ) {
            Element *elem = d->giveElement(dumEl);
            slaMat = dynamic_cast<SlaMaterialExtensionInterface*>(elem->giveMaterial());

            for ( int j = 0; j < elem->giveNumberOfIntegrationRules(); j++ ) {
                iRule = elem->giveIntegrationRule(j);
                gp = iRule->getIntegrationPoint(dumGp);

                slaMat->updateStressDrop(gp,tStep);
            }
        }
    }

    // increment is already added
    incrementOfDisplacementVector.zero();

}

void
SequentiallyLinearStatic :: updateYourself(TimeStep *tStep)
{
    StructuralEngngModel :: updateInternalState(tStep);
    EngngModel :: updateYourself(tStep);
}

void
SequentiallyLinearStatic :: updateStiffnessMatrix(SparseMtrx *answer, TimeStep *tStep,
                                                  CharType type, const UnknownNumberingScheme &s, Domain *domain,
                                                  int elemNum, double scale)
{

    if( elemNum < 1 || elemNum > domain->giveNumberOfElements() ) {
        return;
    }

    IntArray loc;
    FloatMatrix mat, R;
    if ( answer == NULL ) {
        OOFEM_ERROR("NULL pointer encountered.");
    }


    Element *element = domain->giveElement(elemNum);
    if ( !element->isActivated(tStep) ) {
        return;
    }

    domain->giveEngngModel()->giveElementCharacteristicMatrix(mat, elemNum, type, tStep, domain);
    mat.times(scale);
    if ( mat.isNotEmpty() ) {
        element->giveLocationArray(loc, s);
        if ( element->giveRotationMatrix(R) ) {
            mat.rotatedWith(R);
        }

        if ( answer->assemble(loc, mat) == 0 ) {
            OOFEM_ERROR("sparse matrix assemble error");
        }
    }


    int nbc = domain->giveNumberOfBoundaryConditions();
    for ( int i = 1; i <= nbc; ++i ) {
        ActiveBoundaryCondition *bc = dynamic_cast< ActiveBoundaryCondition * >( domain->giveBc(i) );
        if ( bc != NULL ) {
            OOFEM_ERROR("Must be checked how this influences the partial update of stiffness matrix");
        }
    }

    answer->assembleBegin();
    answer->assembleEnd();
}

double
SequentiallyLinearStatic :: giveDiscreteTime(int iStep)
{
    if ( this->fixedSteps ) {
        return this->deltaT * (iStep-1);
    } else {
        if ( ( iStep > 0 ) && ( iStep <= discreteTimes.giveSize() ) ) {
            return ( discreteTimes.at(iStep) );
        }
    }

    OOFEM_ERROR("invalid iStep");
    return 0.0;
}


TimeStep*
SequentiallyLinearStatic :: giveNextStep()
{
    int istep = this->giveNumberOfFirstStep();
    int mStepNum = 1;
    double dt = this->giveDiscreteTime(istep);
    StateCounterType counter = 1;

    if ( currentStep != NULL ) {
        istep =  currentStep->giveNumber() + 1;
        dt = this->giveDiscreteTime(istep) - this->giveDiscreteTime(istep - 1);
        counter = currentStep->giveSolutionStateCounter() + 1;
    }

    if ( previousStep != NULL ) {
        delete previousStep;
    }

    previousStep = currentStep;
    if ( previousStep == NULL ) {
        previousStep = new TimeStep(giveNumberOfTimeStepWhenIcApply(), this, 0, -dt, dt, 0);
    }
    currentStep = new TimeStep(istep, this, mStepNum, this->giveDiscreteTime ( istep ), dt, counter);
    return currentStep;
}


void
SequentiallyLinearStatic :: printIteration(TimeStep *tStep, int avalanche)
{
    int di = 1;
    FILE *fw = this->outputStreamIterations;
    Domain *d = this->giveDomain(di);
    DofManager *dofman;
    StructuralElement *se;
    StructuralMaterialStatus *status;
    IntegrationRule *iRule;
    GaussPoint *gp;
    int i,j,k,l;

    fprintf(fw, "----------------------------------------------------------------------\n");
    fprintf(fw, "Step: %6d  Time: %e Substep: %6d Avalanche: %6d\n", tStep->giveNumber(), tStep->giveTargetTime(),
                                                                    tStep->giveSubStepNumber(), avalanche);
    fprintf(fw, "----------------------------------------------------------------------\n");

    // node displacements
    fprintf(fw, "Node displacements\nID\tdisplacements\n");
    for( i = 1; i <= d->giveNumberOfDofManagers(); i++ ) {
        dofman = d->giveDofManager(i);
        fprintf(fw, "%6d", dofman->giveLabel());
        for ( Dof *dof: *dofman ) {
            fprintf(fw, "\t% e", dof->giveUnknown( VM_Total, tStep ));
        }
        fprintf(fw, "\n");
    }

    // element strains
    FloatArray strain;
    fprintf(fw, "Element strains\nID\tIR\tIP\tstrains\n");
    for ( i = 1; i <= d->giveNumberOfElements(); i++ ) {
        se = dynamic_cast<StructuralElement*>(d->giveElement(i));
        if( !se ) continue;
        for ( j = 0; j < se->giveNumberOfIntegrationRules(); j++ ) {
            iRule = se->giveIntegrationRule(j);
            for ( k = 0; k < iRule->giveNumberOfIntegrationPoints(); k++ ) {
                gp = iRule->getIntegrationPoint(k);
                status = dynamic_cast< StructuralMaterialStatus * >( se->giveMaterial()->giveStatus(gp) );
                strain = status->giveStrainVector();

                fprintf(fw, "%6d  %3d  %3d", se->giveLabel(), iRule->giveNumber(), gp->giveNumber());
                for ( l = 1; l <= strain.giveSize(); l++ ) {
                    fprintf(fw, "\t% e", strain.at(l));
                }
                fprintf(fw, "\n");
            }
        }
    }

    // element stresses
    FloatArray stress;
    fprintf(fw, "Element stresses\nID\tIR\tIP\tstresses\n");
    for ( i = 1; i <= d->giveNumberOfElements(); i++ ) {
        se = dynamic_cast<StructuralElement*>(d->giveElement(i));
        if( !se ) continue;
        for ( j = 0; j < se->giveNumberOfIntegrationRules(); j++ ) {
            iRule = se->giveIntegrationRule(j);
            for ( k = 0; k < iRule->giveNumberOfIntegrationPoints(); k++ ) {
                gp = iRule->getIntegrationPoint(k);
                status = dynamic_cast< StructuralMaterialStatus * >( se->giveMaterial()->giveStatus(gp) );
                stress = status->giveStressVector();

                fprintf(fw, "%6d  %3d  %3d", se->giveLabel(), iRule->giveNumber(), gp->giveNumber());
                for ( l = 1; l <= stress.giveSize(); l++ ) {
                    fprintf(fw, "\t% e", stress.at(l));
                }
                fprintf(fw, "\n");
            }
        }
    }

    // reactions
    FloatArray reactions;
    IntArray dofManMap, dofMap, eqnMap;

    this->buildReactionTable(dofManMap, dofMap, eqnMap, tStep, di);
    this->computeReaction(reactions, tStep, di);
    fprintf(fw, "Reactions\nID\tDoF\tValue\n");
    for ( i = 1; i <= dofManMap.giveSize(); i++ ) {
        fprintf( fw, "%6d  %3d  % e\n",
                d->giveDofManager( dofManMap.at(i) )->giveLabel(),
                dofMap.at(i), reactions.at( eqnMap.at(i) ) );
    }

}

void
SequentiallyLinearStatic::giveDeformationLoadForElement(FloatArray &answer, FloatMatrix &rotationMatG2L, Element *elem, GaussPoint *gp, TimeStep *tStep)
{
  StructuralElement *se = dynamic_cast<StructuralElement*>(elem);
  StructuralInterfaceElement *sie = dynamic_cast<StructuralInterfaceElement*>(elem);

  if( se != NULL && sie != NULL ) {
    OOFEM_ERROR( "Multiple element characteristic - not allowed" );
  }

  if( se ) {
    se->computeStrainVector(answer, gp, tStep);
    rotationMatG2L.clear();
  } else if( sie ) {
    sie->computeSpatialJump(answer,gp,tStep);
    // friend class SequentiallyLinearStatic; must be defined in class "StructuralInterfaceElement"
    sie->computeTransformationMatrixAt(gp, rotationMatG2L);
  } else {
    OOFEM_ERROR( "Unsupported element type" )
  }

}

} // end namespace oofem
