/*
 *
 *                 #####    #####   ######  ######  ###   ###
 *               ##   ##  ##   ##  ##      ##      ## ### ##
 *              ##   ##  ##   ##  ####    ####    ##  #  ##
 *             ##   ##  ##   ##  ##      ##      ##     ##
 *            ##   ##  ##   ##  ##      ##      ##     ##
 *            #####    #####   ##      ######  ##     ##
 *
 *
 *             OOFEM : Object Oriented Finite Element Code
 *
 *               Copyright (C) 1993 - 2013   Borek Patzak
 *
 *
 *
 *       Czech Technical University, Faculty of Civil Engineering,
 *   Department of Structural Mechanics, 166 29 Prague, Czech Republic
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef sequentiallylinearstatic_h
#define sequentiallylinearstatic_h

#include "incrementallinearstatic.h"

///@name Input fields for SequentiallyLinearStatic
//@{
#define _IFT_SequentiallyLinearStatic_Name "sequentlinearstatic"
#define _IFT_SequentiallyLinearStatic_PrintIterations "printiterations"
#define _IFT_SequentiallyLinearStatic_FileName "filename"
//@}

namespace oofem {
class Element;
class GaussPoint;

/**
 * This class implements Sequentially Linear Static Engineering problem.
 * Problem is solved as series of explicit linear solutions.
 * It will solve any nonlinear problem by writing the expressions explicitly
 * @f$ K({}^{n-1} r)\cdot \delta r = R({}^{n-1}r) - F @f$.
 *
 * Supports the changes of static scheme (applying, removing and changing  boundary conditions)
 * during the analysis.
 *
 */

class SequentiallyLinearStatic : public IncrementalLinearStatic
{

private: 
    bool printIterations;
    FILE *outputStreamIterations;
    std::string fileNameIterations;

    void updateStiffnessMatrix(SparseMtrx *answer, TimeStep *tStep,
                               CharType type, const UnknownNumberingScheme &s, Domain *domain,
                               int elemNum, double scale);
    void printIteration(TimeStep *tStep, int avalanche);

    double giveDiscreteTime(int iStep);
    void giveDeformationLoadForElement(FloatArray &answer, FloatMatrix &rotationMatG2L, Element *elem, GaussPoint *gp, TimeStep *tStep);

public:
    SequentiallyLinearStatic(int i, EngngModel * _master = NULL);
    virtual ~SequentiallyLinearStatic();

    virtual IRResultType initializeFrom(InputRecord *ir);

    virtual void solveYourselfAt(TimeStep *tStep);
    virtual void updateYourself(TimeStep *tStep);

    virtual const char *giveInputRecordName() const { return _IFT_SequentiallyLinearStatic_Name; }
    virtual const char *giveClassName() const { return "SequentiallyLinearStatic"; }

    virtual TimeStep *giveNextStep();

};
} // end namespace oofem
#endif // sequentiallylinearstatic_h
