/*
*
*                 #####    #####   ######  ######  ###   ###
*               ##   ##  ##   ##  ##      ##      ## ### ##
*              ##   ##  ##   ##  ####    ####    ##  #  ##
*             ##   ##  ##   ##  ##      ##      ##     ##
*            ##   ##  ##   ##  ##      ##      ##     ##
*            #####    #####   ##      ######  ##     ##
*
*
*             OOFEM : Object Oriented Finite Element Code
*
*               Copyright (C) 1993 - 2011   Borek Patzak
*
*
*
*       Czech Technical University, Faculty of Civil Engineering,
*   Department of Structural Mechanics, 166 29 Prague, Czech Republic
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "shccmat.h"
#include "floatarray.h"
#include "floatmatrix.h"
#include "structuralms.h"
#include "gausspoint.h"
#include "intarray.h"
#include "datastream.h"
#include "contextioerr.h"
#include "timestep.h"
#include "structuralmaterial.h"
#include "isolinearelasticmaterial.h"
#include "structuralcrosssection.h"
#include "mathfem.h"
#include "classfactory.h"

namespace oofem {
  REGISTER_Material(SHCCMaterial);

  // constructor
  SHCCMaterial :: SHCCMaterial(int n, Domain *d) : StructuralMaterial(n, d),
      dDamage(), princStrain(), dprincStrain()
  {
    softType  = ST_Unknown;
    crushType = CT_Unknown;

    linearElasticMaterial = new IsotropicLinearElasticMaterial(n, d);

    E           = 0.0;
    nu          = 0.0;
    epsT0       = 0.0;
    epsTp       = 0.0;
    sigTp       = 0.0;
    wT          = 0.0;
    aT          = 0.0;
    bT          = 0.0;
    bT_closing  = 0.0;
    epsC0       = 0.0;
    epsCp       = 0.0;
    sigCp       = 0.0;
    epsCu_input = 0.0;
    dC          = 0.0;         
    aC          = 0.0;         
    bC          = 0.0;         
    bC_closing  = 0.0;

    wC          = 0.0;
    betaC       = 0.0;
    Le          = 0.0; // Le = element size
    changeState = false;

    ecsMethod   = ECSM_Unknown;
    
    dimension   = d->giveNumberOfSpatialDimensions();

    // state determining parametres
    dDamage.resize(dimension);

    epsTmax_calc = 0.0; 
    sigTmax_calc = 0.0; 
    epsTul_calc  = 0.0; 
    sigTul_calc  = 0.0; 
      
    epsCmin_calc = 0.0; 
    sigCmin_calc = 0.0; 
    epsCul_calc  = 0.0; 
    sigCul_calc  = 0.0; 

  }

  // destructor
  SHCCMaterial :: ~SHCCMaterial()
  {
    delete linearElasticMaterial;
  }

  int
  SHCCMaterial :: hasMaterialModeCapability(MaterialMode mode)
    //
    // returns whether receiver supports the given mode
    //
  {
    if ( ( mode == _PlaneStress ) ) {
      return 1;
    }

    return 0;
  }

  void
  SHCCMaterial :: give3dMaterialStiffnessMatrix(FloatMatrix &answer, MatResponseMode mode, GaussPoint *gp, TimeStep *atTime)
  {
    SHCCMaterialStatus *status = ( SHCCMaterialStatus * ) this->giveStatus(gp);
    // start from the elastic stiffness
    this->giveLinearElasticMaterial()->give3dMaterialStiffnessMatrix(answer, mode, gp, atTime);
  }

  void
  SHCCMaterial :: givePlaneStressStiffMtrx(FloatMatrix &answer, MatResponseMode mode, GaussPoint *gp, TimeStep *atTime)
  {
    SHCCMaterialStatus *status = ( SHCCMaterialStatus * ) this->giveStatus(gp);

    // start from the elastic stiffness
    this->giveLinearElasticMaterial()->givePlaneStressStiffMtrx(answer, mode, gp, atTime);

//    if ( mode == SecantStiffness ) {
//        double alpha;
//        FloatArray  damage = status->giveDamage();
//        FloatMatrix dum(3,3), omega(3,3), Tstress(3,3), Tstrain(3,3), TstrainT(3,3);

//        omega.at(1,1) = 1.0;
//        omega.at(1,2) = damage.at(1);
//        omega.at(2,1) = damage.at(2);
//        omega.at(2,2) = 1.0;
//        omega.at(3,3) = 1.0;

//        alpha   = status->giveAlpha();
//        Tstrain = giveTransformationMatrixStrain(alpha);
//        Tstress = giveTransformationMatrixStress(alpha);
//        TstrainT.beTranspositionOf(Tstrain);

//        dum.beProductOf(TstrainT,omega);
//        omega.beProductOf(dum,Tstress);

//        dum = answer;
//        answer.beProductOf(omega,dum);
//        return;

//        // OOFEM_ERROR( "SHCCMaterial: givePlaneStressStiffMtrx: Secant stiffness matrix is not supported!!!" );

//    } else if ( mode != TangentStiffness ) {
//        return;
//    }

    if ( mode != TangentStiffness && mode != SecantStiffness ) {
      return;
    }

    FloatArray damage = status->giveTempDamage();
    if ( damage(0) == 0 && damage(1) == 0 ) { // elastic loading - elastic stiffness plays the role of tangent stiffness
      return;
    }

    giveCharacteristicMatrix(answer, mode, gp, atTime);

  }

  void  
  SHCCMaterial :: giveCharacteristicMatrix(FloatMatrix &answer, MatResponseMode mode,
                                           GaussPoint *gp, TimeStep *atTime)
  {
    FloatMatrix de, invDe, compliance, T, Tt, dumMat, nuDamaged(2,2);
    FloatArray dcr, activeCrack, principalStressVector, principalStrainVector;
    SHCCMaterialStatus *status = ( SHCCMaterialStatus * ) this->giveStatus(gp);
    
    this->giveLinearElasticMaterial()->givePlaneStressStiffMtrx(de, mode, gp, atTime);
    invDe.beInverseOf(de);
    dcr         = status->giveTempTangent();
    activeCrack = status->giveTempDamage();

    nuDamaged.at(1,1) = nuDamaged.at(2,2) = 1.0;
    nuDamaged.at(1,2) = (1.0-activeCrack.at(1))*nu;
    nuDamaged.at(2,1) = (1.0-activeCrack.at(2))*nu;
    // first we set compliances for normal streses in
    // local coordinate system defined by crackplane
    compliance = invDe;
    for ( int i = 1; i <= 2; i++ ) 
    {
      double Ecracked = giveCrackingModulus(mode, gp, i);
      if( !(status->giveTempActiveState().at(i) == noneSS ||
           (status->giveTempActiveState().at(i) == tensionSS && status->giveTempEpsTmax().at(i) <= epsT0) ||
           (status->giveTempActiveState().at(i) == compressionSS && status->giveTempEpsCmin().at(i) >= epsC0)) ) {

        if ( Ecracked <= SHCC_SMALL ) {
          compliance.at(i,i) *= SHCC_BIGNUMBER;
        } else {
          compliance.at(i,i) = 1. / Ecracked;
        }
      }
    }

    principalStressVector = status->giveTempPrincipalStress();
    principalStrainVector = status->giveTempPrincipalStrain();

    // now remain to set shears
    if ( fabs(principalStrainVector.at(1) - principalStrainVector.at(2)) < SHCC_SMALL ) {
        compliance.at(3,3) = 1. / (E/(2*(1+nu)));
    } else if ( fabs( principalStressVector.at(1) - principalStressVector.at(2)) < SHCC_SMALL ) {
        compliance.at(3,3) = SHCC_BIGNUMBER;
    } else {
        compliance.at(3,3) = 2 * (principalStrainVector.at(1) - principalStrainVector.at(2)) / 
                                 (principalStressVector.at(1) - principalStressVector.at(2));
    }

    // now we invert compliance to get stiffness in reduced space
    de.beInverseOf(compliance);

    T = giveTransformationMatrixStrain( status->giveTempAlpha() );
    Tt.beTranspositionOf(T);
    dumMat.beProductOf(Tt,de);
    answer.beProductOf(dumMat,T);

  }

  double
  SHCCMaterial :: giveCrackingModulus(MatResponseMode rMode, GaussPoint *gp, int i)
  //
  // returns current cracking modulus for i-th crackplane
  // softening modulus represents a relation between the normal crack strain
  // rate and the normal stress rate.
  //
  {
      double Cf;
      SHCCMaterialStatus *status = static_cast< SHCCMaterialStatus * >( this->giveStatus(gp) );

      if( status->giveTempActiveState().at(i) == noneSS ||
          (status->giveTempActiveState().at(i) == tensionSS && status->giveTempEpsTmax().at(i) <= epsT0) ||
          (status->giveTempActiveState().at(i) == compressionSS && status->giveTempEpsCmin().at(i) >= epsC0) ) {
        Cf = 0.0;
      }

      if ( rMode == TangentStiffness ) {
        Cf = status->giveTempTangent().at(i);

      } else { // secant stiffness
        if( status->giveTempActiveState().at(i) == tensionSS ) {
          Cf = status->giveTempSigTmax().at(i) /
                     ( status->giveTempEpsTmax().at(i) -
                       giveEpsTul(status->giveTempEpsTmax().at(i),status->giveTempSigTmax().at(i),i,status) );
        } else { // compression
          Cf = status->giveTempSigCmin().at(i) /
                     ( status->giveTempEpsCmin().at(i) -
                       giveEpsCul(status->giveTempEpsCmin().at(i),status->giveTempSigCmin().at(i),i,status) );
        }
      }

      return Cf;
  }

  void
  SHCCMaterial :: giveRealStressVector(FloatArray &answer, GaussPoint *gp,
                                       const FloatArray &totalStrain, TimeStep *atTime)
  //
  // returns real stress vector in 2d stress space of receiver according to
  // previous level of stress and current
  // strain increment, the only way, how to correctly update gp records
  //
  {
    double alpha;
    FloatArray dumVec, tangent, principalStress, reducedTotalStrainVector;
    FloatMatrix Tstrain, TstrainT;
    
    SHCCMaterialStatus *status = ( SHCCMaterialStatus * ) this->giveStatus(gp);
    
    dumVec.resize(this->dimension);
    principalStress.resize(this->dimension);
    tangent.resize(this->dimension);

    this->initTempStatus(gp);

    this->giveStressDependentPartOfStrainVector(reducedTotalStrainVector, gp, totalStrain, atTime, VM_Total);

    // setting temp value of element size and others
    dDamage.resize(this->dimension);
    princStrain.resize(totalStrain.giveSize());
    dprincStrain.resize(totalStrain.giveSize());

    alpha   = give2DRotationalAngleStrain(reducedTotalStrainVector);
    Tstrain = giveTransformationMatrixStrain(alpha);
    TstrainT.beTranspositionOf(Tstrain);
    princStrain.beProductOf(Tstrain,reducedTotalStrainVector);
    dprincStrain.beDifferenceOf(princStrain, status->givePrincipalStrain());

    status->letTempAlpha(alpha);


    if(atTime->giveTargetTime()==47)
      int ooi = 1;


    // principal stress
    principalStress.zero();
    tangent.zero();
    for( int i=1; i <= dimension; i++ )
    {
      if( status->giveTempElemSize().at(i) > 0.0 ) {
        initParametersSoftening(i,gp,status);
        initParametersCrushing(i,gp,status);
      }

      changeState = false;     

      if( status->giveTempActiveState().at(i) == tensionSS ||
         (status->giveTempActiveState().at(i) == noneSS && princStrain.at(i) > SHCC_SMALL) )
      {
        // determines stress and updates global parameters
        dumVec = giveTensileStress( i, gp, status );

        if( changeState ) 
          dumVec = giveCompressiveStress( i, gp, status );
      }
      else if( status->giveTempActiveState().at(i) == compressionSS ||
              (status->giveTempActiveState().at(i) == noneSS && princStrain.at(i) < -SHCC_SMALL) )
      {
        // determines stress and updates global parameters
        dumVec = giveCompressiveStress( i, gp, status );

        if( changeState ) 
          dumVec = giveTensileStress( i, gp, status );
      } 
      else // if zero load - usually the initiation
      {
        dumVec(0)  = SHCC_ZERO;
        dumVec(1)  = E;

        status->letTempDamageAt(0.0, i);
        dDamage.at(i) = 0.0;
      }

      principalStress.at(i) = dumVec(0);
      tangent.at(i)         = dumVec(1);
    }

    FloatArray _nu(this->dimension),_stress(totalStrain.giveSize());

    //damage(0) = damage(1) = max(damage(0),damage(1));

    _nu(0) = nu*(1.0 - status->giveTempDamage().at(1));
    _nu(1) = nu*(1.0 - status->giveTempDamage().at(2));

    _stress.zero();
    _stress(0) = (principalStress(0) + _nu(0)*principalStress(1))/(1.-_nu(0)*_nu(1));  
    _stress(1) = (principalStress(1) + _nu(1)*principalStress(0))/(1.-_nu(0)*_nu(1));

    answer.beProductOf(TstrainT,_stress);

    // saving temporary parametres
    status->letTempStressVectorBe(answer);
    status->letTempStrainVectorBe(totalStrain);

    status->letTempPrincipalStrain(princStrain);

    status->letTempTangentStiffMat( giveTangetialStiffMatrix(principalStress,tangent,alpha,status) );

    status->letTempTangent(tangent);
    
  }

  void
  SHCCMaterial :: initElementSize(int direction, GaussPoint *gp, SHCCMaterialStatus *status)
  {
      if( status->giveTempElemSize().at(direction) > 0.0 ) {
        Le = status->giveTempElemSize().at(direction);

      } else {
      double alpha = status->giveTempAlpha();
        FloatArray crackPlaneNormal(3);
        crackPlaneNormal.zero();
        if( direction == 1 )
        {
          crackPlaneNormal.at(1) = cos(alpha);
          crackPlaneNormal.at(2) = sin(alpha);
        }
        else
        {
          crackPlaneNormal.at(1) = sin(alpha);
          crackPlaneNormal.at(2) = cos(alpha);
        }

        Le = gp->giveElement()->giveCharacteristicSize(gp, crackPlaneNormal, ecsMethod);
      }
  }

  void
  SHCCMaterial :: initParametersSoftening(int direction, GaussPoint *gp, SHCCMaterialStatus *status)
  {
      // Setup of driving parameters
      // tension softening

      initElementSize(direction,gp,status);

      if( softType == ST_Hermit ) // Hermit function softening
      {
        epsTu.resize(1);
        epsTu.at(1) = epsTp + wT/Le - 2.0*(epsTp-status->giveTempEpsTulForEpsTp().at(direction))/(aT+1.);
      }
      else if( softType == ST_Linear || ST_Polynomial_Cohesive_Crack )
      {
        epsTu.resize(1);
        epsTu.at(1) = status->giveTempEpsTulForEpsTp().at(direction) + wT/Le;
      }
      else if( softType == ST_Exponential_Cohesive_Crack || softType == ST_Polynomial_Cohesive_Crack )
      {
        epsTu.resize(0);
      }
  }

  void
  SHCCMaterial :: initParametersCrushing(int direction, GaussPoint *gp, SHCCMaterialStatus *status)
  {
      initElementSize(direction, gp, status);

      // Setup of driving parameters
      // compression softening
      if( crushType == CT_Hermit )
      {
        epsCu.resize(1);

        wC          = (epsCu_input - status->giveTempEpsCulForEpsCp().at(direction))*dC;
        epsCu.at(1) = epsCp + wC/Le - 2.0*(epsCp - status->giveTempEpsCulForEpsCp().at(direction))/(aC+1.);
      }
      else if( crushType == CT_Thorenfeldt_Cohesive_Crack || crushType == CT_Exponential_Cohesive_Crack )
      {
        epsCu.resize(0);
      }
  }

  // Tension behavior
  FloatArray
  SHCCMaterial :: giveTensileStress(int direction, GaussPoint *gp, SHCCMaterialStatus *status )
  {
    // answer(0) = stress, answer(1) = tengent;
    FloatArray answer(2);


    if( (status->giveTempActiveState().at(direction) == compressionSS || status->giveTempTCrackClosing().at(direction) )
         && !(status->giveTempEpsTmax().at(direction) <= epsT0 && status->giveTempEpsCmin().at(direction) >= epsC0) )
    {
      // loading after compression
      answer = giveTensionAfterCompression( direction, gp, status );
    } 
    else
    {
      answer = giveTensionOrdinaryLoading( direction, gp, status );
    }

    return answer;
  }

  
  FloatArray 
  SHCCMaterial :: giveTensionAfterCompression( int direction, GaussPoint *gp, SHCCMaterialStatus *status )
  {    
    // setup of driving variables
    setTmaxCmin_calc( direction, status );

    double _dummy;
    if( status->giveTempActiveState().at(direction) == compressionSS )
    {
      // setup of EpsTul parametrs on the value of EpsCul
      _dummy = giveClosingStrain( true, direction, status );
      status->letTempEpsTulAt(_dummy, direction);
      status->letTempEpsTpulAt(_dummy, direction);
      status->letTempSigTpulAt(0.0, direction);
    }


    // answer(0) = stress, answer(1) = tengent;
    FloatArray answer(2);

    double _eps, _exponent;
    double epsTmax = status->giveTempEpsTmax().at(direction);
    double sigTmax = status->giveTempSigTmax().at(direction);
    double epsTpul = status->giveTempEpsTpul().at(direction);
    double epsTul  = status->giveTempEpsTul().at(direction);
    double sigTpul = status->giveTempSigTpul().at(direction);
    double epsCmin = status->giveTempEpsCmin().at(direction);
    double sigCmin = status->giveTempSigCmin().at(direction);
    double epsTprl = status->giveTempEpsTprl().at(direction);
    double sigTprl = status->giveTempSigTprl().at(direction);

    // setup of logical parameters 
    status->letTempTCrackClosingAt(true, direction);
    status->letTempActiveStateAt(tensionSS,direction);

    // setup of damage parameter
    status->letTempDamageAt(giveDamage( true, direction, status ), direction);
    dDamage.at(direction) = 0.0;


    // Stress evaluation
    if( dprincStrain.at(direction) > 0.0 )   // loading
    {
      //sirka oblasti na ktere se prejde na normalni zatezovani
      // pri prvnim behu odecneni epsTpul = epsCul => sirka oblasti mezi epsCul a epsTmax,
      // nasledne zmenseni oblasti o neuzaviratelne trhliny
      _dummy  = max( epsT0, epsTmax ) - epsTpul;
      _dummy += (1.0-bC_closing) * giveEpsCul(epsCmin,sigCmin,direction,status);  // zmenseni o oblast neuzaviratelnych trhlin

      // relative strain <0;_dummy>
      _eps  = princStrain.at(direction) - epsTpul;

      if( _eps < _dummy )  // loading with crack closing
      {
        if( sigTmax_calc < SHCC_SMALL ) 
        {
          answer(0) = sigTpul;
          answer(1) = sigTmax_calc / epsTmax_calc;

          // update driving parameters
          status->letTempEpsTprlAt(_eps + epsTpul,direction);
          status->letTempSigTprlAt(answer(0), direction);

          status->letTempEpsTulAt(_eps + epsTpul,direction);
        }
        else
        {
          _exponent = _dummy * sigTmax_calc / ((sigTmax_calc-sigTpul)*epsTmax_calc);

          answer(0) = max( SHCC_ZERO, sigTpul + (sigTmax_calc-sigTpul) *
                                 pow(_eps/_dummy, _exponent) );
          answer(1) = max( SHCC_ZERO, (sigTmax_calc-sigTpul) * _exponent / _dummy *
                                 pow(_eps/_dummy, _exponent-1.0) );

          // update driving parameters
          epsTprl = _eps + epsTpul;
          sigTprl = answer(0);
          status->letTempEpsTprlAt(epsTprl, direction); // ulozeni skutecne hodnoty strainu (neposunute)
          status->letTempSigTprlAt(sigTprl, direction);

          // ulozeni skutecne hodnoty strainu epsTul (neposunute)
          _dummy  = min( giveEpsTul(epsTmax, sigTmax, direction, status) +
                         (1.0-bC_closing) * giveEpsCul(epsCmin, sigCmin, direction, status),
                         epsTprl-(epsTmax_calc/sigTmax_calc) * answer(0) );

          status->letTempEpsTulAt(_dummy, direction);

        }
      }
      else  // normal loading
      {
        status->letTempTCrackClosingAt(false, direction);
        answer = giveTensionOrdinaryLoading( direction, gp, status );
      }
    }
    else  // unloading
    {
      _eps = princStrain.at(direction) - epsTul;

      if( _eps >= 0.0 ) 
      {
        if( fabs(epsTprl-epsTul) < SHCC_SMALL )
        {
          answer(0) = SHCC_ZERO;
          answer(1) = SHCC_ZERO;
        } 
        else
        {
          _dummy   = _eps / (epsTprl-epsTul);

          answer(0) = max( SHCC_ZERO, sigTprl*pow(_dummy, aT) );
          answer(1) = max( SHCC_ZERO, sigTprl * aT * pow(_dummy, aT-1.0)/(epsTprl-epsTul) );
        }

        // update driving parameters
        status->letTempEpsTpulAt(_eps + epsTul, direction);
        status->letTempSigTpulAt(answer(0), direction);
      }
      else
      {
        answer = giveCompressiveStress( direction, gp, status );
      }
    }

    return answer;
  }

  
  FloatArray 
  SHCCMaterial :: giveTensionOrdinaryLoading( int direction, GaussPoint *gp, SHCCMaterialStatus *status )
  {
    // answer(0) = stress, answer(1) = tangent;
    FloatArray answer(2);
    
    double _dummy, _eps, _epsT0p, _sigT0p;
    double epsTmax = status->giveTempEpsTmax().at(direction);
    double sigTmax = status->giveTempSigTmax().at(direction);

    // setup
    status->letTempTCrackClosingAt(false, direction);
    status->letTempActiveStateAt(tensionSS, direction);

    _eps = giveShiftedStrain( true, direction, status );

    _epsT0p = epsT0 + 0.00*(epsTp - epsT0);
    _sigT0p = E*epsT0 + 0.00*(sigTp-E*epsT0);

    if( (_eps-epsTmax) > -SHCC_SMALL )    // initial loading - failure envelope
    {
      if( _eps <= epsT0 ) 
      {
        answer(0) = E*_eps;
        answer(1) = E;    
      }       
      else if( _eps <= _epsT0p ) 
      {
        _dummy   = (_epsT0p-_eps)/(_epsT0p-epsT0);

        answer(0) = _sigT0p - (_sigT0p-epsT0*E)* pow(_dummy, E*(_epsT0p-epsT0)/(_sigT0p-epsT0*E));
        answer(1) = max( SHCC_ZERO, E * pow(_dummy, (E*(_epsT0p-epsT0)/(_sigT0p-epsT0*E))-1.0) );   
      }
      else if( _eps <= epsTp ) 
      {
        _dummy   = (_eps-_epsT0p)/(epsTp-_epsT0p);

        answer(0) = _sigT0p + (sigTp-_sigT0p)*(-2.0 * pow(_dummy, 3.0) + 3.0 * pow(_dummy, 2.0));
        //answer(1) = max( SHCC_ZERO, (sigTp-_sigT0p)*(-6.0* pow(_dummy, 2.0) + 6.0*_dummy)/(epsTp-_epsT0p) );

        //answer(0) = epsT0p*E + (sigTp-epsT0p*E)*(1-(epsTp-_eps)/(epsTp-epsT0p));
        answer(1) = max( SHCC_ZERO,  (sigTp-_epsT0p*E) / (epsTp-_epsT0p) );

      } 
      else  
      {
        // fixing the element size - for the first time when it occures
        if( status->giveTempElemSize().at(direction) == 0.0 )
        {
          status->letTempEpsTulForEpsTpAt(giveEpsTul(epsTp, sigTp, direction, status), direction);

          initParametersSoftening(direction,gp,status);

          status->letTempElemSizeAt(Le, direction);

          if( epsTu(0) <= epsTp )
            OOFEM_ERROR("SHCCMaterial: epsTu <= epsTp for element no. %d (%d), Le = %e!!!", gp->giveElement()->giveNumber(), gp->giveNumber(), Le );
      
        }


        if( softType == ST_Hermit ) // Hermit function softening
        {
          if( _eps < epsTu(0) )
          {
            _dummy    = (_eps-epsTp)/(epsTu(0)-epsTp);
            answer(0) = max( SHCC_ZERO, sigTp*(2.0 * pow(_dummy, 3.0) - 3.0 * pow(_dummy, 2.0) + 1.0) );
            answer(1) = min( -SHCC_ZERO, sigTp*(6.0 * pow(_dummy, 2.0) - 6.0*_dummy)/(epsTu(0)-epsTp) );
          }
          else
          {
            answer(0) = SHCC_ZERO;
            answer(1) = -SHCC_ZERO;
          }
        }
        else if( softType == ST_Linear ) // linear softening
        {
          if( _eps < epsTu(0) )
          {
            _dummy    = (_eps-epsTp)/(epsTu(0)-epsTp);
            answer(0) = max( SHCC_ZERO, sigTp*(1.0-_dummy) );
            answer(1) = -sigTp/(epsTu(0)-epsTp);
          }
          else
          {
            answer(0) = SHCC_ZERO;
            answer(1) = -SHCC_ZERO;
          }
        }
        else if( softType == ST_Exponential_Cohesive_Crack )
        {
          // exponential cohesive crack - iteration needed
          double R, Lhs, help, epsF; 
          double _epsTul = status->giveTempEpsTulForEpsTp().at(direction);
          int nite = 0;
          // iteration to achieve objectivity
          // we are looking for a state in which the elastic stress (unloading) is equal to
          // the stress from crack-opening relation
          // ef has now the meaning of strain
          epsF   = wT/status->giveTempElemSize().at(direction);
          _dummy = pow(sigTmax/sigTp,1./aT)*(epsTp-_epsTul);
          do {
            nite++;
            help    = _eps - _epsTul - _dummy;
            R       = exp(-help / epsF) - pow(_dummy/(epsTp-_epsTul),aT);
            Lhs     = exp(-help / epsF) / epsF - aT*pow(_dummy/(epsTp-_epsTul),aT-1.0)/(epsTp-_epsTul);
            _dummy -= R / Lhs;
            if ( nite > 40 ) {
              OOFEM_ERROR("SHCC: computeDamageParamForCohesiveCrack::algorithm not converging - can be caused by large element size");
            }
          } while ( fabs(R) >= epsT0 * SHCC_ITERATION_LIMIT );

          help      = _eps - _epsTul - _dummy;
          answer(0) = max( SHCC_ZERO, sigTp*exp(-help/epsF) );
          answer(1) = min( -SHCC_ZERO, -sigTp*exp(-help/epsF)/epsF  );
        }
        else if( softType == ST_Polynomial_Cohesive_Crack )
        {
          // Nonlinear Tension Softening (Moelands and Reinhardt) - viz DIANA manual
          if( _eps < epsTu(0) )
          {
            double R, Lhs, help, epsF; 
            double _epsTul = status->giveTempEpsTulForEpsTp().at(direction);
            int nite = 0;
            // iteration to achieve objectivity
            // we are looking for a state in which the elastic stress (unloading) is equal to
            // the stress from crack-opening relation
            // ef has now the meaning of strain
            epsF   = wT/status->giveTempElemSize().at(direction);
            //_dummy = pow(sigTmax(direction)/sigTp,1./aT)*(epsTp-_epsTul);
            _dummy = (epsTp-_epsTul);
            do {
              nite++;
              help    = _eps - _epsTul - _dummy;
              R       = pow(1.0-help/epsF,aT) - pow(_dummy/(epsTp-_epsTul),aT);
              Lhs     = aT*pow(1.0-help/epsF,aT-1.0) / epsF - aT*pow(_dummy/(epsTp-_epsTul),aT-1.0)/(epsTp-_epsTul);
              _dummy -= R / Lhs;
              if ( nite > 40 ) {
                OOFEM_ERROR("SHCC: computeDamageParamForCohesiveCrack: algorithm not converging\nFor polynomial cohesive crack it can be caused by bad parameter aT,element size, etc.");
              }
            } while ( fabs(R) >= epsT0 * SHCC_ITERATION_LIMIT );

            help      = _eps - _epsTul - _dummy;
            answer(0) = max( SHCC_ZERO, sigTp * pow(1.0-help/epsF,aT) );
            answer(1) = min( -SHCC_ZERO, -aT * sigTp * pow(1.0-help/epsF,aT-1.0) / epsF );
          }
          else
          {
            answer(0) = SHCC_ZERO;
            answer(1) = -SHCC_ZERO;
          }
        }
      }

      /* update driving parameters */
      epsTmax = _eps;
      sigTmax = answer(0);

      status->letTempEpsTmaxAt(epsTmax, direction);
      status->letTempSigTmaxAt(sigTmax, direction);
      status->letTempEpsTprlAt(epsTmax, direction);
      status->letTempSigTprlAt(sigTmax, direction);

      status->letTempEpsTulAt(giveEpsTul(epsTmax, sigTmax, direction, status), direction);

      // damage parameter
      status->letTempDamageAt(giveDamage(true, direction, status), direction);
      if( isDamageGrowth(true,direction,status) )
      {
        double dEpsTul = bT;
        double epsTul  = status->giveTempEpsTul().at(direction);
        
        dDamage.at(direction) =  -answer(1)/((epsTmax-epsTul)*E) + sigTmax*(1.0-dEpsTul)/
                                 (E*pow(epsTmax-epsTul, 2.0));
      }
      else
        dDamage.at(direction) = 0.0;
                                
    } 
    else // unloading, reloading    
    {

      double epsTul = giveEpsTul(epsTmax, sigTmax, direction, status);
      status->letTempEpsTulAt(epsTul, direction);

      if( epsTmax <= epsT0 ) // elastic behavior
      {
        answer(0) = max( SHCC_ZERO, E*_eps );
        answer(1) = E;

        if( _eps < 0.0 ) changeState = true;
      } 
      else if( (softType == ST_Hermit || softType == ST_Linear || softType == ST_Polynomial_Cohesive_Crack) &&
                epsTp <= epsTmax && epsTu(0) <= epsTmax )
      { // this is some kind of tricky because of the nonlinear unloading - has to be probably checked and changed
        answer(0) = SHCC_ZERO;
        answer(1) = SHCC_ZERO;

        if( (_eps-epsTul) < 0.0 ) changeState = true;
      }
      else if( dprincStrain.at(direction) < 0.0 ) // unloading
      {
        setTmaxTul_calc( true, direction, status );  // sets up (un)loading parameters and changes loading state

        _dummy    = (_eps-epsTul)/(epsTmax_calc-epsTul);
        answer(0) = max( SHCC_ZERO, sigTmax_calc * pow(_dummy, aT) );
        answer(1) = max( SHCC_ZERO, sigTmax_calc * aT * pow(_dummy, aT-1.0)/(epsTmax_calc-epsTul) );

        if( (_eps-epsTul) < 0.0 )
        {
          answer(0)   = 0.0;
          answer(1)   = 0.0;
          changeState = true;
        }

        // update driving parameters
        status->letTempEpsTpulAt(max(epsTul,_eps), direction);
        status->letTempSigTpulAt(max(0.0,answer(0)), direction);
      }   
      else // reloading
      {
        setTmaxTul_calc( false, direction, status );  // sets up (un)loading parameters and change loading state

        answer(0) = max( SHCC_ZERO, sigTul_calc+(sigTmax-sigTul_calc)
                                *(_eps-epsTul_calc)/(epsTmax-epsTul_calc) );
        answer(1) = max( SHCC_ZERO, (sigTmax-sigTul_calc)/(epsTmax-epsTul_calc) );

        // update driving parameters
        status->letTempEpsTprlAt(_eps, direction);
        status->letTempSigTprlAt(answer(0), direction);
      }
  
      // damage parameter
      status->letTempDamageAt(giveDamage(true, direction, status), direction);
      dDamage.at(direction) = 0.0;
    }

    return answer;
  }


  FloatArray 
  SHCCMaterial :: giveCompressiveStress( int direction, GaussPoint *gp, SHCCMaterialStatus *status )
  {     
    // answer(0) = stress, answer(1) = tengent;
    FloatArray answer(2);

    if( (status->giveTempActiveState().at(direction) == tensionSS || status->giveTempCCrackClosing().at(direction))
      && !(status->giveTempEpsTmax().at(direction) <= epsT0 && status->giveTempEpsCmin().at(direction) >= epsC0) )
    { 
      //loading after tension
      answer = giveCompressionAfterTension( direction, gp, status );
    }
    else
    { 
      answer = giveCompressionOrdinaryLoading( direction, gp, status );
    }  

    return answer;
  }


  FloatArray 
  SHCCMaterial :: giveCompressionAfterTension( int direction, GaussPoint *gp, SHCCMaterialStatus *status )
  { 

    // set up of driving variables
    setTmaxCmin_calc( direction, status );

    double _dummy;
    if( status->giveTempActiveState().at(direction) == tensionSS )
    {
        // setup of EpsCul parametrs on the value of EpsTul
        _dummy = giveClosingStrain( false, direction, status );
        status->letTempEpsCulAt(_dummy, direction);
        status->letTempEpsCpulAt(_dummy, direction);
        status->letTempSigCpulAt(0.0, direction);
    }


    // answer(0) = stress, answer(1) = tengent;
    FloatArray answer(2);

    double _eps, _exponent;
    double epsTmax = status->giveTempEpsTmax().at(direction);
    double sigTmax = status->giveTempSigTmax().at(direction);
    double epsCpul = status->giveTempEpsCpul().at(direction);
    double epsCul  = status->giveTempEpsCul().at(direction);
    double sigCpul = status->giveTempSigCpul().at(direction);
    double epsCmin = status->giveTempEpsCmin().at(direction);
    double sigCmin = status->giveTempSigCmin().at(direction);
    double epsCprl = status->giveTempEpsCprl().at(direction);
    double sigCprl = status->giveTempSigCprl().at(direction);

    // setup of logical parameters 
    status->letTempCCrackClosingAt(true, direction);
    status->letTempActiveStateAt(compressionSS, direction);

    // setup of damage parameter 
    status->letTempDamageAt(giveDamage(false, direction, status), direction);
    dDamage.at(direction) = 0.0;


    // stress evaluation
    if( dprincStrain.at(direction) < 0.0 )
    {
      //sirka oblasti na ktere se prejde na normalni zatezovani
      // pri prvnim behu odecneni epsCpul = epsTul => sirka oblasti mezi epsTul a epsCmin,
      // nasledne zmenseni oblasti o neuzaviratelne trhliny
      _dummy  = min( epsC0, epsCmin ) - epsCpul;
      _dummy += (1.0-bT_closing) * giveEpsTul(epsTmax,sigTmax,direction,status);  // zmenseni o oblast neuzaviratelnych trhlin

      // relative strain <_dummy;0.0>
      _eps = princStrain.at(direction) - epsCpul;

      if( _eps > _dummy )  // loading with crack closing
      {
        if( sigCmin_calc > -SHCC_SMALL ) 
        {
          answer(0) = sigCpul;
          answer(1) = (sigCmin_calc/epsCmin_calc);

          // update driving parameters
          status->letTempEpsCprlAt(_eps + epsCpul,direction);
          status->letTempSigCprlAt(answer(0), direction);

          status->letTempEpsCulAt(_eps + epsCpul,direction);
        }
        else  
        {
          _exponent = _dummy*sigCmin_calc/((sigCmin_calc-sigCpul)*epsCmin_calc);

          answer(0) = min( -SHCC_ZERO, sigCpul + (sigCmin_calc-sigCpul) *
                                  pow(_eps/_dummy, _exponent) );

          answer(1) = max( SHCC_ZERO, (sigCmin_calc-sigCpul) * (_exponent)/_dummy *
                                 pow(_eps/_dummy, _exponent-1.0) );

          // update driving parameters
          epsCprl = _eps + epsCpul;
          sigCprl = answer(0);
          status->letTempEpsCprlAt(epsCprl, direction); // ulozeni skutecne hodnoty strainu (neposunute)
          status->letTempSigCprlAt(sigCprl, direction);

          // ulozeni skutecne hodnoty strainu epsCul (neposunute)
          _dummy  = max( giveEpsCul(epsCmin, sigCmin, direction, status) +
                         (1.0-bT_closing) * giveEpsTul(epsTmax,sigTmax,direction, status),
                         epsCprl-(epsCmin_calc/sigCmin_calc) * answer(0) );

          status->letTempEpsCulAt(_dummy, direction);
        }
      }
      else  // normal loading
      {
        status->letTempCCrackClosingAt(false, direction);
        answer = giveCompressionOrdinaryLoading( direction, gp, status );
      }
    }
    else  // unloading
    {
      _eps = princStrain.at(direction) - epsCul;

      if( _eps <= 0.0 ) 
      {
        if( fabs(epsCprl-epsCul)<SHCC_ZERO )
        {
          answer(0) = -SHCC_ZERO;
          answer(1) = SHCC_ZERO;
        }
        else   
        {
          _dummy   = _eps/(epsCprl-epsCul);

          answer(0) = min( -SHCC_ZERO, sigCprl* pow(_dummy, aC) );
          answer(1) = max( SHCC_ZERO, sigCprl * aC * pow(_dummy, (aC-1.0))/(epsCprl-epsCul) );
        }

        // update driving parameters
        status->letTempEpsCpulAt(_eps + epsCul, direction);
        status->letTempSigCpulAt(answer(0), direction);
      }
      else
      {
        answer = giveTensileStress( direction, gp, status );
      }
    }

    return answer;
  }

  
  FloatArray 
  SHCCMaterial :: giveCompressionOrdinaryLoading( int direction, GaussPoint *gp, SHCCMaterialStatus *status )
  {
    // answer(0) = stress, answer(1) = tengent;
    FloatArray answer(2);
    
    double _dummy, _eps;
    double epsCmin = status->giveTempEpsCmin().at(direction);
    double sigCmin = status->giveTempSigCmin().at(direction);

    // setup
    status->letTempCCrackClosingAt(false, direction);
    status->letTempActiveStateAt(compressionSS, direction);

    _eps = giveShiftedStrain( false, direction, status );

    if( (_eps-epsCmin) < SHCC_SMALL  )   // initial loading - failure envelope
    {
      if( _eps >= epsC0 ) 
      {
        answer(0) = E*_eps;
        answer(1) = E;
      }
      else if( _eps >= epsCp ) 
      {
        answer(0) = sigCp - (sigCp-epsC0*E) * pow((epsCp-_eps)/(epsCp-epsC0), E*(epsCp-epsC0)/(sigCp-epsC0*E));

        _dummy    = E * pow((epsCp-_eps)/(epsCp-epsC0), (E*(epsCp-epsC0)/(sigCp-epsC0*E))-1.0);
        answer(1) = max( SHCC_ZERO, _dummy );

        // linear curve - test
        //_dummy    = (_eps-epsC0)/(epsCp-epsC0);
        //answer(0) = E*epsC0 + _dummy * (sigCp - E*epsC0);
        //answer(1) = max( SHCC_ZERO, _dummy * (sigCp - E*epsC0)/(epsCp-epsC0) );
      } 
      else
      {
        // fixing the element size - in the first occasion when epsCp is exceeded
        if( status->giveTempElemSize().at(direction) == 0.0 )
        {
          status->letTempEpsCulForEpsCpAt(giveEpsCul(epsCp, sigCp, direction, status), direction);

          initParametersCrushing(direction,gp,status);

          status->letTempElemSizeAt(Le, direction);
        }

        if( crushType == CT_Hermit ) {
          if( _eps > epsCu(0) ) 
          {
            _dummy    = (_eps-epsCp)/(epsCu(0)-epsCp);
            answer(0) = min( -SHCC_ZERO, sigCp*(2.0 * pow(_dummy, 3.) - 3.0 * pow(_dummy, 2.) + 1.0) );

            answer(1) = min( -SHCC_ZERO, sigCp*(6.0 * pow(_dummy, 2.) - 6.0 * _dummy)/(epsCu(0)-epsCp) ); 
          }    
          else
          {
            answer(0) = -SHCC_ZERO;
            answer(1) = -SHCC_ZERO;
          }
        }
        else if( crushType == CT_Thorenfeldt_Cohesive_Crack )
        {

          // Thorenfeldt based cohesive function for crushing ("overlapping") - iteration needed
          double R, Lhs, help, epsF; 
          double _epsCul = status->giveTempEpsCulForEpsCp().at(direction);
          int nite = 0;
          // iteration to achieve objectivity
          // we are looking for a state in which the elastic stress (unloading) is equal to
          // the stress from crack-opening relation
          // ef has now the meaning of strain
          epsF   = wC/status->giveTempElemSize().at(direction);
          //_dummy = pow(sigCmin(direction)/sigCp,1./aC)*(epsCp-_epsCul);
          _dummy = epsCp-_epsCul;
          do {
            nite++;
            help    = _eps - _epsCul - _dummy;
            R       =   (1+help/epsF) * betaC/(betaC-1+pow(1+help/epsF,betaC))
                      - pow(_dummy/(epsCp-_epsCul),aC);
            Lhs     = - betaC/(betaC-1+pow(1+help/epsF,betaC))/epsF
                      + (1+help/epsF) * betaC/pow(betaC-1+pow(1+help/epsF,betaC),2.0)*betaC*pow(1+help/epsF,betaC-1.0)/epsF
                      - aC*pow(_dummy/(epsCp-_epsCul),aC-1.0)/(epsCp-_epsCul);
            _dummy -= R / Lhs;
            if ( nite > 40 ) {
              OOFEM_ERROR("SHCC: computeDamageParamForOverlapping: algorithm not converging - can be caused by large element size");
            }
          } while ( fabs(R) >= fabs(epsC0 * SHCC_ITERATION_LIMIT) );

          help      = _eps - _epsCul - _dummy;
          answer(0) = min( -SHCC_ZERO, sigCp * (1+help/epsF) * betaC/(betaC-1+pow(1+help/epsF,betaC)) );
          answer(1) =  - sigCp * (betaC/(betaC-1+pow(1+help/epsF,betaC))/epsF
                       - (1+help/epsF) * betaC/pow(betaC-1+pow(1+help/epsF,betaC),2.0)*betaC*pow(1+help/epsF,betaC-1.0)/epsF) ;

        }
        else if( crushType == CT_Exponential_Cohesive_Crack )
        {

          // exponential cohesive function for crushing ("overlapping") - iteration needed
          double R, Lhs, help, epsF; 
          double _epsCul = status->giveTempEpsCulForEpsCp().at(direction);
          int nite = 0;
          // iteration to achieve objectivity
          // we are looking for a state in which the elastic stress (unloading) is equal to
          // the stress from crack-opening relation
          // ef has now the meaning of strain
          epsF   = wC/status->giveTempElemSize().at(direction);
          _dummy = pow(sigCmin/sigCp,1./aC)*(epsCp-_epsCul);
          do {
            nite++;
            help    = _eps - _epsCul - _dummy;
            R       =   exp(-help / epsF)
                      - pow(_dummy/(epsCp-_epsCul),aC);
            Lhs     =   exp(-help / epsF) / epsF
                      - aC*pow(_dummy/(epsCp-_epsCul),aC-1.0)/(epsCp-_epsCul);
            _dummy -= R / Lhs;
            if ( nite > 40 ) {
              OOFEM_ERROR("SHCC: computeDamageParamForOverlapping: algorithm not converging - can be caused by large element size");
            }
          } while ( fabs(R) >= fabs(epsC0 * SHCC_ITERATION_LIMIT) );

          help      = _eps - _epsCul - _dummy;
          answer(0) = min( -SHCC_ZERO, sigCp * exp(-help / epsF) );
          answer(1) =  - sigCp * exp(-help / epsF) / epsF ;

        }
      }

      // update driving parameters
      epsCmin = _eps;
      sigCmin = answer(0);

      status->letTempEpsCminAt(epsCmin, direction);
      status->letTempSigCminAt(sigCmin, direction);
      status->letTempEpsCprlAt(epsCmin, direction);
      status->letTempSigCprlAt(sigCmin, direction);

      status->letTempEpsCulAt(giveEpsCul(epsCmin, sigCmin, direction, status), direction);

      // damage parameter
      status->letTempDamageAt(giveDamage(false, direction, status), direction);
      if( isDamageGrowth(false,direction,status) )
      {
        double dEpsCul = bC;
        double epsCul = status->giveTempEpsCul().at(direction);
        
        dDamage.at(direction) = -answer(1)/((epsCmin-epsCul)*E) + sigCmin*(1.0-dEpsCul)/
                              (E*pow(epsCmin-epsCul, 2.0));
      }
      else
        dDamage.at(direction) = 0.0;

    }       
    else // unloading, reloading  
    {
      double epsCul = giveEpsCul( epsCmin, sigCmin, direction, status );
      status->letTempEpsCulAt(epsCul, direction);

      if( epsC0 <= epsCmin)
      {
        answer(0) = min( -SHCC_ZERO, E*_eps );
        answer(1) = E;

        if( _eps > 0.0 ) changeState = true;
      }
      else if( ( crushType == CT_Hermit ) && epsCp >= epsCmin && epsCu(0) >= epsCmin )
      {  // this is some kind of tricky because of the nonlinear unloading - has to be probably checked and changed

        answer(0) = -SHCC_ZERO;
        answer(1) = -SHCC_ZERO;

        // update driving parameters
        if( (_eps-epsCul) > 0.0 ) changeState = true;

      }
      else if( dprincStrain.at(direction) > 0.0 ) // unloading
      {
        setCminCul_calc( true, direction, status );

        _dummy    = (_eps-epsCul)/(epsCmin_calc-epsCul);
        answer(0) = min( -SHCC_ZERO, sigCmin_calc * pow(_dummy, aC) );                     
        answer(1) = max( SHCC_ZERO, sigCmin_calc * aC * pow(_dummy, aC-1.0)/(epsCmin_calc-epsCul) );

        // update driving parameters
        status->letTempEpsCpulAt(min(epsCul,0.0), direction);
        status->letTempSigCpulAt(min(0.0,answer(0)), direction);

        if( (_eps-epsCul) > 0.0 )   // for cyclic loading
        {
          answer(0)   = 0.0;
          answer(1)   = 0.0;
          changeState = true;
        }
      }                               
      else // reloading
      {
        setCminCul_calc( false, direction, status );

        answer(0) = min( -SHCC_ZERO, sigCul_calc+(sigCmin-sigCul_calc)
                                *(_eps-epsCul_calc)/(epsCmin-epsCul_calc) );
        answer(1) = min( -SHCC_ZERO, (sigCmin-sigCul_calc)/(epsCmin-epsCul_calc) );

        // update driving parameters
        status->letTempEpsCprlAt(_eps, direction);
        status->letTempSigCprlAt(answer(0), direction);
      }

      // damage parameter
      status->letTempDamageAt(giveDamage(false, direction, status), direction);
      dDamage.at(direction) = 0.0;
    }

    return answer;
  }


  // general
  void
  SHCCMaterial :: setTmaxTul_calc(bool unloadingState, int direction, SHCCMaterialStatus *status )
  { 
    if( unloadingState ) 
    {
      epsTmax_calc = status->giveTempEpsTprl().at(direction);
      sigTmax_calc = status->giveTempSigTprl().at(direction);

      epsTul_calc  = status->giveTempEpsTul().at(direction);
      sigTul_calc  = 0.0;
    }
    else
    {
      epsTmax_calc = status->giveTempEpsTmax().at(direction);
      sigTmax_calc = status->giveTempSigTmax().at(direction);

      epsTul_calc  = status->giveTempEpsTpul().at(direction);
      sigTul_calc  = status->giveTempSigTpul().at(direction);
    } 
  }

  
  void 
  SHCCMaterial :: setCminCul_calc( bool unloadingState, int direction, SHCCMaterialStatus *status )
  {
    if( unloadingState ) 
    {
      epsCmin_calc = status->giveTempEpsCprl().at(direction);
      sigCmin_calc = status->giveTempSigCprl().at(direction);

      epsCul_calc  = status->giveTempEpsCul().at(direction);
      sigCul_calc  = 0.0;
    }   
    else
    {
      epsCmin_calc = status->giveTempEpsCmin().at(direction);
      sigCmin_calc = status->giveTempSigCmin().at(direction);

      epsCul_calc  = status->giveTempEpsCpul().at(direction);
      sigCul_calc  = status->giveTempSigCpul().at(direction);
    }
  }

  
  void 
  SHCCMaterial :: setTmaxCmin_calc( int direction, SHCCMaterialStatus *status )
  {
    double epsTmax = status->giveTempEpsTmax().at(direction);
    double sigTmax = status->giveTempSigTmax().at(direction);
    double epsCmin = status->giveTempEpsCmin().at(direction);
    double sigCmin = status->giveTempSigCmin().at(direction);

    if( epsTmax <= epsT0 )
    {
      epsTmax_calc = epsT0;
      sigTmax_calc = E*epsT0;
    }
    else 
    {
      epsTmax_calc = epsTmax - giveEpsTul(epsTmax,sigTmax,direction,status);
      sigTmax_calc = sigTmax;

    }  

    if( epsCmin >= epsC0 )
    {
      epsCmin_calc = epsC0;
      sigCmin_calc = E*epsC0;
    }
    else 
    {
      epsCmin_calc = epsCmin - giveEpsCul(epsCmin,sigCmin,direction,status);
      sigCmin_calc = sigCmin;
    }
  }

  
  double 
  SHCCMaterial :: giveShiftedStrain( bool tensileLoad, int direction, SHCCMaterialStatus *status )
  {
    double _shiftedStrain = princStrain.at(direction);

    if( tensileLoad ) {
      double epsCmin = status->giveTempEpsCmin().at(direction);
      double sigCmin = status->giveTempSigCmin().at(direction);
      _shiftedStrain -= (1.0-bC_closing) * giveEpsCul( epsCmin, sigCmin, direction, status );

    } else {
      double epsTmax = status->giveTempEpsTmax().at(direction);
      double sigTmax = status->giveTempSigTmax().at(direction);
      _shiftedStrain -= (1.0-bT_closing) * giveEpsTul( epsTmax, sigTmax, direction, status );
    }

    return( _shiftedStrain );
  }

  
  double 
  SHCCMaterial :: giveClosingStrain( bool tensileLoad, int direction, SHCCMaterialStatus *status )
  {
    double _value;

    if( tensileLoad ) 
    {
      if( status->giveCCrackClosing().at(direction) ) {
        _value = status->giveTempEpsCul().at(direction);
      } else {
        double epsCmin = status->giveTempEpsCmin().at(direction);
        double sigCmin = status->giveTempSigCmin().at(direction);
        _value = giveEpsCul( epsCmin, sigCmin, direction, status );
      }
    }
    else
    {
      if( status->giveTCrackClosing().at(direction) ) {
        _value = status->giveTempEpsTul().at(direction);
      } else {
        double epsTmax = status->giveTempEpsTmax().at(direction);
        double sigTmax = status->giveTempSigTmax().at(direction);
        _value = giveEpsTul( epsTmax, sigTmax, direction, status );
      }
    }

    return( _value );
  }

  
  double
  SHCCMaterial :: giveDamage( bool tensileLoad, int direction, SHCCMaterialStatus *status )
  {
    double _value = 0.0;
    double damage_old = status->giveDamage().at(direction);

    if( tensileLoad )
    {
      double epsTmax = status->giveTempEpsTmax().at(direction);
      double sigTmax = status->giveTempSigTmax().at(direction);
      if( epsTmax <= epsT0 )
        _value = 0.0;
      else
      {
        if( softType == ST_Hermit || softType == ST_Linear || softType == ST_Polynomial_Cohesive_Crack ) // Hermit function softening
        {
          if( epsTmax <= epsTp || epsTmax < epsTu.at(1) )
            _value =  max( 0.0, 1.0 - sigTmax/
                      ((epsTmax - giveEpsTul(epsTmax, sigTmax, direction, status))*E) );
          else
            _value = 1.0;
        }
        else if( softType == ST_Exponential_Cohesive_Crack ) 
        {
          _value =  max( 0.0, 1.0 - sigTmax/((epsTmax - giveEpsTul(epsTmax, sigTmax, direction, status))*E) );
        }
        else
        {
          OOFEM_ERROR( "SHCCMaterial: giveDamage: Unknown softType!!!" );
        }
      }
    }
    else
    {
      double epsCmin = status->giveTempEpsCmin().at(direction);
      double sigCmin = status->giveTempSigCmin().at(direction);
      if( epsCmin >= epsC0 )
        _value = 0.0;
      else 
      {
        if( crushType == CT_Hermit )
        {
          if( epsCmin >= epsCp || epsCmin > epsCu.at(1) )
            _value = max(0.0, 1.0 - sigCmin/
                     ((epsCmin - giveEpsCul(epsCmin, sigCmin, direction, status))*E) );
          else
            _value = 1.0;
        }
        else if( crushType == CT_Thorenfeldt_Cohesive_Crack || crushType == CT_Exponential_Cohesive_Crack )
        {
          _value = max(0.0, 1.0 - sigCmin/((epsCmin - giveEpsCul(epsCmin, sigCmin, direction, status))*E) );
        }
        else
        {
          OOFEM_ERROR( "SHCCMaterial: giveDamage: Unknown crushType!!!" );
        }

      }
    }

    return( min( SHCC_MAXDAMAGE, max(damage_old, _value) ) );
  }

  bool 
  SHCCMaterial :: isDamageGrowth( bool tensileLoad, int direction, SHCCMaterialStatus *status )
  {
    double damage_old = status->giveDamage().at(direction);;
    
    if( damage_old == giveDamage(tensileLoad,direction,status) )
      return false;
    else
      return true;
  }

  double 
  SHCCMaterial :: giveEpsTul( double eps, double sig, int direction, SHCCMaterialStatus *status )
  {
    // eps, sig => point on the failure envelope
    double dum;
    double epsTul_old = status->giveEpsTul().at(direction);

    if( eps <= epsT0 ) 
      dum = 0.0;
    else 
      dum = bT*(eps-epsT0);   


    return( max(dum,epsTul_old) );
  }

  
  double 
  SHCCMaterial :: giveEpsCul( double eps, double sig, int direction, SHCCMaterialStatus *status )
  {
    // eps, sig => point on the failure envelope
    double dum;
    double epsCul_old = status->giveEpsCul().at(direction);

    if( eps >= epsC0 ) 
      dum = 0.0;
    else 
      dum = bC*(eps-epsC0);
    

    return( min(dum,epsCul_old) );

  }

  FloatMatrix 
  SHCCMaterial :: giveRotationMatrix( double angle ) 
  { 
    FloatMatrix answer(2,2);

    answer(0,0) = cos( angle );
    answer(1,1) = cos( angle );

    answer(0,1) = sin( angle );
    answer(1,0) = - answer(0,1);

    return answer;
  }

  
  double
  SHCCMaterial :: give2DRotationalAngleStrain( const FloatArray &strain )
  {
    double _angle;

    if( fabs(strain(0)-strain(1)) < SHCC_ZERO )
    {
      if( fabs(strain(2)) < SHCC_ZERO )
        _angle = 0.0;
      else if( (strain(0)-strain(1)) >= 0.0 )
        _angle = 0.25*M_PI;
      else
        _angle = -0.25*M_PI;
    }
    else
      _angle = 0.5 * atan(strain(2)/(strain(0)-strain(1)));

    return( _angle );
  }

  
  FloatMatrix
  SHCCMaterial :: giveTransformationMatrixStrain( double angle )
  { 
    double _n11, _n22, _n12, _n21;
    FloatMatrix answer(3,3);

    _n11 = cos(angle);
    _n22 = cos(angle);
    _n12 = sin(angle);
    _n21 = -sin(angle);


    answer(0,0) = _n11*_n11;
    answer(1,1) = _n22*_n22;

    answer(0,1) = _n12*_n12;
    answer(1,0) = _n21*_n21;

    answer(0,2) = _n11*_n12;
    answer(1,2) = _n21*_n22;

    answer(2,0) = 2.0*_n11*_n21;
    answer(2,1) = 2.0*_n22*_n12;
    answer(2,2) = _n11*_n22 + _n12*_n21; 

    return answer;
  }

  
  FloatMatrix
  SHCCMaterial :: giveTransformationMatrixStress( double angle )
  { 
    double _n11, _n22, _n12, _n21;
    FloatMatrix answer(3,3);

    _n11 = cos(angle);
    _n22 = cos(angle);
    _n12 = sin(angle);
    _n21 = -sin(angle);


    answer(0,0) = _n11*_n11;
    answer(1,1) = _n22*_n22;

    answer(0,1) = _n12*_n12;
    answer(1,0) = _n21*_n21;

    answer(0,2) = 2.0*_n11*_n12;
    answer(1,2) = 2.0*_n21*_n22;

    answer(2,0) = _n11*_n21;
    answer(2,1) = _n22*_n12;
    answer(2,2) = _n11*_n22 + _n12*_n21;  

    return answer;
  }

  FloatMatrix&
  SHCCMaterial :: giveTangetialStiffMatrix( FloatArray &principalStress, FloatArray &tangent, double rotAngle, SHCCMaterialStatus *status )
  {
    FloatArray _dNu(2), _nu(2), _stress(2);
    FloatMatrix answer(3,3), _T, _Tt, _dumMat;
    FloatArray damage(status->giveTempDamage());

    _nu(0) = nu*(1.0 - damage(0));
    _nu(1) = nu*(1.0 - damage(1));

    _stress(0) = (principalStress(0) + _nu(0)*principalStress(1))/(1.-_nu(0)*_nu(1));  
    _stress(1) = (principalStress(1) + _nu(1)*principalStress(0))/(1.-_nu(0)*_nu(1));
    

    _dNu(0) = - nu * dDamage(0);
    _dNu(1) = - nu * dDamage(1);  

    answer(0,0) = (tangent(0) + principalStress(1)*_dNu(0))/(1.-_nu(0)*_nu(1)) + 
                  (principalStress(0) + _nu(0)*principalStress(1))/pow(1.-_nu(0)*_nu(1), 2.)*_nu(1)*_dNu(0);
    answer(0,1) = (_nu(0)*tangent(1))/(1.-_nu(0)*_nu(1)) + 
                  (principalStress(0) + _nu(0)*principalStress(1))/pow(1.-_nu(0)*_nu(1), 2.)*_nu(0)*_dNu(1);

    answer(1,0) = (_nu(1)*tangent(0))/(1.-_nu(0)*_nu(1)) + 
                  (principalStress(1) + _nu(1)*principalStress(0))/pow(1.-_nu(0)*_nu(1), 2.)*_nu(1)*_dNu(0);
    answer(1,1) = (tangent(1) + principalStress(0)*_dNu(1))/(1.-_nu(0)*_nu(1)) + 
                  (principalStress(1) + _nu(1)*principalStress(0))/pow(1.-_nu(0)*_nu(1), 2.)*_nu(0)*_dNu(1);

    if( fabs(_stress(0)-_stress(1)) < SHCC_SMALL && fabs(princStrain(0)-princStrain(1)) < SHCC_SMALL )
    {
      answer(2,2) = (answer(0,0) - answer(0,1))/2.0;
      //answer(2,2) = E/(2*(1.+nu));
    }
    else    
    {
      answer(2,2) = (_stress(0)-_stress(1))/(2.*(princStrain(0)-princStrain(1)));
    }

  
    _T = giveTransformationMatrixStrain( rotAngle );
    _Tt.beTranspositionOf(_T);
    _dumMat.beProductOf(_Tt,answer);
    answer.beProductOf(_dumMat,_T);

    return answer;
  }

  IRResultType
  SHCCMaterial :: initializeFrom(InputRecord *ir)
  {
    IRResultType result;                   // Required by IR_GIVE_FIELD macro

    StructuralMaterial :: initializeFrom(ir);
    linearElasticMaterial->initializeFrom(ir); // takes care of elastic constants

    E  = ( ( IsotropicLinearElasticMaterial * ) linearElasticMaterial )->giveYoungsModulus();
    nu = ( ( IsotropicLinearElasticMaterial * ) linearElasticMaterial )->givePoissonsRatio();

    IR_GIVE_FIELD(ir, epsT0,       _IFT_SHCCMaterial_epsT0);       
    IR_GIVE_FIELD(ir, epsTp,       _IFT_SHCCMaterial_epsTp);        
    IR_GIVE_FIELD(ir, sigTp,       _IFT_SHCCMaterial_sigTp);       
    IR_GIVE_FIELD(ir, wT,          _IFT_SHCCMaterial_wT);          
    IR_GIVE_FIELD(ir, aT,          _IFT_SHCCMaterial_aT);          
    IR_GIVE_FIELD(ir, bT,          _IFT_SHCCMaterial_bT);         
    IR_GIVE_FIELD(ir, bT_closing,  _IFT_SHCCMaterial_bT_closing);
    IR_GIVE_FIELD(ir, epsC0,       _IFT_SHCCMaterial_epsC0);
    IR_GIVE_FIELD(ir, epsCp,       _IFT_SHCCMaterial_epsCp);       
    IR_GIVE_FIELD(ir, sigCp,       _IFT_SHCCMaterial_sigCp);                
    IR_GIVE_FIELD(ir, aC,          _IFT_SHCCMaterial_aC);          
    IR_GIVE_FIELD(ir, bC,          _IFT_SHCCMaterial_bC);        
    IR_GIVE_FIELD(ir, bC_closing,  _IFT_SHCCMaterial_bC_closing); 

    aC = floor(aC+0.5); // rounding
    bC = floor(bC+0.5); // rounding

    int sType = 0;
    IR_GIVE_OPTIONAL_FIELD(ir, sType,   _IFT_SHCCMaterial_softType);
    switch ( sType ) {
      case 1: softType = ST_Linear;
        break;
      case 2: softType = ST_Exponential_Cohesive_Crack;
        break;
      case 3: softType = ST_Hermit;
        break;
      default: 
        softType = ST_Polynomial_Cohesive_Crack;
    }

    int cType = 0;
    IR_GIVE_OPTIONAL_FIELD(ir, cType,   _IFT_SHCCMaterial_crushType);
    switch ( cType ) {
      case 1: {
        crushType = CT_Hermit;
        IR_GIVE_FIELD(ir, dC,          _IFT_SHCCMaterial_dC); 
        IR_GIVE_FIELD(ir, epsCu_input, _IFT_SHCCMaterial_epsCu_input); 

        // aby to nervalo - nepouziji se
        IR_GIVE_OPTIONAL_FIELD(ir, wC,          _IFT_SHCCMaterial_wC);        
        IR_GIVE_OPTIONAL_FIELD(ir, betaC,       _IFT_SHCCMaterial_betaC); 
        break;
      }
      case 2: {
        crushType = CT_Exponential_Cohesive_Crack;
        IR_GIVE_FIELD(ir, wC,          _IFT_SHCCMaterial_wC);                

        // aby to nervalo - nepouziji se
        IR_GIVE_OPTIONAL_FIELD(ir, betaC,       _IFT_SHCCMaterial_betaC); 
        IR_GIVE_OPTIONAL_FIELD(ir, dC,          _IFT_SHCCMaterial_dC); 
        IR_GIVE_OPTIONAL_FIELD(ir, epsCu_input, _IFT_SHCCMaterial_epsCu_input); 
        break;
      }
      default: {
        crushType = CT_Thorenfeldt_Cohesive_Crack;
        IR_GIVE_FIELD(ir, wC,          _IFT_SHCCMaterial_wC);        
        IR_GIVE_FIELD(ir, betaC,       _IFT_SHCCMaterial_betaC); 

        // aby to nervalo - nepouziji se
        IR_GIVE_OPTIONAL_FIELD(ir, dC,          _IFT_SHCCMaterial_dC); 
        IR_GIVE_OPTIONAL_FIELD(ir, epsCu_input, _IFT_SHCCMaterial_epsCu_input); 
      }
    }


    int ecsm = 0;
    IR_GIVE_OPTIONAL_FIELD(ir, ecsm, _IFT_SHCCMaterial_ecsm);
    switch ( ecsm ) {
      case 1: ecsMethod = ECSM_SquareRootOfArea;
        break;
      case 2: ecsMethod = ECSM_ProjectionCentered;
        break;
      case 3: ecsMethod = ECSM_Oliver1;
        break;
      case 4: ecsMethod = ECSM_Oliver1modified;
        break;
      default: ecsMethod = ECSM_Projection;
    }

    return IRRT_OK;
  }

  int
  SHCCMaterial :: giveIPValue(FloatArray &answer, GaussPoint *aGaussPoint, InternalStateType type, TimeStep *atTime)
  {
      //MaterialStatus *status = this->giveStatus(aGaussPoint);
      if ( type == IST_DamageScalar ) {
          answer.resize(1);
          answer.zero();
          ///@todo Actually export the relevant damage value here!
          //answer.at(1) = status->giveDamage();
          return 1;
      } else if ( type == IST_DamageTensor ) {
          answer.resize(6);
          answer.zero();

          SHCCMaterialStatus *status = ( SHCCMaterialStatus * ) this->giveStatus(aGaussPoint);
          FloatArray dummy = status->giveDamage();
          answer.at(1) = dummy.at(1);
          answer.at(2) = dummy.at(2);
          return 1;
      } else {
          return StructuralMaterial :: giveIPValue(answer, aGaussPoint, type, atTime);
      }
  }

  MaterialStatus *
  SHCCMaterial :: CreateStatus(GaussPoint *gp) const
  {
    SHCCMaterialStatus *status = new SHCCMaterialStatus(1, SHCCMaterial :: domain, gp);
    return status;
  }


  MaterialStatus *
  SHCCMaterial :: giveStatus(GaussPoint *gp) const
  {
      MaterialStatus *status = static_cast< MaterialStatus * >( gp->giveMaterialStatus() );
      if ( status == NULL ) {
          // create a new one
          status = this->CreateStatus(gp);

          if ( status != NULL ) {
              gp->setMaterialStatus( status, this->giveNumber() );
          }
      }

      return status;
  }

  /*******************************************************/
  SHCCMaterialStatus :: SHCCMaterialStatus(int n, Domain *d, GaussPoint *g) : StructuralMaterialStatus(n, d, g),
      tempEpsTmax(), tempSigTmax(), tempEpsTprl(), tempSigTprl(), tempEpsTpul(), tempSigTpul(), tempEpsTul(), tempEpsTulForEpsTp(),
      tempEpsCmin(), tempSigCmin(), tempEpsCprl(), tempSigCprl(), tempEpsCpul(), tempSigCpul(), tempEpsCul(), tempEpsCulForEpsCp(),
      tempPrincipalStrain(), tempPrincipalStress(),
      tempActiveState(), tempTCrackClosing(), tempCCrackClosing(),
      tempTangentStiffMat(),
      tempDamage(), tempTangent(), tempElemSize(),

      epsTmax(), sigTmax(), epsTprl(), sigTprl(), epsTpul(), sigTpul(), epsTul(), epsTulForEpsTp(),
      epsCmin(), sigCmin(), epsCprl(), sigCprl(), epsCpul(), sigCpul(), epsCul(), epsCulForEpsCp(),
      principalStrain(), principalStress(),
      activeState(), tCrackClosing(), cCrackClosing(),
      tangentStiffMat(),
      damage(), tangent(), elemSize()
  {
    // init state variables
    int dumSize = 0,dumSizeSS = 0;

    // agrees with dimension of the problem
    dumSize   = d->giveNumberOfSpatialDimensions();
    // number of stress and strain elements
    dumSizeSS = StructuralMaterial :: giveSizeOfVoigtSymVector( g->giveMaterialMode() );

    /// temp values ///

    // loading parametres
    tempEpsTmax.resize(dumSize);
    tempSigTmax.resize(dumSize); 
    tempEpsTprl.resize(dumSize); 
    tempSigTprl.resize(dumSize); 
    tempEpsTpul.resize(dumSize); 
    tempSigTpul.resize(dumSize); 
    tempEpsTul.resize(dumSize);  
    tempEpsTulForEpsTp.resize(dumSize);

    tempEpsCmin.resize(dumSize); 
    tempSigCmin.resize(dumSize); 
    tempEpsCprl.resize(dumSize); 
    tempSigCprl.resize(dumSize); 
    tempEpsCpul.resize(dumSize); 
    tempSigCpul.resize(dumSize); 
    tempEpsCul.resize(dumSize);  
    tempEpsCulForEpsCp.resize(dumSize);

    tempPrincipalStrain.resize(dumSizeSS);
    tempPrincipalStress.resize(dumSizeSS);

    // state determining parametres
    tempActiveState.resize(dumSize);     
    tempTCrackClosing.resize(dumSize);   
    tempCCrackClosing.resize(dumSize);

    tempTangentStiffMat.resize(dumSizeSS,dumSizeSS);

    tempDamage.resize(dumSize);
    tempTangent.resize(dumSize);
    tempElemSize.resize(dumSize);
    tempAlpha = 0;

    /// equilibriated variables
    // loading parametres
    epsTmax.resize(dumSize); 
    sigTmax.resize(dumSize); 
    epsTprl.resize(dumSize); 
    sigTprl.resize(dumSize); 
    epsTpul.resize(dumSize); 
    sigTpul.resize(dumSize); 
    epsTul.resize(dumSize);
    epsTulForEpsTp.resize(dumSize);

    epsCmin.resize(dumSize); 
    sigCmin.resize(dumSize); 
    epsCprl.resize(dumSize); 
    sigCprl.resize(dumSize); 
    epsCpul.resize(dumSize); 
    sigCpul.resize(dumSize); 
    epsCul.resize(dumSize);  
    epsCulForEpsCp.resize(dumSize);

    principalStrain.resize(dumSizeSS);
    principalStress.resize(dumSizeSS);

    // state determining parametres
    activeState.resize(dumSize);     
    tCrackClosing.resize(dumSize);   
    cCrackClosing.resize(dumSize);   
    for( int i = 1; i<=dumSize; i++ ) {
        activeState.at(i) = 0; // SHCCMaterial::noneSS
        tCrackClosing.at(i) = false;
        cCrackClosing.at(i) = false;
    }

    tangentStiffMat.resize(dumSizeSS,dumSizeSS);

    damage.resize(dumSize);
    tangent.resize(dumSize);
    elemSize.resize(dumSize);
    alpha = 0;

    initTempStatus();
  }


  SHCCMaterialStatus :: ~SHCCMaterialStatus()
  {}


  void
  SHCCMaterialStatus :: printOutputAt(FILE *file, TimeStep *tStep)
  {
    /// print state to output stream

    StructuralMaterialStatus :: printOutputAt(file, tStep);
    fprintf(file, "status { alpha %e ", this->alpha);
    fprintf(file, "damage " );
    for( int i=1; i<=damage.giveSize(); i++ ) {
      fprintf(file, "%f ", this->damage.at(i));
    }
    fprintf(file, "}\n");
  }

  // initialize temporary state variables according to equilibrated state vars
  void
  SHCCMaterialStatus :: initTempStatus()
  {
    StructuralMaterialStatus :: initTempStatus();

    // loading parametres                   
    tempEpsTmax = epsTmax;              
    tempSigTmax = sigTmax;              
    tempEpsTprl = epsTprl;              
    tempSigTprl = sigTprl;              
    tempEpsTpul = epsTpul;              
    tempSigTpul = sigTpul;            
    tempEpsTul  = epsTul;  
    tempEpsTulForEpsTp = epsTulForEpsTp;

    tempEpsCmin = epsCmin;                 
    tempSigCmin = sigCmin;                
    tempEpsCprl = epsCprl;                
    tempSigCprl = sigCprl;                
    tempEpsCpul = epsCpul;                
    tempSigCpul = sigCpul;                      
    tempEpsCul  = epsCul;   
    tempEpsCulForEpsCp = epsCulForEpsCp;

    tempPrincipalStrain = principalStrain;
    tempPrincipalStress = principalStress;

    // state determining parametres              
    tempActiveState      = activeState;                     
    tempTCrackClosing    = tCrackClosing;                           
    tempCCrackClosing    = cCrackClosing; 

    tempTangentStiffMat  = tangentStiffMat;

    tempDamage           = damage;
    tempTangent          = tangent;
    tempElemSize         = elemSize;
    tempAlpha            = alpha;

  }

  // Called when equilibrium reached, set equilibrated vars according to temporary (working) ones.
  void
  SHCCMaterialStatus :: updateYourself(TimeStep *atTime)
  {
    StructuralMaterialStatus :: updateYourself(atTime);

    // loading parametres                   
    epsTmax = tempEpsTmax;              
    sigTmax = tempSigTmax;              
    epsTprl = tempEpsTprl;              
    sigTprl = tempSigTprl;              
    epsTpul = tempEpsTpul;              
    sigTpul = tempSigTpul;            
    epsTul  = tempEpsTul;    
    epsTulForEpsTp = tempEpsTulForEpsTp;

    epsCmin = tempEpsCmin;                 
    sigCmin = tempSigCmin;                
    epsCprl = tempEpsCprl;                
    sigCprl = tempSigCprl;                
    epsCpul = tempEpsCpul;                
    sigCpul = tempSigCpul;                      
    epsCul =  tempEpsCul;  
    epsCulForEpsCp = tempEpsCulForEpsCp;

    principalStrain = tempPrincipalStrain;
    principalStress = tempPrincipalStress;

    // state determining parametres                 
    activeState      = tempActiveState;                        
    tCrackClosing    = tempTCrackClosing;                            
    cCrackClosing    = tempCCrackClosing;     

    tangentStiffMat  = tempTangentStiffMat;

    damage           = tempDamage;
    tangent          = tempTangent;
    elemSize         = tempElemSize;
    alpha            = tempAlpha;

  }
} // end namespace oofem
