/*
*
*                 #####    #####   ######  ######  ###   ###
*               ##   ##  ##   ##  ##      ##      ## ### ##
*              ##   ##  ##   ##  ####    ####    ##  #  ##
*             ##   ##  ##   ##  ##      ##      ##     ##
*            ##   ##  ##   ##  ##      ##      ##     ##
*            #####    #####   ##      ######  ##     ##
*
*
*             OOFEM : Object Oriented Finite Element Code
*
*               Copyright (C) 1993 - 2011   Borek Patzak
*
*
*
*       Czech Technical University, Faculty of Civil Engineering,
*   Department of Structural Mechanics, 166 29 Prague, Czech Republic
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef shcc_h
#define shcc_h

#include "material.h"
#include "structuralmaterial.h"
#include "structuralms.h"
#include "isolinearelasticmaterial.h"
#include "intarray.h"

///@name Input fields for IsotropicDamageMaterial1
//@{
#define _IFT_SHCCMaterial_Name "shcc"
#define _IFT_SHCCMaterial_epsT0 "epst0"        
#define _IFT_SHCCMaterial_epsTp "epstp"       
#define _IFT_SHCCMaterial_sigTp "sigtp"       
#define _IFT_SHCCMaterial_wT    "wt"       
#define _IFT_SHCCMaterial_aT    "at"       
#define _IFT_SHCCMaterial_bT    "bt"       
#define _IFT_SHCCMaterial_bT_closing "bt_closing"  
#define _IFT_SHCCMaterial_epsC0      "epsc0"  
#define _IFT_SHCCMaterial_epsCp      "epscp"  
#define _IFT_SHCCMaterial_sigCp      "sigcp"  
#define _IFT_SHCCMaterial_epsCu_input "epscu" 
#define _IFT_SHCCMaterial_dC          "dc" 
#define _IFT_SHCCMaterial_aC          "ac" 
#define _IFT_SHCCMaterial_bC          "bc" 
#define _IFT_SHCCMaterial_bC_closing  "bc_closing" 
#define _IFT_SHCCMaterial_wC          "wc"
#define _IFT_SHCCMaterial_betaC       "betac"
#define _IFT_SHCCMaterial_ecsm        "ecsm"

#define _IFT_SHCCMaterial_softType    "softtype"
#define _IFT_SHCCMaterial_crushType   "crushtype"
//@}

#define SHCC_ITERATION_LIMIT 1.e-9
#define SHCC_ZERO     1.0e-12
#define SHCC_SMALL    1.0e-8
#define SHCC_BIGNUMBER 1.0e8
#define SHCC_MAXDAMAGE 0.999999


namespace oofem {
  /**
  * This class implements associated MaterialStatus for SHCCMaterial.
  */
  class SHCCMaterialStatus : public StructuralMaterialStatus
  {
  protected:
    // loading parametres
    FloatArray epsTmax; 
    FloatArray sigTmax; 
    FloatArray epsTprl; 
    FloatArray sigTprl; 
    FloatArray epsTpul; 
    FloatArray sigTpul; 
    FloatArray epsTul;  
    FloatArray epsTulForEpsTp;

    FloatArray epsCmin; 
    FloatArray sigCmin; 
    FloatArray epsCprl; 
    FloatArray sigCprl; 
    FloatArray epsCpul; 
    FloatArray sigCpul; 
    FloatArray epsCul;  
    FloatArray epsCulForEpsCp;

    FloatArray principalStrain,principalStress;

    // state determining parametres
    IntArray activeState;
    IntArray tCrackClosing;
    IntArray cCrackClosing;

    FloatMatrix tangentStiffMat;

    FloatArray damage, tangent;
    FloatArray elemSize;
    double alpha; // rotation of principal axis

    // temp values

    // loading parametres
    FloatArray tempEpsTmax; 
    FloatArray tempSigTmax; 
    FloatArray tempEpsTprl; 
    FloatArray tempSigTprl; 
    FloatArray tempEpsTpul; 
    FloatArray tempSigTpul; 
    FloatArray tempEpsTul;  
    FloatArray tempEpsTulForEpsTp;

    FloatArray tempEpsCmin; 
    FloatArray tempSigCmin; 
    FloatArray tempEpsCprl; 
    FloatArray tempSigCprl; 
    FloatArray tempEpsCpul; 
    FloatArray tempSigCpul; 
    FloatArray tempEpsCul;  
    FloatArray tempEpsCulForEpsCp;

    FloatArray tempPrincipalStrain,tempPrincipalStress;  

    // state determining parametres
    IntArray tempActiveState;
    IntArray tempTCrackClosing;
    IntArray tempCCrackClosing;

    FloatMatrix tempTangentStiffMat;

    FloatArray tempDamage, tempTangent;
    FloatArray tempElemSize;
    double tempAlpha;

  public:
    /// Constructor
    SHCCMaterialStatus(int n, Domain *d, GaussPoint *g);
    /// Destructor
    virtual ~SHCCMaterialStatus();

    void printOutputAt(FILE *file, TimeStep *tStep);

    // definition
    virtual const char *giveClassName() const { return "SHCCMaterialStatus"; }

    virtual void initTempStatus();
    virtual void updateYourself(TimeStep *tStep);

    contextIOResultType saveContext(FILE *stream, void *obj = NULL);
    contextIOResultType restoreContext(FILE *stream, void *obj = NULL);

    FloatArray &giveEpsTmax() { return epsTmax; }
    FloatArray &giveSigTmax() { return sigTmax; }
    FloatArray &giveEpsTprl() { return epsTprl; }
    FloatArray &giveSigTprl() { return sigTprl; }
    FloatArray &giveEpsTpul() { return epsTpul; }
    FloatArray &giveSigTpul() { return sigTpul; }
    FloatArray &giveEpsTul()  { return epsTul;  }
    FloatArray &giveEpsTulForEpsTp()  { return epsTulForEpsTp;  }

    FloatArray &giveEpsCmin() { return epsCmin; }
    FloatArray &giveSigCmin() { return sigCmin; }
    FloatArray &giveEpsCprl() { return epsCprl; }
    FloatArray &giveSigCprl() { return sigCprl; }
    FloatArray &giveEpsCpul() { return epsCpul; }
    FloatArray &giveSigCpul() { return sigCpul; }
    FloatArray &giveEpsCul()  { return epsCul;  }
    FloatArray &giveEpsCulForEpsCp()  { return epsCulForEpsCp;  }

    FloatArray &givePrincipalStrain()  { return principalStrain;  }
    FloatArray &givePrincipalStress()  { return principalStress;  }

    IntArray &giveActiveState     () { return activeState;      }
    IntArray &giveTCrackClosing   () { return tCrackClosing;    }
    IntArray &giveCCrackClosing   () { return cCrackClosing;    }

    FloatMatrix &giveTangentStiffMat() { return tangentStiffMat; }

    FloatArray &giveDamage() { return damage; }
    FloatArray &giveTangent() { return tangent; }
    FloatArray &giveElemSize() { return elemSize; }
    double giveAlpha() { return alpha; }

    // temp variables
    FloatArray &giveTempEpsTmax() { return tempEpsTmax; }
    FloatArray &giveTempSigTmax() { return tempSigTmax; }
    FloatArray &giveTempEpsTprl() { return tempEpsTprl; }
    FloatArray &giveTempSigTprl() { return tempSigTprl; }
    FloatArray &giveTempEpsTpul() { return tempEpsTpul; }
    FloatArray &giveTempSigTpul() { return tempSigTpul; }
    FloatArray &giveTempEpsTul()  { return tempEpsTul;  }
    FloatArray &giveTempEpsTulForEpsTp()  { return tempEpsTulForEpsTp;  }

    FloatArray &giveTempEpsCmin() { return tempEpsCmin; }
    FloatArray &giveTempSigCmin() { return tempSigCmin; }
    FloatArray &giveTempEpsCprl() { return tempEpsCprl; }
    FloatArray &giveTempSigCprl() { return tempSigCprl; }
    FloatArray &giveTempEpsCpul() { return tempEpsCpul; }
    FloatArray &giveTempSigCpul() { return tempSigCpul; }
    FloatArray &giveTempEpsCul()  { return tempEpsCul;  }
    FloatArray &giveTempEpsCulForEpsCp()  { return tempEpsCulForEpsCp;  }

    FloatArray &giveTempPrincipalStrain()  { return tempPrincipalStrain;  }
    FloatArray &giveTempPrincipalStress()  { return tempPrincipalStress;  }

    IntArray &giveTempActiveState     () { return tempActiveState;      }
    IntArray &giveTempTCrackClosing   () { return tempTCrackClosing;    }
    IntArray &giveTempCCrackClosing   () { return tempCCrackClosing;    }

    FloatMatrix &giveTempTangentStiffMat() { return tempTangentStiffMat; }

    FloatArray &giveTempDamage() { return tempDamage; }
    FloatArray &giveTempTangent() { return tempTangent; }
    FloatArray &giveTempElemSize() { return tempElemSize; }
    double giveTempAlpha() { return tempAlpha; }

    void letEpsTmax(FloatArray &values) { epsTmax = values; }
    void letSigTmax(FloatArray &values) { sigTmax = values; }
    void letEpsTprl(FloatArray &values) { epsTprl = values; }
    void letSigTprl(FloatArray &values) { sigTprl = values; }
    void letEpsTpul(FloatArray &values) { epsTpul = values; }
    void letSigTpul(FloatArray &values) { sigTpul = values; }
    void letEpsTul (FloatArray &values) { epsTul  = values; }
    void letEpsTulForEpsTp(FloatArray &values)  { epsTulForEpsTp = values; }

    void letEpsCmin(FloatArray &values) { epsCmin = values; }
    void letSigCmin(FloatArray &values) { sigCmin = values; }
    void letEpsCprl(FloatArray &values) { epsCprl = values; }
    void letSigCprl(FloatArray &values) { sigCprl = values; }
    void letEpsCpul(FloatArray &values) { epsCpul = values; }
    void letSigCpul(FloatArray &values) { sigCpul = values; }
    void letEpsCul (FloatArray &values) { epsCul  = values; }
    void letEpsCulForEpsCp(FloatArray &values)  { epsCulForEpsCp = values; }

    void letPrincipalStrain(FloatArray &values) { principalStrain  = values; }
    void letPrincipalStress(FloatArray &values) { principalStress  = values; }

    void letActiveState     (IntArray &values) { activeState      = values; }
    void letTCrackClosing   (IntArray &values) { tCrackClosing    = values; }
    void letCCrackClosing   (IntArray &values) { cCrackClosing    = values; }

    void letTangentStiffMat(FloatMatrix &values) { tangentStiffMat = values; }

    void letDamage(FloatArray &values) { damage = values; }
    void letTangent(FloatArray &values) { tangent = values; }
    void letElemSize(FloatArray &values) { elemSize = values; }

    void letAlpha(double &values) { alpha = values; }

    void letTempEpsTmax(FloatArray &values) { tempEpsTmax = values; }
    void letTempSigTmax(FloatArray &values) { tempSigTmax = values; }
    void letTempEpsTprl(FloatArray &values) { tempEpsTprl = values; }
    void letTempSigTprl(FloatArray &values) { tempSigTprl = values; }
    void letTempEpsTpul(FloatArray &values) { tempEpsTpul = values; }
    void letTempSigTpul(FloatArray &values) { tempSigTpul = values; }
    void letTempEpsTul (FloatArray &values) { tempEpsTul  = values; }
    void letTempEpsTulForEpsTp(FloatArray &values)  { tempEpsTulForEpsTp = values; }

    void letTempEpsTmaxAt(double val, int dim) { tempEpsTmax.at(dim) = val; }
    void letTempSigTmaxAt(double val, int dim) { tempSigTmax.at(dim) = val; }
    void letTempEpsTprlAt(double val, int dim) { tempEpsTprl.at(dim) = val; }
    void letTempSigTprlAt(double val, int dim) { tempSigTprl.at(dim) = val; }
    void letTempEpsTpulAt(double val, int dim) { tempEpsTpul.at(dim) = val; }
    void letTempSigTpulAt(double val, int dim) { tempSigTpul.at(dim) = val; }
    void letTempEpsTulAt (double val, int dim) { tempEpsTul.at(dim)  = val; }
    void letTempEpsTulForEpsTpAt(double val, int dim)  { tempEpsTulForEpsTp.at(dim) = val; }


    void letTempEpsCmin(FloatArray &values) { tempEpsCmin = values; }
    void letTempSigCmin(FloatArray &values) { tempSigCmin = values; }
    void letTempEpsCprl(FloatArray &values) { tempEpsCprl = values; }
    void letTempSigCprl(FloatArray &values) { tempSigCprl = values; }
    void letTempEpsCpul(FloatArray &values) { tempEpsCpul = values; }
    void letTempSigCpul(FloatArray &values) { tempSigCpul = values; }
    void letTempEpsCul (FloatArray &values) { tempEpsCul  = values; }
    void letTempEpsCulForEpsCp(FloatArray &values)  { tempEpsCulForEpsCp = values; }

    void letTempEpsCminAt(double val, int dim) { tempEpsCmin.at(dim) = val; }
    void letTempSigCminAt(double val, int dim) { tempSigCmin.at(dim) = val; }
    void letTempEpsCprlAt(double val, int dim) { tempEpsCprl.at(dim) = val; }
    void letTempSigCprlAt(double val, int dim) { tempSigCprl.at(dim) = val; }
    void letTempEpsCpulAt(double val, int dim) { tempEpsCpul.at(dim) = val; }
    void letTempSigCpulAt(double val, int dim) { tempSigCpul.at(dim) = val; }
    void letTempEpsCulAt (double val, int dim) { tempEpsCul.at(dim)  = val; }
    void letTempEpsCulForEpsCpAt(double val, int dim)  { tempEpsCulForEpsCp.at(dim) = val; }

    void letTempPrincipalStrain(FloatArray &values) { tempPrincipalStrain  = values; }
    void letTempPrincipalStress(FloatArray &values) { tempPrincipalStress  = values; }

    void letTempActiveState     (IntArray &values) { tempActiveState      = values; }
    void letTempTCrackClosing   (IntArray &values) { tempTCrackClosing    = values; }
    void letTempCCrackClosing   (IntArray &values) { tempCCrackClosing    = values; }

    void letTempActiveStateAt     (int val, int dim) { tempActiveState.at(dim)      = val; }
    void letTempTCrackClosingAt   (int val, int dim) { tempTCrackClosing.at(dim)    = val; }
    void letTempCCrackClosingAt   (int val, int dim) { tempCCrackClosing.at(dim)    = val; }

    void letTempTangentStiffMat(FloatMatrix &values) { tempTangentStiffMat = values; }

    void letTempDamageAt(double val, int dim) { tempDamage.at(dim) = val; }
    void letTempDamage(FloatArray &values) { tempDamage = values; }

    void letTempTangent(FloatArray &values) { tempTangent = values; }
    void letTempElemSize(FloatArray &values) { tempElemSize = values; }
    void letTempElemSizeAt(double val, int dim) { tempElemSize.at(dim) = val; }
    void letTempAlpha(double &values) { tempAlpha = values; }

  };

  /**
  * This class implements a SHCC material in a finite element problem.
  *
  * SHCC is NonLinear elasto-plastic material with damage model of SHCC
  * Cracking is described using Rotating Crack Model based maximum crack opening criterion.
  */
  class SHCCMaterial : public StructuralMaterial
  {


  protected:
    enum SofteningType { ST_Unknown, ST_Linear, ST_Exponential_Cohesive_Crack, ST_Hermit, ST_Polynomial_Cohesive_Crack };
    enum CrushingType { CT_Unknown, CT_Hermit, CT_Thorenfeldt_Cohesive_Crack, CT_Exponential_Cohesive_Crack };

    /// Parameter specifying the type of softening.
    SofteningType softType;
    /// Parameter specifying the type of crushing.
    CrushingType crushType;

    /// Reference to the basic elastic material.
    LinearElasticMaterial *linearElasticMaterial;

    double E;
    double nu;    
    double epsT0;
    double epsTp;      
    double sigTp;      
    double wT;         
    double aT;         
    double bT;         
    double bT_closing; 
    double epsC0;      
    double epsCp;      
    double sigCp;      
    double epsCu_input;
    double dC;         
    double aC;         
    double bC;         
    double bC_closing;
    double betaC;
    int    dimension;

    FloatArray epsTu, epsCu; // They are length 0 if they are not needed

    double wC, Le; // Le = element size
    bool changeState;

    FloatArray dDamage;
    FloatArray princStrain, dprincStrain;
    double epsTmax_calc, sigTmax_calc, epsTul_calc, sigTul_calc;
    double epsCmin_calc, sigCmin_calc, epsCul_calc, sigCul_calc;

    enum SHCC_StressState  { noneSS = 0, tensionSS = 1, compressionSS = 2 };

    /// Method used for evaluation of characteristic element size
    ElementCharSizeMethod ecsMethod;

    void initElementSize(int direction, GaussPoint *gp, SHCCMaterialStatus *status);
    void initParametersSoftening(int direction, GaussPoint *gp, SHCCMaterialStatus *status);
    void initParametersCrushing(int direction, GaussPoint *gp, SHCCMaterialStatus *status);

    double giveCrackingModulus(MatResponseMode rMode, GaussPoint *gp, int i);

  public:
    /// Constructor
    SHCCMaterial(int n, Domain *d);
    /// Destructor
    virtual ~SHCCMaterial();


    // identification and auxiliary functions
    virtual const char *giveInputRecordName() const { return _IFT_SHCCMaterial_Name; }  
    virtual const char *giveClassName() const { return "SHCCMaterial"; }
    virtual IRResultType initializeFrom(InputRecord *ir);

    virtual void give3dMaterialStiffnessMatrix(FloatMatrix &answer,
                                               MatResponseMode mode,
                                               GaussPoint *gp,
                                               TimeStep *tStep);
    virtual void givePlaneStressStiffMtrx(FloatMatrix &answer,
                                          MatResponseMode mode,
                                          GaussPoint *gp,
                                          TimeStep *atTime);

    virtual void giveRealStressVector(FloatArray &answer, GaussPoint *gp,
                                      const FloatArray &reducedStrain, TimeStep *tStep);

    virtual void giveRealStressVector_3d(FloatArray &answer, GaussPoint *gp, const FloatArray &reducedE, TimeStep *tStep)
    { OOFEM_ERROR("SHCCmat :: 3D stress is not implemented"); }
    virtual void giveRealStressVector_PlaneStrain(FloatArray &answer, GaussPoint *gp, const FloatArray &reducedE, TimeStep *tStep)
    { OOFEM_ERROR("SHCCmat :: Plane strain is not implemented"); }
    virtual void giveRealStressVector_PlaneStress(FloatArray &answer, GaussPoint *gp, const FloatArray &reducedE, TimeStep *tStep)
    { this->giveRealStressVector(answer, gp, reducedE, tStep); }
    virtual void giveRealStressVector_1d(FloatArray &answer, GaussPoint *gp, const FloatArray &reducedE, TimeStep *tStep)
    { OOFEM_ERROR("SHCCmat :: 1D proble is not implemented"); }


    virtual void  giveCharacteristicMatrix(FloatMatrix &answer, MatResponseMode mode,
                                           GaussPoint *gp, TimeStep *atTime);
    

    virtual int hasMaterialModeCapability(MaterialMode);
    virtual MaterialStatus *CreateStatus(GaussPoint *gp) const;
    virtual MaterialStatus *giveStatus(GaussPoint *gp) const;

    virtual bool isCharacteristicMtrxSymmetric(MatResponseMode rMode) { return false; }

    // Returns a reference to the basic elastic material.
    LinearElasticMaterial *giveLinearElasticMaterial() { return linearElasticMaterial; }

    virtual int giveIPValue(FloatArray &answer, GaussPoint *gp, InternalStateType type, TimeStep *tStep);
  protected:
    // general functions
    virtual void setTmaxTul_calc( bool unloadingState, int direction, SHCCMaterialStatus *status );
    virtual void setCminCul_calc( bool unloadingState, int direction, SHCCMaterialStatus *status );
    virtual void setTmaxCmin_calc( int direction, SHCCMaterialStatus *status );
    virtual double giveShiftedStrain( bool tensileLoad, int direction, SHCCMaterialStatus *status );
    virtual double giveClosingStrain( bool tensileLoad, int direction, SHCCMaterialStatus *status );
    virtual double giveDamage( bool tensileLoad, int direction, SHCCMaterialStatus *status );
    virtual bool isDamageGrowth( bool tensileLoad, int direction, SHCCMaterialStatus *status );
    virtual double giveEpsTul( double eps, double sig, int direction, SHCCMaterialStatus *status );
    virtual double giveEpsCul( double eps, double sig, int direction, SHCCMaterialStatus *status );


    FloatMatrix giveRotationMatrix( double angle );
    double give2DRotationalAngleStrain( const FloatArray &strain );
    FloatMatrix giveTransformationMatrixStrain( double angle );
    FloatMatrix giveTransformationMatrixStress( double angle );

    // stress evaluation
    FloatArray giveTensileStress( int direction, GaussPoint *gp, SHCCMaterialStatus *status );
    FloatArray giveTensionAfterCompression( int direction, GaussPoint *gp, SHCCMaterialStatus *status );
    FloatArray giveTensionOrdinaryLoading( int direction, GaussPoint *gp, SHCCMaterialStatus *status );
    FloatArray giveCompressiveStress( int direction, GaussPoint *gp, SHCCMaterialStatus *status );
    FloatArray giveCompressionAfterTension( int direction, GaussPoint *gp, SHCCMaterialStatus *status );
    FloatArray giveCompressionOrdinaryLoading( int direction, GaussPoint *gp, SHCCMaterialStatus *status );

    FloatMatrix &giveTangetialStiffMatrix( FloatArray &principalStress, FloatArray &tangent, double rotAngle, SHCCMaterialStatus *status );


  };
} // end namespace oofem
#endif // shcc_h
