/*
*
*                 #####    #####   ######  ######  ###   ###
*               ##   ##  ##   ##  ##      ##      ## ### ##
*              ##   ##  ##   ##  ####    ####    ##  #  ##
*             ##   ##  ##   ##  ##      ##      ##     ##
*            ##   ##  ##   ##  ##      ##      ##     ##
*            #####    #####   ##      ######  ##     ##
*
*
*             OOFEM : Object Oriented Finite Element Code
*
*               Copyright (C) 1993 - 2011   Borek Patzak
*
*
*
*       Czech Technical University, Faculty of Civil Engineering,
*   Department of Structural Mechanics, 166 29 Prague, Czech Republic
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "shccmatsla.h"
#include "floatarray.h"
#include "floatmatrix.h"
#include "structuralms.h"
#include "gausspoint.h"
#include "intarray.h"
#include "datastream.h"
#include "contextioerr.h"
#include "timestep.h"
#include "structuralmaterial.h"
#include "isolinearelasticmaterial.h"
#include "structuralcrosssection.h"
#include "mathfem.h"
#include "classfactory.h"


namespace oofem {
REGISTER_Material(SHCCMaterialSla);

// constructor
SHCCMaterialSla :: SHCCMaterialSla(int n, Domain *d) : SHCCMaterial(n, d), SlaMaterialExtensionInterface()
{
}


// destructor
SHCCMaterialSla :: ~SHCCMaterialSla()
{
}


void
SHCCMaterialSla :: givePlaneStressStiffMtrx(FloatMatrix &answer, MatResponseMode mode, GaussPoint *gp, TimeStep *atTime)
{
  SHCCMaterialStatusSla *status = ( SHCCMaterialStatusSla * ) this->giveStatus(gp);

  // start from the elastic stiffness - elastic material, no damage
  if( status->giveTempActiveState().at(1) == SHCCMaterialStatusSla::LS_none &&
      status->giveTempActiveState().at(2) == SHCCMaterialStatusSla::LS_none ) {
    this->giveLinearElasticMaterial()->givePlaneStressStiffMtrx(answer, mode, gp, atTime);
    return;
  }

  this->givePlaneStressStiffMtrxPrincDir(answer,MatResponseMode::SecantStiffness,gp,atTime);

  FloatMatrix Tstrain(3,3), TstrainT(3,3), dum(3,3);
  Tstrain = giveTransformationMatrixStrain(status->giveTempAlpha());
  TstrainT.beTranspositionOf(Tstrain);

  dum.beProductOf(TstrainT,answer);
  answer.beProductOf(dum,Tstrain);
}


void
SHCCMaterialSla :: givePlaneStressStiffMtrxPrincDir(FloatMatrix &answer, MatResponseMode mode, GaussPoint *gp, TimeStep *atTime)
{
  SHCCMaterialStatusSla *status = ( SHCCMaterialStatusSla * ) this->giveStatus(gp);

  // start from the elastic stiffness - elastic material, no damage
  if( status->giveTempActiveState().at(1) == SHCCMaterialStatusSla::LS_none &&
      status->giveTempActiveState().at(2) == SHCCMaterialStatusSla::LS_none ) {
    this->giveLinearElasticMaterial()->givePlaneStressStiffMtrx(answer, mode, gp, atTime);
    return;
  }

  // already damaged material
  double E1 = status->giveCurrentLSStiffnessRatio(1) * E;
  double E2 = status->giveCurrentLSStiffnessRatio(2) * E;
  double nuc = giveNu(status);
  double nu1 = nuc*(1.- status->giveTempDamageAt(1));
  double nu2 = nuc*(1.- status->giveTempDamageAt(2));

  answer.resize(3,3);
  answer.zero();
  answer.at(1,1) = E1;
  answer.at(1,2) = nu1*E2;
  answer.at(2,1) = nu2*E1;
  answer.at(2,2) = E2;
  answer.times(1./(1.-nu1*nu2));
  answer.at(3,3) = min(E1,E2)/(2.0*(1.0+min(nu1,nu2)));
  //answer.at(3,3) = sqrt(E1*E2)/(2.0*(1.0+sqrt(nu1*nu2)));
}


void
SHCCMaterialSla :: giveRealStressVector(FloatArray &answer, GaussPoint *gp,
                                        const FloatArray &totalStrain, TimeStep *tStep)
{
  SHCCMaterialStatusSla *status = static_cast< SHCCMaterialStatusSla * >( this->giveStatus(gp) );
  FloatMatrix stiffMat;
  int rsize = StructuralMaterial :: giveSizeOfVoigtSymVector( gp->giveMaterialMode() );
  FloatArray reducedTotalStrainVector(rsize), princStrain(rsize),  dPStrainVec(rsize), pStressVec(rsize);

  // subtract stress independent part
  // note: eigenStrains (temperature) is not contained in mechanical strain stored in gp
  // therefore it is necessary to subtract always the total eigen strain value
  this->giveStressDependentPartOfStrainVector(reducedTotalStrainVector, gp, totalStrain, tStep, VM_Total);


  FloatMatrix Tstrain(3,3), TstrainT(3,3);
  Tstrain = giveTransformationMatrixStrain(status->giveTempAlpha());
  TstrainT.beTranspositionOf(Tstrain);

  princStrain.beProductOf(Tstrain,reducedTotalStrainVector);
  dPStrainVec.beDifferenceOf(princStrain, status->giveTempPrincipalStrain());

  // stress increment + previous stress
  this->givePlaneStressStiffMtrxPrincDir(stiffMat,MatResponseMode::SecantStiffness,gp,tStep);
  pStressVec.beProductOf(stiffMat,dPStrainVec);
  pStressVec.add(status->giveTempPrincipalStress());

  // total stress - previous stress calculated with the same damage - can be simply added
  answer.beProductOf(TstrainT, pStressVec);

}


IRResultType
SHCCMaterialSla :: initializeFrom(InputRecord *ir)
{
  SHCCMaterial :: initializeFrom(ir);

  if( SHCCsla_simplifiedCompression ) {
    epsC0 = sigCp / E; // compresion has plateau insted of hardening part
  }

  SlaMaterialExtensionInterface :: initializeFrom(ir);

  if( !SHCCsla_clossingPart ) {
    bT = bT*(1.-bT_closing);
    bC = bC*(1.-bC_closing);

    bT_closing = bC_closing = 0.0;
  }

  return IRRT_OK;
}


double
SHCCMaterialSla :: computeLambda(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep, double minLambda)
{

  // lambda is between 0 and DBL_MAX, but because we load with full unscaled bc => max value of lambda used = 1
  SHCCMaterialStatusSla *status = static_cast< SHCCMaterialStatusSla * >( this->giveStatus(gp) );
  int rsize = StructuralMaterial :: giveSizeOfVoigtSymVector( gp->giveMaterialMode() );
  FloatArray psigVecOld(rsize), dPStressVec(rsize);
  int loadState;
  IntArray bounds(2);
  double lamMin, lam, dumLam, strLim;


  dPStressVec = givePrincStressIncrement(gp,totalStrain,tStep);

  if( dPStressVec.computeNorm() < SHCC_ZERO ) {
    return INFINITY;
  }

  this->initSLA(gp,totalStrain, tStep);

  psigVecOld = status->givePrincStressLastIter();

  lamMin = INFINITY;
  for( int i=1; i<=dimension; i++ ) {
    if( fabs(dPStressVec.at(i)) < SHCC_ZERO ) {
      continue;
    }

    loadState = status->giveTempActiveState().at(i);
    bounds = status->giveBoundsForLS(loadState);

    // at least one lambda in this run should be positive or equal to zero
    lam = INFINITY;
    for( int j=1; j<=2; j++) { // check for upper and lower limit
      strLim = status->giveStrengthAt(status->giveStrengthIndex(bounds.at(j),i));

      if( (bounds.at(j) == SHCCMaterialStatusSla::bound_unloadComp ||
           bounds.at(j) == SHCCMaterialStatusSla::bound_unloadTen) &&
          status->giveLSChange().at(i) && fabs(strLim-psigVecOld.at(i)) < SHCC_ZERO ) {
        continue;
      }

      dumLam = (strLim-psigVecOld.at(i))/dPStressVec.at(i);

      if( dumLam < 0.0 ) {
        continue;
      } else if( dumLam < lam ) {
        lam = dumLam;
      }
    }


    if( lam == INFINITY ) {
      lamMin = 0.0;
    } else if( lam < lamMin ) {
      lamMin = lam;
    }
  }

  return lamMin;
}


void
SHCCMaterialSla :: initSLA(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep)
{
  SHCCMaterialStatusSla *status = static_cast< SHCCMaterialStatusSla * >( this->giveStatus(gp) );

  if( status->giveStrengthAt(1) >= 0. ) {
    return;
  }

  const double ET2 = (sigTp-E*epsT0)/(epsTp-epsT0);
  const double EC2 = (sigCp-E*epsC0)/(epsCp-epsC0);

  double sigT = E * ((1+this->giveStrengthRange())*E*epsT0-ET2*epsT0)/(E-ET2);
  double sigC = E * ((1+this->giveStrengthRange())*E*epsC0-EC2*epsC0)/(E-EC2);

  FloatArray strengthLim(status->giveStrength().giveSize());
  for( int i = 1; i<=dimension; i++ ) {
    strengthLim.at(status->giveStrengthIndex(SHCCMaterialStatusSla::bound_tension, i)) = sigT;
    strengthLim.at(status->giveStrengthIndex(SHCCMaterialStatusSla::bound_compression, i)) = sigC;

    strengthLim.at(status->giveStrengthIndex(SHCCMaterialStatusSla::bound_unloadTen, i)) = sigC;
    strengthLim.at(status->giveStrengthIndex(SHCCMaterialStatusSla::bound_unloadComp, i)) = sigT;


    status->letTempEpsTmaxAt(sigT/E, i);
    status->letTempSigTmaxAt(sigT, i);
    status->letTempEpsCminAt(sigC/E, i);
    status->letTempSigCminAt(sigC, i);
  }

  status->letStrengthBe(strengthLim);

  status->updateYourself(tStep);

}


void
SHCCMaterialSla :: updateHistoryParameters(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep, bool activeDamage)
{

  SHCCMaterialStatusSla *status = static_cast< SHCCMaterialStatusSla * >( this->giveStatus(gp) );
  FloatArray stress;
  int rsize = StructuralMaterial :: giveSizeOfVoigtSymVector( gp->giveMaterialMode() );
  FloatArray epsVec(rsize), epsVecOld(rsize), depsVec(rsize), dumVec(rsize), pStressLastIter(rsize);
  FloatMatrix Tstrain(rsize,rsize);
  double alpha;

  pStressLastIter = status->givePrincStressLastIter();
  pStressLastIter.add(givePrincStressIncrement(gp,totalStrain,tStep));

  this->giveStressDependentPartOfStrainVector(epsVec, gp, totalStrain, tStep, VM_Total);
  this->giveStressDependentPartOfStrainVector(epsVecOld, gp, status->giveTempStrainVector(), tStep->givePreviousStep(), VM_Total);
  depsVec.beDifferenceOf(epsVec,epsVecOld);

  this->giveRealStressVector(stress,gp,totalStrain,tStep);

  status->letTempStrainVectorBe(totalStrain);
  status->letTempStressVectorBe(stress);
  status->letPrincStressLastIter(pStressLastIter);

  initSLA(gp,totalStrain,tStep);

  // checking of stress change
  if( status->isLSChange() ) {
    if( status->giveTempActiveState().at(1) == SHCCMaterialStatusSla::LS_none &&
        status->giveTempActiveState().at(2) == SHCCMaterialStatusSla::LS_none ) {
      alpha = give2DRotationalAngleStrain(epsVec);
    } else {
      alpha = status->giveTempAlpha();
    }
    Tstrain = giveTransformationMatrixStrain(alpha);
    dumVec.beProductOf(Tstrain,depsVec);
    for( int i = 1; i<=dimension; i++ ) {
      if( fabs(dumVec.at(i)) > SHCC_ZERO ) {
        status->letLSChangeAtBe(0,i);
      }
    }
  }

  if( !activeDamage ) {
    status->updateYourself(tStep);
    return;
  }

  // the stress drop is calculated in the new principal direction=> update of princ directions, alpha
  FloatArray stressAfterDropVec(rsize), dStressVec(rsize);
  FloatArray pStressVec(rsize);
  FloatMatrix TstrainT(rsize,rsize);

  alpha = give2DRotationalAngleStrain(epsVec);
  status->letTempAlpha(alpha);
  Tstrain = giveTransformationMatrixStrain(alpha);
  TstrainT.beTranspositionOf(Tstrain);

  princStrain.beProductOf(Tstrain,epsVec);
  dprincStrain.beDifferenceOf(princStrain,status->giveTempPrincipalStrain());

  // all directions have to be upgrade
  for( int i=1; i<=dimension; i++ ) {

    int loadState = status->giveTempActiveState().at(i);
    FloatArray strength = status->giveStrengthInDir(i);

    // initial loading = elasticity - the loading state is defined
    if( loadState == SHCCMaterialStatusSla::LS_none ) {
      if( princStrain.at(i) >= strength.at(SHCCMaterialStatusSla::bound_tension)/E - SHCC_ZERO ) {
        loadState = SHCCMaterialStatusSla::LS_tension;
      } else if( princStrain.at(i) <= strength.at(SHCCMaterialStatusSla::bound_compression)/E + SHCC_ZERO ){
        loadState = SHCCMaterialStatusSla::LS_compression;
      } else {
        loadState = SHCCMaterialStatusSla::LS_none;
      }

      if( loadState != SHCCMaterialStatusSla::LS_none ) {
        status->letTempActiveStateAt(loadState,i);
        if( SHCCsla_clossingPart ) {
          status->letStrengthAtBe(giveMinStressForCrackClosing(), status->giveStrengthIndex(SHCCMaterialStatusSla::bound_unloadTen,i));
          status->letStrengthAtBe(-1.0*giveMinStressForCrackClosing(), status->giveStrengthIndex(SHCCMaterialStatusSla::bound_unloadComp,i));
        } else {
          status->letStrengthAtBe(0.0, status->giveStrengthIndex(SHCCMaterialStatusSla::bound_unloadTen,i));
          status->letStrengthAtBe(0.0, status->giveStrengthIndex(SHCCMaterialStatusSla::bound_unloadComp,i));
        }
      }
    }


    switch( loadState ) {
    case SHCCMaterialStatusSla::LS_tension: {
      loadStateTension(pStressVec, i, gp, status);
      break;
    }
    case SHCCMaterialStatusSla::LS_compression: {
      loadStateCompression(pStressVec, i, gp, status);
      break;
    }
    case SHCCMaterialStatusSla::LS_crackClosing: {
      if( !SHCCsla_clossingPart ) {
        OOFEM_ERROR( "This should not happend" );
      }
      loadStateCrackClosing(pStressVec, i, gp, status);
      break;
    }
    case SHCCMaterialStatusSla::LS_none: {
      pStressVec.at(i) = E*princStrain.at(i);
      break;
    }
    default:
      OOFEM_ERROR( "I am not sure that this should be allowed" )
    }
  }

  double nuc = giveNu(status);
  double nu1 = nuc*(1.0 - status->giveTempDamageAt(1));
  double nu2 = nuc*(1.0 - status->giveTempDamageAt(2));

  stressAfterDropVec.zero();
  stressAfterDropVec.at(1) = (pStressVec.at(1) + nu1*pStressVec.at(2)) / (1.-nu1*nu2);
  stressAfterDropVec.at(2) = (pStressVec.at(2) + nu2*pStressVec.at(1)) / (1.-nu1*nu2);

  dStressVec.beProductOf(TstrainT,stressAfterDropVec);
  dStressVec.subtract(status->giveTempStressVector());

  status->letDStressVectorBe(dStressVec);
  status->letTempPrincipalStrain(princStrain);
  status->letTempPrincipalStress(pStressVec);
  status->letPrincStressLastIter(pStressVec);
  status->updateYourself(tStep);

}

void
SHCCMaterialSla :: loadStateTension(FloatArray &pStressVec, int dir, GaussPoint *gp, SHCCMaterialStatusSla *status)
{

  double eps = giveShiftedStrain( true, dir, status );
  double epsTul = status->giveTempEpsTul().at(dir);
  double Ed, sig, newStrength;
  FloatArray strength = status->giveStrengthInDir(dir);
  FloatArray stiffRatio = status->giveLSStiffnessRatioInDir(dir);

  status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_tension, dir);

  if( eps >= status->giveTempEpsTmax().at(dir) - SHCC_ZERO ) { // increase of damage

    newStrength = strength.at(SHCCMaterialStatusSla::bound_tension);
    while( eps >= status->giveTempEpsTmax().at(dir) - SHCC_ZERO ) {

      sig = newStrength - 2*this->giveStrengthRange()*E*epsT0;
      if( sig <= 0.0 ) {
        pStressVec.at(dir) = 0.0;

        status->letTempEpsTmaxAt(status->giveTempEpsTmax().at(dir), dir);
        status->letTempSigTmaxAt(0.0, dir);
        status->letTempDamageAt(SHCC_MAXDAMAGE, dir);
        status->letStrengthAtBe(INFINITY,status->giveStrengthIndex(SHCCMaterialStatusSla::LS_tension, dir));
        status->letLSStiffnessRatioAtBe(SHCC_ZERO,SHCCMaterialStatusSla::LS_tension, dir);
        status->letTempEpsTulAt(giveEpsTul(status->giveTempEpsTmax().at(dir),0.0,dir,status),dir);
        break;
      }

      newStrength = giveNewStrengthTension(dir, status->giveTempEpsTmax().at(dir),sig,gp,status);
    }

    pStressVec.at(dir) = status->giveTempSigTmax().at(dir) + (eps-status->giveTempEpsTmax().at(dir)) * E*status->giveCurrentLSStiffnessRatio(dir);
    status->letStrengthAtBe(newStrength,status->giveStrengthIndex(SHCCMaterialStatusSla::bound_tension, dir));

    // update of crack closing (reopening) part
    // strains designated for the transitional part
    double dum;

    // part of the tension and compression unloading
    dum = giveMinStressForCrackClosing()/(E*status->giveLSStiffnessRatioInDir(dir).at(SHCCMaterialStatusSla::LS_tension)) +
        giveMinStressForCrackClosing()/(E*stiffRatio.at(SHCCMaterialStatusSla::LS_compression));
    // part of the crack closing
    dum = dum + fabs( bT_closing * status->giveTempEpsTul().at(dir) );

    status->letLSStiffnessRatioAtBe( 2.0*giveMinStressForCrackClosing()/dum/E, SHCCMaterialStatusSla::LS_crackClosing, dir);


  } else { // no damage increase => partial loading/unloading or state change

    double dumEps1 = status->giveTempEpsTul().at(dir) +
        giveMinStressForCrackClosing() / (E*stiffRatio.at(SHCCMaterialStatusSla::LS_tension));
    double dumEps2 = 2.0*giveMinStressForCrackClosing() / (E*stiffRatio.at(SHCCMaterialStatusSla::LS_crackClosing));
    double dumEps3 = (status->giveTempEpsCul().at(dir) - status->giveTempEpsCmin().at(dir)) -
        giveMinStressForCrackClosing() / (E*stiffRatio.at(SHCCMaterialStatusSla::LS_compression));

    if( dumEps1 < eps ) { //No Load state change - still tension
      Ed = E * stiffRatio.at(SHCCMaterialStatusSla::LS_tension);
      pStressVec.at(dir) = (eps-epsTul) * Ed;

      if( pStressVec.at(dir)-giveMinStressForCrackClosing() < SHCC_ZERO ) {
        if( SHCCsla_clossingPart ) {
          status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_crackClosing, dir);
        } else {
          status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_compression, dir);
        }
        status->letLSChangeAtBe(1, dir);
      }

    } else if( SHCCsla_clossingPart && dumEps1 - eps < dumEps2 ) { // crack closing
      Ed = E * stiffRatio.at(SHCCMaterialStatusSla::LS_crackClosing);
      pStressVec.at(dir) = giveMinStressForCrackClosing() - (dumEps1 - eps) * Ed;
      status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_crackClosing, dir);
      status->letLSChangeAtBe(1,dir);

    } else if( dumEps1 - eps < dumEps2 + dumEps3 ) { // compression (reloding - no damage increase)
      Ed = E * stiffRatio.at(SHCCMaterialStatusSla::LS_compression);
      pStressVec.at(dir) = - giveMinStressForCrackClosing() - (dumEps1 - eps - dumEps2) * Ed;
      status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_compression, dir);
      status->letLSChangeAtBe(1, dir);

    } else { // increase of damage in compression
      loadStateCompression(pStressVec, dir, gp, status);
    }

  }
}


void
SHCCMaterialSla :: loadStateCompression(FloatArray &pStressVec, int dir, GaussPoint *gp, SHCCMaterialStatusSla *status)
{
  double eps = giveShiftedStrain( false, dir, status );
  double epsCul = status->giveTempEpsCul().at(dir);
  double Ed, sig, newStrength;
  FloatArray strength = status->giveStrengthInDir(dir);
  FloatArray stiffRatio = status->giveLSStiffnessRatioInDir(dir);

  status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_compression, dir);

  if( eps <= status->giveTempEpsCmin().at(dir) + SHCC_ZERO ) { // increase of damage

    newStrength = strength.at(SHCCMaterialStatusSla::bound_compression);
    while( eps <= status->giveTempEpsCmin().at(dir) + SHCC_ZERO ) {

      sig = newStrength - 2*this->giveStrengthRange()*E*epsC0;
      if( sig >= 0.0 ) {
        pStressVec.at(dir) = 0.0;

        status->letTempEpsCminAt(status->giveTempEpsCmin().at(dir), dir);
        status->letTempSigCminAt(0.0, dir);
        status->letTempDamageAt(SHCC_MAXDAMAGE, dir);
        status->letStrengthAtBe(INFINITY,status->giveStrengthIndex(SHCCMaterialStatusSla::LS_compression, dir));
        status->letLSStiffnessRatioAtBe(SHCC_ZERO,SHCCMaterialStatusSla::LS_compression, dir);
        status->letTempEpsCulAt(giveEpsCul(status->giveTempEpsCmin().at(dir),0.0,dir,status), dir);
        break;
      }

      newStrength = giveNewStrengthCompression(dir,status->giveTempEpsCmin().at(dir),sig,gp,status);
    }

    pStressVec.at(dir) = status->giveTempSigCmin().at(dir) + (eps-status->giveTempEpsCmin().at(dir)) * E*status->giveCurrentLSStiffnessRatio(dir);
    status->letStrengthAtBe(newStrength,status->giveStrengthIndex(SHCCMaterialStatusSla::bound_compression, dir));


    // update of crack closing (reopening) part
    // strains designated for the transitional part
    double dum;

    // part of the tension and compression unloading
    dum = giveMinStressForCrackClosing()/(E*stiffRatio.at(SHCCMaterialStatusSla::LS_tension)) +
        giveMinStressForCrackClosing()/(E*status->giveLSStiffnessRatioInDir(dir).at(SHCCMaterialStatusSla::LS_compression));
    // part of the crack closing
    dum = dum + fabs(bC_closing * status->giveTempEpsCul().at(dir));

    status->letLSStiffnessRatioAtBe( 2.0*giveMinStressForCrackClosing()/dum/E, SHCCMaterialStatusSla::LS_crackClosing, dir);


  } else { // no damage increase => partial loading/unloading or state change

    double dumEps1 = status->giveTempEpsCul().at(dir) -
        giveMinStressForCrackClosing() / (E*stiffRatio.at(SHCCMaterialStatusSla::LS_compression));
    double dumEps2 = -2.0*giveMinStressForCrackClosing() / (E*stiffRatio.at(SHCCMaterialStatusSla::LS_crackClosing));
    double dumEps3 = (status->giveTempEpsTul().at(dir) - status->giveTempEpsTmax().at(dir)) +
        giveMinStressForCrackClosing() / (E*stiffRatio.at(SHCCMaterialStatusSla::LS_tension));

    if( dumEps1 > eps ) { //No Load state change - still compression
      Ed = E * stiffRatio.at(SHCCMaterialStatusSla::LS_compression);
      pStressVec.at(dir) = (eps-epsCul) * Ed;

      if( pStressVec.at(dir)+giveMinStressForCrackClosing() > -SHCC_ZERO ) {
        if( SHCCsla_clossingPart ) {
          status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_crackClosing, dir);
        } else {
          status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_tension, dir);
        }
        status->letLSChangeAtBe(1, dir);
      }

    } else if( SHCCsla_clossingPart && dumEps1 - eps > dumEps2 ) { // crack closing
      Ed = E * stiffRatio.at(SHCCMaterialStatusSla::LS_crackClosing);
      pStressVec.at(dir) = -giveMinStressForCrackClosing() - (dumEps1 - eps) * Ed;
      status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_crackClosing, dir);
      status->letLSChangeAtBe(1, dir);

    } else if( dumEps1 - eps > dumEps2 + dumEps3 ) { // tension (reloding - no damage increase)
      Ed = E * stiffRatio.at(SHCCMaterialStatusSla::LS_tension);
      pStressVec.at(dir) = giveMinStressForCrackClosing() - (dumEps1 - eps - dumEps2) * Ed;
      status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_tension, dir);
      status->letLSChangeAtBe(1, dir);

    } else { // increase of damage in tension
      loadStateTension(pStressVec, dir, gp, status);
    }

  }
}


void
SHCCMaterialSla :: loadStateCrackClosing(FloatArray &pStressVec, int dir, GaussPoint *gp, SHCCMaterialStatusSla *status)
{
  // stress from previous equilibriated step
  double equilibriatedStress = status->giveTempPrincipalStress().at(dir);
  double Ed, sig;
  FloatArray strength = status->giveStrengthInDir(dir);
  FloatArray stiffRatio = status->giveLSStiffnessRatioInDir(dir);

  Ed = E * stiffRatio.at(SHCCMaterialStatusSla::LS_crackClosing);
  sig = equilibriatedStress + dprincStrain.at(dir) * Ed;

  status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_crackClosing, dir);

  if( sig > strength.at(SHCCMaterialStatusSla::bound_unloadComp) + SHCC_ZERO &&
      sig < strength.at(SHCCMaterialStatusSla::bound_unloadTen) - SHCC_ZERO ) { // no load state change

    pStressVec.at(dir) = sig;

    if( pStressVec.at(dir)-giveMinStressForCrackClosing() > -SHCC_ZERO ) {
      status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_tension, dir);
      status->letLSChangeAtBe(1, dir);
    } else if( pStressVec.at(dir)+giveMinStressForCrackClosing() < SHCC_ZERO ) {
      status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_compression, dir);
      status->letLSChangeAtBe(1, dir);
    }

  } else if( sig >= strength.at(SHCCMaterialStatusSla::bound_unloadTen) - SHCC_ZERO ) { // tension
    Ed = E * stiffRatio.at(SHCCMaterialStatusSla::LS_tension);

    sig = Ed * (giveShiftedStrain(true, dir, status) - status->giveTempEpsTul().at(dir));
    if( sig < strength.at(SHCCMaterialStatusSla::bound_tension) ) { // normal loading in tension
      pStressVec.at(dir) = sig;
      status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_tension, dir);
      status->letLSChangeAtBe(1, dir);

    } else {
      loadStateTension(pStressVec, dir, gp, status);
    }

    // degradation of crack clossing stiffness
    status->letLSStiffnessRatioAtBe( SHCCsla_minStressRatioDegadation*status->giveLSStiffnessRatioInDir(dir).at(SHCCMaterialStatusSla::LS_crackClosing),
                                     SHCCMaterialStatusSla::LS_crackClosing, dir );
    int idx = status->giveStrengthIndex(SHCCMaterialStatusSla::bound_unloadTen, dir);
    status->letStrengthAtBe( SHCCsla_minStressRatioDegadation*status->giveStrengthAt(idx), idx);
    idx = status->giveStrengthIndex(SHCCMaterialStatusSla::bound_unloadComp, dir);
    status->letStrengthAtBe( SHCCsla_minStressRatioDegadation*status->giveStrengthAt(idx), idx);

  } else { // compression
    Ed = E * stiffRatio.at(SHCCMaterialStatusSla::LS_compression);

    sig = Ed * (giveShiftedStrain(false, dir, status) - status->giveTempEpsCul().at(dir));
    if( sig > strength.at(SHCCMaterialStatusSla::bound_compression) ) { // normal loading in tension
      pStressVec.at(dir) = sig;
      status->letTempActiveStateAt(SHCCMaterialStatusSla::LS_compression, dir);
      status->letLSChangeAtBe(1, dir);

    } else {
      loadStateCompression(pStressVec, dir, gp, status);
    }

    // degradation of crack clossing stiffness
    status->letLSStiffnessRatioAtBe( SHCCsla_minStressRatioDegadation*status->giveLSStiffnessRatioInDir(dir).at(SHCCMaterialStatusSla::LS_crackClosing),
                                     SHCCMaterialStatusSla::LS_crackClosing, dir );
    int idx = status->giveStrengthIndex(SHCCMaterialStatusSla::bound_unloadTen, dir);
    status->letStrengthAtBe( SHCCsla_minStressRatioDegadation*status->giveStrengthAt(idx), idx);
    idx = status->giveStrengthIndex(SHCCMaterialStatusSla::bound_unloadComp, dir);
    status->letStrengthAtBe( SHCCsla_minStressRatioDegadation*status->giveStrengthAt(idx), idx);

  }
}

double
SHCCMaterialSla :: giveNewStrengthTension(int dir, double eps, double sig, GaussPoint *gp, SHCCMaterialStatusSla *status)
{
  double newStrength, epsTmax, unusedSig = 0.0;
  double epsTul = giveEpsTul(eps, unusedSig, dir, status);
  double Ed = sig/(eps-epsTul);

  bool softPart = true;
  if( eps <= epsTp ) { // elastic and hardening part
    const double ET2 = (sigTp-E*epsT0)/(epsTp-epsT0);
    epsTmax = ((1+this->giveStrengthRange())*E*epsT0-ET2*epsT0+Ed*epsTul)/(Ed-ET2);
    newStrength = Ed * (epsTmax-epsTul);

    if( epsTmax <= epsTp ) { // ensures the correct limits
      softPart = false;
    }
  }

  if( softPart ) { // softening part

    // fixing the element size - for the first time when it occures
    if( status->giveTempElemSize().at(dir) == 0.0 )
    {
      status->letTempEpsTulForEpsTpAt(epsTul, dir);
      status->letEdforEpsTpAtBe(Ed, dir);
      initParametersSoftening(dir,gp,status);
      status->letTempElemSizeAt(Le, dir);

      if( epsTu.at(1) <= epsTp ) {
        OOFEM_ERROR("SHCCMaterialSla: epsTu <= epsTp for element no. %d (%d), Le = %e!!!", gp->giveElement()->giveNumber(), gp->giveNumber(), Le );
      }

    } else {
      initParametersSoftening(dir,gp,status);
    }


    if( softType == ST_Hermit ) // Hermit function softening
    {
      double epsTrial = eps;
      double dum, sig1, sig2, R, Lhs;
      double sR = this->giveStrengthRange();
      int niter = 0;

      do {
        niter++;

        dum = (epsTrial-epsTp)/(epsTu.at(1)-epsTp);
        sig1 = Ed*(epsTrial-epsTul);
        sig2 = sigTp*(2.0 * pow(dum, 3.0) - 3.0 * pow(dum, 2.0) + 1.0) + sR*E*epsT0;

        R        = sig1 - sig2;
        Lhs      = Ed - sigTp*(6.0 * dum*dum - 6.0*dum)/(epsTu.at(1)-epsTp);
        epsTrial -= R / Lhs;
        if ( niter > 100 ) {
          OOFEM_ERROR("SHCC: computeStrengthForCohesiveCrack::algorithm not converging - can be caused by large element size");
        }
      } while ( niter == 0 || fabs(R) >= epsT0 * SHCC_ITERATION_LIMIT );

      newStrength = Ed * (epsTrial-epsTul);
      epsTmax = epsTrial;

    } else if( softType == ST_Linear ) { // linear softening
      const double ET3 =  -sigTp/(epsTu.at(1)-epsTp);
      epsTmax = ((1+this->giveStrengthRange())*sigTp-ET3*epsTp+Ed*epsTul)/(Ed-ET3);
      newStrength = Ed * (epsTmax-epsTul);

    }
    else if( softType == ST_Exponential_Cohesive_Crack ) { // exponential cohesive crack - iteration needed
      double epsTrial = eps;
      double dum, sig1, sig2, epsF, R, Lhs;
      double epsTuTp = status->giveTempEpsTulForEpsTp().at(dir);
      double EdTp = status->giveEdforEpsTp().at(dir);
      double sR = this->giveStrengthRange();
      int niter = 0;

      epsF = wT/status->giveTempElemSize().at(dir);
      do {
        niter++;

        sig1 = Ed*(epsTrial-epsTul);
        dum = epsTrial - epsTuTp - sig1/EdTp;
        sig2 = sigTp*exp(-dum/epsF) + sR*E*epsT0;

        R        = sig1 - sig2;
        Lhs      = Ed + sig2*(1.0-Ed/EdTp)/epsF;
        epsTrial -= R / Lhs;
        if ( niter > 100 ) {
          OOFEM_ERROR("SHCC: computeStrengthForCohesiveCrack::algorithm not converging - can be caused by large element size");
        }
      } while ( niter == 0 || fabs(R) >= epsT0 * SHCC_ITERATION_LIMIT );

      newStrength = Ed * (epsTrial-epsTul);
      epsTmax = epsTrial;

    } else if( softType == ST_Polynomial_Cohesive_Crack ) { // Nonlinear Tension Softening (Moelands and Reinhardt) - viz DIANA manual
      double epsTrial = eps;
      double dum, sig1, sig2, epsF, R, Lhs;
      double epsTuTp = status->giveTempEpsTulForEpsTp().at(dir);
      double EdTp = status->giveEdforEpsTp().at(dir);
      double sR = this->giveStrengthRange();
      int niter = 0;

      epsF = wT/status->giveTempElemSize().at(dir);
      do {
        niter++;

        sig1 = Ed*(epsTrial-epsTul);
        dum  = epsTrial - epsTuTp - sig1/EdTp;
        sig2 = sigTp * pow(1.0-dum/epsF,aT) + sR*E*epsT0;

        R        = sig1 - sig2;
        Lhs      = Ed + sigTp * aT * pow(1.0-dum/epsF,aT-1.0)*(1.0-Ed/EdTp)/epsF;
        epsTrial -= R / Lhs;
        if ( niter > 100 ) {
          OOFEM_ERROR("SHCC: computeStrengthForCohesiveCrack::algorithm not converging - can be caused by large element size");
        }
      } while ( niter == 0 || fabs(R) >= epsT0 * SHCC_ITERATION_LIMIT );

      newStrength = Ed * (epsTrial-epsTul);
      epsTmax = epsTrial;
    }
  }

  status->letTempSigTmaxAt(sig,dir); // change just for the damage calculation
  double dumDamage = giveDamage(true,dir,status); // uses previous tempEpsTmax and tempSigTmax

  status->letLSStiffnessRatioAtBe(Ed/E,SHCCMaterialStatusSla::LS_tension,dir);
  status->letTempEpsTmaxAt(epsTmax,dir);
  status->letTempSigTmaxAt(newStrength,dir);
  status->letTempEpsTulAt(epsTul,dir);
  status->letTempDamageAt(dumDamage, dir);

  return newStrength;
}


double
SHCCMaterialSla :: giveNewStrengthCompression(int dir, double eps, double sig, GaussPoint *gp, SHCCMaterialStatusSla *status)
{
  double newStrength, epsCmin, unusedSig = 0.0;
  double epsCul = giveEpsCul(eps, unusedSig, dir, status);
  double Ed = sig/(eps-epsCul);

  bool crushPart = true;
  if( eps >= epsCp ) { // elastic part and hardening

    if( SHCCsla_simplifiedCompression ) {
      newStrength = (1.0+this->giveStrengthRange()) * E * epsC0;
      epsCmin     = epsCul + newStrength/Ed;

    } else {
      double epsTrial = eps;
      double dum, sig1, sig2, R, Lhs;
      double sR = this->giveStrengthRange();
      int niter = 0;

      do {
        niter++;

        dum = epsCp - epsTrial;
        sig1 = Ed*(epsTrial-epsCul);
        sig2 = sigCp - (sigCp-epsC0*E) * pow(dum/(epsCp-epsC0), E*(epsCp-epsC0)/(sigCp-epsC0*E)) + sR*E*epsC0;

        R        = sig1 - sig2;
        Lhs      = Ed + E * pow(dum/(epsCp-epsC0), E*(epsCp-epsC0)/(sigCp-epsC0*E)-1.0);
        epsTrial -= R / Lhs;
        if ( niter > 100 ) {
          OOFEM_ERROR("SHCC: computeStrengthForCrushing::algorithm not converging - can be caused by large element size");
        }
      } while ( niter == 0 || fabs(R) >= fabs(epsC0) * SHCC_ITERATION_LIMIT );

      newStrength = Ed * (epsTrial-epsCul);
      epsCmin = epsTrial;
    }

    if( epsCmin >= epsCp ) { // ensures the correct limits
      crushPart = false;
    }

  }

  if( crushPart ) { // crushing part
    // fixing the element size - in the first occasion when epsCp is exceeded
    if( status->giveTempElemSize().at(dir) == 0.0 )
    {
      status->letTempEpsCulForEpsCpAt(epsCul, dir);
      status->letEdforEpsCpAtBe(Ed, dir);
      initParametersCrushing(dir,gp,status);
      status->letTempElemSizeAt(Le, dir);

    } else {
      initParametersCrushing(dir,gp,status);
    }


    if( crushType == CT_Hermit ) {
      double epsTrial = eps;
      double dum, sig1, sig2, R, Lhs;
      double sR = this->giveStrengthRange();
      int niter = 0;

      do {
        niter++;

        dum = (epsTrial-epsCp)/(epsCu(0)-epsCp);
        sig1 = Ed*(epsTrial-epsCul);
        sig2 = sigCp*(2.0 * pow(dum, 3.0) - 3.0 * pow(dum, 2.0) + 1.0) + sR*E*epsC0;

        R        = sig1 - sig2;
        Lhs      = Ed - sigCp*(6.0 * dum*dum - 6.0*dum)/(epsCu(0)-epsCp);
        epsTrial -= R / Lhs;
        if ( niter > 100 ) {
          OOFEM_ERROR("SHCC: computeStrengthForCrushing::algorithm not converging - can be caused by large element size");
        }
      } while ( niter == 0 || fabs(R) >= fabs(epsC0) * SHCC_ITERATION_LIMIT );

      newStrength = Ed * (epsTrial-epsCul);
      epsCmin = epsTrial;


    } else if( crushType == CT_Thorenfeldt_Cohesive_Crack ) { // Thorenfeldt based cohesive function for crushing ("overlapping") - iteration needed
      double epsTrial = eps;
      double dum, sig1, sig2, epsF, R, Lhs;
      double epsCuTp = status->giveTempEpsCulForEpsCp().at(dir);
      double EdCp = status->giveEdforEpsCp().at(dir);
      double sR = this->giveStrengthRange();
      int niter = 0;

      epsF = wC/status->giveTempElemSize().at(dir);
      do {
        niter++;

        sig1 = Ed*(epsTrial-epsCul);
        dum = epsTrial - epsCuTp - sig1/EdCp;
        sig2 = sigCp * (1+dum/epsF) * betaC/(betaC-1+pow(1+dum/epsF,betaC)) + sR*E*epsC0;

        R        = sig1 - sig2;
        Lhs      = Ed - sigCp*(1.0-Ed/EdCp)/epsF * betaC/(betaC-1+pow(1+dum/epsF,betaC)) +
                        sigCp * (1+dum/epsF) * betaC/pow(betaC-1+pow(1+dum/epsF,betaC),2.0) *
                        betaC*pow(1+dum/epsF,betaC-1.0) *(1.0-Ed/EdCp)/epsF;
        epsTrial -= R / Lhs;
        if ( niter > 100 ) {
          OOFEM_ERROR("SHCC: computeStrengthForCrushing::algorithm not converging - can be caused by large element size");
        }
      } while ( niter == 0 || fabs(R) >= fabs(epsC0) * SHCC_ITERATION_LIMIT );

      newStrength = Ed * (epsTrial-epsCul);
      epsCmin = epsTrial;

    } else if( crushType == CT_Exponential_Cohesive_Crack ) { // exponential cohesive function for crushing ("overlapping") - iteration needed
      double epsTrial = eps;
      double dum, sig1, sig2, epsF, R, Lhs;
      double epsCuTp = status->giveTempEpsCulForEpsCp().at(dir);
      double EdCp = status->giveEdforEpsCp().at(dir);
      double sR = this->giveStrengthRange();
      int niter = 0;

      epsF = wC/status->giveTempElemSize().at(dir);
      do {
        niter++;

        sig1 = Ed*(epsTrial-epsCul);
        dum = epsTrial - epsCuTp - sig1/EdCp;
        sig2 = sigCp*exp(-dum/epsF) + sR*E*epsC0;

        R        = sig1 - sig2;
        Lhs      = Ed + sig2*(1.0-Ed/EdCp)/epsF;
        epsTrial -= R / Lhs;
        if ( niter > 100 ) {
          OOFEM_ERROR("SHCC: computeStrengthForCrushing::algorithm not converging - can be caused by large element size");
        }
      } while ( niter == 0 || fabs(R) >= fabs(epsC0) * SHCC_ITERATION_LIMIT );

      newStrength = Ed * (epsTrial-epsCul);
      epsCmin = epsTrial;
    }
  }

  status->letTempSigCminAt(sig,dir); // change just for the damage calculation
  double dumDamage = giveDamage(false,dir,status); // uses previous tempEpsCmin and tempSigCmin

  status->letLSStiffnessRatioAtBe(Ed/E,SHCCMaterialStatusSla::LS_compression,dir);
  status->letTempEpsCminAt(epsCmin,dir);
  status->letTempSigCminAt(newStrength,dir);
  status->letTempEpsCulAt(epsCul,dir);
  status->letTempDamageAt(dumDamage, dir);

  return newStrength;
}


FloatArray
SHCCMaterialSla :: givePrincStressIncrement(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep)
{
  SHCCMaterialStatusSla *status = static_cast< SHCCMaterialStatusSla * >( this->giveStatus(gp) );
  int rsize = StructuralMaterial :: giveSizeOfVoigtSymVector( gp->giveMaterialMode() );
  FloatArray epsVec(rsize), epsVecOld(rsize), depsVec(rsize);

  this->giveStressDependentPartOfStrainVector(epsVec, gp, totalStrain, tStep, VM_Total);
  this->giveStressDependentPartOfStrainVector(epsVecOld, gp, status->giveTempStrainVector(), tStep->givePreviousStep(), VM_Total);
  depsVec.beDifferenceOf(epsVec,epsVecOld);
  FloatMatrix Tstrain, Tstress;

  // we can afford to calculate the increment of principal strain in this way,
  // because the saved value and the current one are rotated by the same angle
  double alpha;
  if( status->giveTempActiveState().at(1) == SHCCMaterialStatusSla::LS_none &&
      status->giveTempActiveState().at(2) == SHCCMaterialStatusSla::LS_none ) {
    alpha = give2DRotationalAngleStrain(epsVec);
  } else {
    alpha = status->giveTempAlpha();
  }
  Tstrain = giveTransformationMatrixStrain(alpha);
  Tstress = giveTransformationMatrixStress(alpha);
  dprincStrain.beProductOf(Tstrain,depsVec);

  this->initSLA(gp,totalStrain,tStep);

  FloatArray dPStressVec(rsize);
  double Edum;

  // only the pricipal directions are updated - the rest is not used
  for( int i=1; i<=dimension; i++ ) {
    Edum = status->giveCurrentLSStiffnessRatio(i) * E;
    dPStressVec.at(i) = Edum * dprincStrain.at(i);
  }

  return dPStressVec;
}

double
SHCCMaterialSla :: giveShiftedStrain( bool tensileLoad, int direction, SHCCMaterialStatus *status )
{
  double _shiftedStrain = princStrain.at(direction);

  if( tensileLoad ) {
    _shiftedStrain -= (1.0-bC_closing) * status->giveTempEpsCul().at(direction);

  } else {
    _shiftedStrain -= (1.0-bT_closing) * status->giveTempEpsTul().at(direction);
  }

  return( _shiftedStrain );
}

double
SHCCMaterialSla :: giveEpsTul( double eps, double sig, int direction, SHCCMaterialStatus *status )
{
  // eps, sig => point on the failure envelope
  double dum;
  double epsTul_old = status->giveEpsTul().at(direction);
  const double ET2 = (sigTp-E*epsT0)/(epsTp-epsT0);
  double sigT = E * ((1+this->giveStrengthRange())*E*epsT0-ET2*epsT0)/(E-ET2);


  if( eps <= sigT/E )
    dum = 0.0;
  else
    dum = bT*(eps-sigT/E);


  return( max(dum,epsTul_old) );
}


double
SHCCMaterialSla :: giveEpsCul( double eps, double sig, int direction, SHCCMaterialStatus *status )
{
  // eps, sig => point on the failure envelope
  double dum;
  double epsCul_old = status->giveEpsCul().at(direction);
  const double EC2 = (sigCp-E*epsC0)/(epsCp-epsC0);
  double sigC = E * ((1+this->giveStrengthRange())*E*epsC0-EC2*epsC0)/(E-EC2);


  if( eps >= sigC/E )
    dum = 0.0;
  else
    dum = bC*(eps-sigC/E);


  return( min(dum,epsCul_old) );

}

//----------------------------------------------------------------------
MaterialStatus *
SHCCMaterialSla :: CreateStatus(GaussPoint *gp) const
{
  SHCCMaterialStatusSla *status = new SHCCMaterialStatusSla(1, this->giveDomain(), gp);
  return status;
}


MaterialStatus *
SHCCMaterialSla :: giveStatus(GaussPoint *gp) const
{
  MaterialStatus *status = static_cast< MaterialStatus * >( gp->giveMaterialStatus() );
  if ( status == NULL ) {
    // create a new one
    status = this->CreateStatus(gp);

    if ( status != NULL ) {
      gp->setMaterialStatus( status, this->giveNumber() );
    }
  }

  return status;
}


double
SHCCMaterialSla :: giveNu(SHCCMaterialStatusSla *status)
{
  for( int i = 1; i < dimension; i++ ) {
    if( status->giveTempActiveState().at(i) == SHCCMaterialStatusSla::LS_crackClosing ) {
      return 0.0;
    }
  }

  return nu;
}

/*******************************************************/
SHCCMaterialStatusSla :: SHCCMaterialStatusSla(int n, Domain *d, GaussPoint *g) : SHCCMaterialStatus(n, d, g),
  SlaMaterialStatusExtensionInterface(n,d,g,StructuralMaterial::giveSizeOfVoigtSymVector(g->giveMaterialMode()),4*d->giveNumberOfSpatialDimensions()),
  LSStiffnessRatio(), LSChange(),
  princStressLastIter(), EdforEpsTp(), EdforEpsCp()
{

  EdforEpsTp.resize(d->giveNumberOfSpatialDimensions());
  EdforEpsCp.resize(d->giveNumberOfSpatialDimensions());
  LSChange.resize(d->giveNumberOfSpatialDimensions());
  for( int i=1; i<=d->giveNumberOfSpatialDimensions(); i++ ) {
    activeState.at(i) = SHCCsla_LS::LS_none;
    tempActiveState.at(i) = SHCCsla_LS::LS_none;
    LSChange.at(i) = 0;
  }

  LSStiffnessRatio.resize(3,d->giveNumberOfSpatialDimensions());
  for( int i=1; i<=LSStiffnessRatio.giveNumberOfRows(); i++ ) {
    for( int j=1; j<=LSStiffnessRatio.giveNumberOfColumns(); j++ ) {
      LSStiffnessRatio.at(i,j) = 1.0;
    }
  }

  princStressLastIter.resize(StructuralMaterial::giveSizeOfVoigtSymVector(g->giveMaterialMode()));

}


IntArray
SHCCMaterialStatusSla :: giveBoundsForLS( SHCCsla_LS LS )
{
  return giveBoundsForLS(static_cast<int>(LS));
}


IntArray
SHCCMaterialStatusSla :: giveBoundsForLS( int i )
{
  IntArray answer(2);
  switch( i )
  {
  case LS_none: {
    answer.at(1) = bound_tension;
    answer.at(2) = bound_compression;
    break;
  }
  case LS_tension: {
    answer.at(1) = bound_tension;
    answer.at(2) = bound_unloadTen;
    break;
  }
  case LS_crackClosing: {
    answer.at(1) = bound_unloadTen;
    answer.at(2) = bound_unloadComp;
    break;
  }
  case LS_compression: {
    answer.at(1) = bound_unloadComp;
    answer.at(2) = bound_compression;
    break;
  }
  default:
    OOFEM_ERROR( "Unsupported loading state" )
  }

  return answer;
}

FloatArray
SHCCMaterialStatusSla :: giveStrengthInDir( int dim )
{
  FloatArray answer(bound_count);
  int shift = bound_count * (dim-1);
  IntArray idx = { bound_tension, bound_unloadTen, bound_unloadComp, bound_compression };
  idx.add(shift);

  answer.beSubArrayOf(strengthSLA, idx);
  return answer;
}

void
SHCCMaterialStatusSla :: printOutputAt(FILE *file, TimeStep *tStep)
{
  StructuralMaterialStatus :: printOutputAt(file, tStep);
  fprintf(file, "status { alpha %e ", this->alpha);
  fprintf(file, "damage " );
  for( int i=1; i<=damage.giveSize(); i++ ) {
    fprintf(file, "%f ", this->damage.at(i));
  }
  fprintf(file, "}\n");
}

} // end namespace oofem
