/*
*
*                 #####    #####   ######  ######  ###   ###
*               ##   ##  ##   ##  ##      ##      ## ### ##
*              ##   ##  ##   ##  ####    ####    ##  #  ##
*             ##   ##  ##   ##  ##      ##      ##     ##
*            ##   ##  ##   ##  ##      ##      ##     ##
*            #####    #####   ##      ######  ##     ##
*
*
*             OOFEM : Object Oriented Finite Element Code
*
*               Copyright (C) 1993 - 2011   Borek Patzak
*
*
*
*       Czech Technical University, Faculty of Civil Engineering,
*   Department of Structural Mechanics, 166 29 Prague, Czech Republic
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef shccsla_h
#define shccsla_h

#include "shccmat.h"
#include "slamaterialextensioninterface.h"

///@name Input fields for IsotropicDamageMaterial1
//@{
#define _IFT_SHCCMaterialSla_Name "shccsla"
//@}

#define SHCCsla_minStressRatio 1.e-2 // +/-SHCCsla_minStressRatio*E*epsT0 = stress where the crack closing/opening starts
#define SHCCsla_minStressRatioDegadation 1.0 // degradates the SHCCsla_minStressRatio when cyclic loading happens
#define SHCCsla_clossingPart 1 // if 0 - no closing part of unloading is defined otherwise it is

namespace oofem {
  /**
  * This class implements SLA extension of associated MaterialStatus for SHCCMaterialsla.
  */
  class SHCCMaterialStatusSla : public SHCCMaterialStatus, public SlaMaterialStatusExtensionInterface
  {
  private:
    #define SHCCsla_simplifiedCompression 1

  public:
    enum SHCCsla_bound {bound_none = 0, bound_tension = 1, bound_unloadTen = 2, bound_unloadComp = 3, bound_compression = 4, bound_count = 4 };
    enum SHCCsla_LS { LS_none = 0, LS_tension = 1, LS_crackClosing = 2, LS_compression = 3 }; // loading states

  protected:
    FloatMatrix LSStiffnessRatio;
    IntArray LSChange;
    FloatArray princStressLastIter;
    FloatArray EdforEpsTp;
    FloatArray EdforEpsCp;

  public:
    /// Constructor
    SHCCMaterialStatusSla(int n, Domain *d, GaussPoint *g);
    /// Destructor
    virtual ~SHCCMaterialStatusSla(){}

    // definition
    virtual const char *giveClassName() const { return "SHCCMaterialStatusSla"; }
    virtual void printOutputAt(FILE *file, TimeStep *tStep);

    double giveTempDamageAt(int dim) { return this->giveDamage().at(dim); }

    int giveNumberOfBounds() { return bound_count; }
    IntArray giveBoundsForLS( SHCCsla_LS LS );
    IntArray giveBoundsForLS( int i );
    FloatMatrix &giveLSStiffnessRatio() { return LSStiffnessRatio; }
    double giveCurrentLSStiffnessRatio( int dim ) { return (tempActiveState.at(dim)==bound_none) ? 1.0 : LSStiffnessRatio.at(tempActiveState.at(dim),dim); }
    FloatArray giveLSStiffnessRatioInDir( int dim ) { FloatArray answer; LSStiffnessRatio.copyColumn(answer, dim); return answer; }
    int giveStrengthIndex( int bound, int dim ) { return bound+(dim-1)*bound_count; }
    FloatArray giveStrengthInDir( int dim );
    IntArray &giveLSChange() { return LSChange; }
    bool isLSChange() { for(int i=1; i<=LSChange.giveSize(); i++) {if(LSChange.at(i)!=0) return true;} return false;}
    FloatArray &givePrincStressLastIter() { return princStressLastIter; }
    FloatArray &giveEdforEpsTp() { return EdforEpsTp; }
    FloatArray &giveEdforEpsCp() { return EdforEpsCp; }

    void letLSStiffnessRatioBe( FloatMatrix &v ) { LSStiffnessRatio = v; }
    void letLSStiffnessRatioAtBe( double val, int LS, int dim ) { LSStiffnessRatio.at(LS,dim) = val; }
    void letLSChangeAtBe( int val, int dim ) { LSChange.at(dim) = val; }
    void letPrincStressLastIter( FloatArray &v ) { princStressLastIter = v; }
    void letEdforEpsTpAtBe( double val, int dim ) { EdforEpsTp.at(dim) = val; }
    void letEdforEpsCpAtBe( double val, int dim ) { EdforEpsCp.at(dim) = val; }

  };

  /**
  * This class implements SLA extension of SHCC material in a finite element problem.
  *
  * SHCC is NonLinear elasto-plastic material with damage model of SHCC
  * Cracking is described using Rotating Crack Model based maximum crack opening criterion.
  */
  class SHCCMaterialSla : public SHCCMaterial, public SlaMaterialExtensionInterface
  {
  private:
      void givePlaneStressStiffMtrxPrincDir(FloatMatrix &answer, MatResponseMode mode, GaussPoint *gp, TimeStep *atTime);

      void loadStateTension(FloatArray &pStressVec, int dir, GaussPoint *gp, SHCCMaterialStatusSla *status);
      void loadStateCompression(FloatArray &pStressVec, int dir, GaussPoint *gp, SHCCMaterialStatusSla *status);
      void loadStateCrackClosing(FloatArray &pStressVec, int dir, GaussPoint *gp, SHCCMaterialStatusSla *status);

      double giveNewStrengthTension(int dir, double eps, double sig, GaussPoint *gp, SHCCMaterialStatusSla *status );
      double giveNewStrengthCompression(int dir, double eps, double sig, GaussPoint *gp, SHCCMaterialStatusSla *status);
      FloatArray givePrincStressIncrement(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep);

      double giveMinStressForCrackClosing() { return SHCCsla_minStressRatio*E*epsT0; }
      double giveNu(SHCCMaterialStatusSla *status);

  protected:
      virtual double giveShiftedStrain( bool tensileLoad, int direction, SHCCMaterialStatus *status );
      virtual double giveEpsTul( double eps, double sig, int direction, SHCCMaterialStatus *status );
      virtual double giveEpsCul( double eps, double sig, int direction, SHCCMaterialStatus *status );

  public:
    /// Constructor
    SHCCMaterialSla(int n, Domain *d);
    /// Destructor
    virtual ~SHCCMaterialSla();

    // identification and auxiliary functions
    virtual IRResultType initializeFrom(InputRecord *ir);
    virtual const char *giveInputRecordName() const { return _IFT_SHCCMaterialSla_Name; }
    virtual const char *giveClassName() const { return "SHCCMaterialSla"; }

    virtual MaterialStatus *CreateStatus(GaussPoint *gp) const;
    virtual MaterialStatus *giveStatus(GaussPoint *gp) const;

    virtual void givePlaneStressStiffMtrx(FloatMatrix &answer,
                                          MatResponseMode mode,
                                          GaussPoint *gp,
                                          TimeStep *atTime);

    virtual void giveRealStressVector(FloatArray &answer, GaussPoint *gp,
                                      const FloatArray &totalStrain, TimeStep *tStep);


    // extension
    virtual double computeLambda(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep, double minLambda = INFINITY);
    virtual void initSLA(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep);
    virtual void updateHistoryParameters(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep, bool activeDamage);

  };
} // end namespace oofem
#endif // shccmat_h
