/*
 *
 *                 #####    #####   ######  ######  ###   ###
 *               ##   ##  ##   ##  ##      ##      ## ### ##
 *              ##   ##  ##   ##  ####    ####    ##  #  ##
 *             ##   ##  ##   ##  ##      ##      ##     ##
 *            ##   ##  ##   ##  ##      ##      ##     ##
 *            #####    #####   ##      ######  ##     ##
 *
 *
 *             OOFEM : Object Oriented Finite Element Code
 *
 *               Copyright (C) 1993 - 2013   Borek Patzak
 *
 *
 *
 *       Czech Technical University, Faculty of Civil Engineering,
 *   Department of Structural Mechanics, 166 29 Prague, Czech Republic
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "slamaterialextensioninterface.h"
#include "structuralinterfacematerialstatus.h"
#include "structuralms.h"
#include "dynamicinputrecord.h"

namespace oofem {

SlaMaterialExtensionInterface::SlaMaterialExtensionInterface()
{
    strengthRange = 0.1; // range should be > 1 && range < Inf
}


IRResultType
SlaMaterialExtensionInterface :: initializeFrom(InputRecord *ir)
{
    IRResultType result;                // Required by IR_GIVE_FIELD macro

    double dum = -1.0;
    IR_GIVE_OPTIONAL_FIELD(ir, dum, _IFT_SlaMaterialExtensionInterface_strengthRange);
    if( dum > 0.0 && dum < INFINITY ) {
        letStrengthRangeBe(dum);
    }

    return IRRT_OK;
}


void
SlaMaterialExtensionInterface :: giveInputRecord(DynamicInputRecord &input)
{
    input.setField(this->strengthRange, _IFT_SlaMaterialExtensionInterface_strengthRange);
}

void
SlaMaterialExtensionInterface :: updateStressDrop(GaussPoint *gp, TimeStep *tStep)
{
    SlaMaterialStatusExtensionInterface *status1 = dynamic_cast< SlaMaterialStatusExtensionInterface * >( gp->giveMaterialStatus() );
    StructuralMaterialStatus *status2 = dynamic_cast< StructuralMaterialStatus * >( gp->giveMaterialStatus() );
    if( status2 ) {
      FloatArray stress(status2->giveTempStressVector());
      stress.add(status1->giveDStressVector());

      status2->letTempStressVectorBe(stress);

      stress.zero();
      status1->letDStressVectorBe(stress);

      status2->updateYourself(tStep);

      return;
    }

    StructuralInterfaceMaterialStatus *status3 = dynamic_cast< StructuralInterfaceMaterialStatus * >( gp->giveMaterialStatus() );
    if( status3 ) {
      FloatArray traction3D(status3->giveTempJump());
      traction3D.add(status1->giveDStressVector());

      status3->letTempTractionBe(traction3D);

      traction3D.zero();
      status1->letDStressVectorBe(traction3D);

      status3->updateYourself(tStep);

      return;
    }


    OOFEM_ERROR( "Function must be redefined to take into account the given element status" );

}




//----------------------------------------------------------------------
SlaMaterialStatusExtensionInterface :: SlaMaterialStatusExtensionInterface(int n, Domain *d, GaussPoint *g, int numOfStresses, int numOfLimits) :
    strengthSLA(), dStressVector()
{
    dStressVector.resize(numOfStresses);

    strengthSLA.resize(numOfLimits);
    for( int i = 1; i<=numOfLimits; i++ ) {
        strengthSLA.at(i) = -1.0;
    }
}

}
