/*
 *
 *                 #####    #####   ######  ######  ###   ###
 *               ##   ##  ##   ##  ##      ##      ## ### ##
 *              ##   ##  ##   ##  ####    ####    ##  #  ##
 *             ##   ##  ##   ##  ##      ##      ##     ##
 *            ##   ##  ##   ##  ##      ##      ##     ##
 *            #####    #####   ##      ######  ##     ##
 *
 *
 *             OOFEM : Object Oriented Finite Element Code
 *
 *               Copyright (C) 1993 - 2013   Borek Patzak
 *
 *
 *
 *       Czech Technical University, Faculty of Civil Engineering,
 *   Department of Structural Mechanics, 166 29 Prague, Czech Republic
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef slamaterialextensioninterface_h
#define slamaterialextensioninterface_h

///@name slamaterialextensioninterface
//@{
#define _IFT_SlaMaterialExtensionInterface_strengthRange "strengthrange"
//@}

#include "gausspoint.h"

namespace oofem {
#define EPS_SLA 1.e-10
#define MAX_ITER_SLA 1e3

class SlaMaterialStatusExtensionInterface
{
/**
 * This class implements associated extension for Sequentially Linear Static Engineering problem.
 *
 */
protected:
    FloatArray strengthSLA;
    FloatArray dStressVector; // stress jump caused by the sudden decrease of stiffness

public:
    /// Constructor
    SlaMaterialStatusExtensionInterface(int n, Domain * d, GaussPoint * g, int numOfStresses, int numOfLimits);
    /// Destructor
    virtual ~SlaMaterialStatusExtensionInterface() { }

    virtual const char *giveClassName() const { return "SlaMaterialStatusExtensionInterface"; }

    double giveStrengthAt(const int i) const
    { return strengthSLA.at(i); }
    void letStrengthAtBe( const double val, const int i )
    { strengthSLA.at(i) = val; }
    const FloatArray &giveStrength() const
    { return strengthSLA; }
    void letStrengthBe( const FloatArray &v )
    { strengthSLA = v; }

    /// Returns the const pointer to receiver's dStressVector.
    const FloatArray &giveDStressVector() const { return dStressVector; }
    /// Assigns dStressVector to given vector v.
    void letDStressVectorBe(const FloatArray &v) { dStressVector = v; }

};

class SlaMaterialExtensionInterface
{

/**
 * This class implements function for Sequentially Linear Static Engineering problem,
 * e.i. extension of materials for SLS solver
 *
 */
private:
    double strengthRange;

public:
    /// Constructor
    SlaMaterialExtensionInterface();
    /// Destructor
    virtual ~SlaMaterialExtensionInterface(){}

    double giveStrengthRange() const
    { return strengthRange; }
    void letStrengthRangeBe( const double val )
    { strengthRange = val; }

    virtual IRResultType initializeFrom(InputRecord *ir);
    virtual void giveInputRecord(DynamicInputRecord &input);
    virtual const char *giveClassName() const { return "SlaMaterialExtensionInterface"; }

    // functions needed to be defined in SLA materials
    virtual double computeLambda(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep, double minLambda = INFINITY) = 0;
    virtual void initSLA(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep) = 0;
    virtual void updateHistoryParameters(GaussPoint *gp, FloatArray &totalStrain, TimeStep *tStep, bool activeDamage) = 0;
    // optional
    virtual void updateStressDrop(GaussPoint *gp, TimeStep *tStep);
};
}  // end namespace oofem
#endif // slamaterialextensioninterface_h
